with Ada.Text_IO;
use Ada.Text_IO;

procedure Convert is

  subtype Scalar is Long_Long_Float;
  type ScalVec is array(Integer range <>) of Scalar;
  type ScalMat is array (Integer range <>, Integer range <>) of Scalar;

  package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
  package Txt_Scal_IO is new Ada.Text_IO.Float_IO(Scalar);
  use TXT_Int_IO, Txt_Scal_IO;

  function Max(A,B: Scalar) return Scalar renames Scalar'Max;

  procedure SetZero(P: in out ScalMat) is
  begin
    for I in P'Range(1) loop
      for J in P'Range(2) loop
        P(I,J) := Scalar(0);
      end loop;
    end loop;
  end SetZero;

  procedure Show(P: in ScalMat) is
  begin
    for I in P'Range(1) loop
      Put_Line("-------------------------------------------------");
      for J in P'Range(2) loop
        if P(I,J) /= Scalar(0) then
          Put("P(" & I'Img & "," & J'Img & ") := ");  -- gnat specific
          Put(Standard_Output,P(I,J));
          Put_Line(";");
        end if;
      end loop;
    end loop;
  end Show;

--- now stuff from [EKW]

  N: constant Integer := 20;
  N1: constant Integer := N+1;
  N2: constant Integer := N+2;
  N3: constant Integer := N+3;
  NN: constant Integer := N1*N2/2;
  NH: constant Integer := NN+1;
  NG: constant Integer := NN+2;
  NC: constant Integer := (N*N+6*N+4)/4;

  SZER:   constant Scalar := 0.0;
  SONE:   constant Scalar := 1.0;
  STWO:   constant Scalar := 2.0;
  SFIVE:  constant Scalar := 5.0;
  STEN:   constant Scalar := 10.0;
  SHALF:  constant Scalar := SONE/STWO;
  SFIFTH: constant Scalar := SONE/SFIVE;

  procedure FZERO(VB: in out ScalVec) is
  begin
    for K in VB'Range loop VB(K) := SZER; end loop;
  end FZERO;

  function SVALUE(V: in Scalvec; SX,SY: in Scalar) return Scalar is
    K: Integer;
    SZ,SZT: Scalar;
  begin
--- ITEST stuff left out
    SZ := SZER;
    K := NN;
    SZT := V(K);
    for I in 1 .. N loop
      SZ := SZ*SY+SZT;
      K := K-1;
      SZT := V(K);
      for J in 1 .. I loop
        K := K-1;
        SZT := SX*SZT+V(K);
      end loop;
    end loop;
    return SZ*SY+SZT+V(NH)*(Max(abs(SX),abs(SY))**N1)+V(NG);
  end SVALUE;

  procedure DERX(VF: in ScalVec; VDF: out ScalVec) is
    KDF,KF: Integer;
  begin
    KDF := 1;
    KF := 2;
    for I in 1 .. N loop
      for J in 1 .. N1-I loop
        VDF(KDF) := Scalar(J)*VF(KF);
        KDF := KF;
        KF := KF+1;
      end loop;
      VDF(KDF) := SZER;
      KDF := KF;
      KF := KF+1;
    end loop;
    VDF(KDF) := SZER;
    VDF(NH) := SZER;
    VDF(NG) := SZER;
  end DERX;

  procedure DERY(VF: in ScalVec; VDF: out ScalVec) is
    KDF,KF: Integer;
  begin
    KDF := 1;
    KF := N2;
    for I in 1 .. N loop
      for J in 1 .. N1-I loop
        VDF(KDF) := Scalar(I)*VF(KF);
        KDF := KDF+1;
        KF := KF+1;
      end loop;
      VDF(KDF) := SZER;
      KDF := KDF+1;
    end loop;
    VDF(KDF) := SZER;
    VDF(NH) := SZER;
    VDF(NG) := SZER;
  end DERY;

  function SDYVALUE(V: in Scalvec; SX,SY: in Scalar) return Scalar is
    VDY: ScalVec(V'Range);
  begin
--- ITEST stuff left out
    DERY(V,VDY);
    return SVALUE(VDY,SX,SY);
  end SDYVALUE;

  procedure PROJ(V: in ScalVec; Alpha,Tau: in Scalar; P10,P01,PY10,PY01: out Scalar) is
    X0,X1: Scalar;
  begin
    X0 := -Alpha/Tau;
    X1 := (SONE-Alpha)/Tau;
    P10 := SVALUE(V,X1,X0);
    P01 := SVALUE(V,X0,X1);
    PY10 := SDYVALUE(V,X1,X0)/Tau;
    PY01 := SDYVALUE(V,X0,X1)/Tau;
  end PROJ;

  procedure UNPACK(VA: in ScalVec; Alpha,Tau: in Scalar; VB: out ScalVec) is
    IP,K,KD,IJ,KK,KKD,KH,LS,LE: Integer;
    A1,A2,P10,P01,DP10,DP01,X1,X0,DD,CC,BB,AA: Scalar;
  begin
    FZERO(VB);
    VB(NH) := VA(NC);
    IP := 1;
    K := N2+N;
    KD := N-1;
    for I in 3 .. N1 loop
      VB(K) := VA(IP);
      K := K+KD;
      KD := KD-1;
      IP := IP+1;
    end loop;
    IJ := N3/2;
    K := 2;
    KD := N2;
    for J in 2 .. IJ loop
      KK := K;
      KKD := KD-1;
      KH := K;
      A2 := Scalar(J-1);
      LS := J-1;
      LE := N2-J;
      for L in LS .. LE loop
        if not ((L<3) and (J=2)) then
          A1 := Scalar(L);
          VB(KK) := VA(IP)/(SONE+A2/A1);
          VB(KH) := VB(KK)*A2/A1;
          if KH=KK then VB(KH) := VA(IP); end if;
          IP := IP+1;
        end if;
        KK := KK+KKD;
        KKD := KKD-1;
        KH := KH+1;
      end loop;
      K := K+KD;
      KD := KD-1;
    end loop;
    PROJ(VB,Alpha,Tau,P10,P01,DP10,DP01);
    X1 := (SONE-Alpha)/Tau;
    X0 := -Alpha/Tau;
    DD := (DP01-DP10)*Tau*Tau*SHALF;
    CC := -Tau*DP10+DD*X1*STWO;
    BB := (P01-P10)*Tau+(-DD)*(X1+X0)+CC;
    AA := P10+BB*X1+CC*X0;
    AA := -(AA+DD*X1*(STWO*X0+X1));
    VB(1) := VB(1)+AA;
    VB(2) := VB(2)+BB;
    VB(N2) := VB(N2)+CC;
    VB(3) := VB(3)+DD;
    VB(N3) := VB(N3)+DD*STWO;
  end UNPACK;

  function Input return ScalVec is
    D: ScalVec(1 .. NC);              -- is 128 in [EKW]
  begin
    FZERO(D);

    D(1):=    -0.302218250463128873E-001;
    D(2):=     0.154830664195353367E-002;
    D(3):=    -0.129578370556081318E-003;
    D(4):=     0.108523586647550293E-004;
    D(5):=    -0.987372844807154123E-006;

    D(6):=     0.935463250541952263E-007;
    D(7):=    -0.915390414245415699E-008;
    D(8):=     0.917640087941869132E-009;
    D(9):=    -0.937621886025294370E-010;
    D(10):=    0.972897515408567785E-011;

    D(11):=   -0.102231566401946393E-011;
    D(12):=    0.108518062717437766E-012;
    D(13):=   -0.115942535488959455E-013;
    D(14):=    0.123599515512696595E-014;
    D(15):=   -0.128585000700047043E-015;

    D(16):=    0.124514720295842190E-016;
    D(17):=   -0.103253974956103418E-017;
    D(18):=    0.639459674649833898E-019;
    D(19):=   -0.221947015573720828E-020;
    D(20):=    0.791742899101695183E-002;

    D(21):=   -0.741859716302175834E-003;
    D(22):=    0.750979401365909050E-004;
    D(23):=   -0.789543633541830282E-005;
    D(24):=    0.851409854814131693E-006;
    D(25):=   -0.934137538514596987E-007;

    D(26):=    0.103792560799945019E-007;
    D(27):=   -0.116433094642290927E-008;
    D(28):=    0.131593518407010228E-009;
    D(29):=   -0.149612482867518388E-010;
    D(30):=    0.170853124534413310E-011;

    D(31):=   -0.195425608391321948E-012;
    D(32):=    0.222107065134780887E-013;
    D(33):=   -0.245419372762764689E-014;
    D(34):=    0.251247438081158377E-015;
    D(35):=   -0.218208096671618356E-016;

    D(36):=    0.138215493093105955E-017;
    D(37):=   -0.463194073132168188E-019;
    D(38):=    0.626224565328406937E-002;
    D(39):=   -0.160675645442736537E-002;
    D(40):=    0.204716420599186307E-003;

    D(41):=   -0.259467491588060176E-004;
    D(42):=    0.327320570884602615E-005;
    D(43):=   -0.411205831371326772E-006;
    D(44):=    0.514710323846259959E-007;
    D(45):=   -0.642219993765179074E-008;

    D(46):=    0.799081151329943560E-009;
    D(47):=   -0.991748882180860719E-010;
    D(48):=    0.122762272148014029E-010;
    D(49):=   -0.151302549229142636E-011;
    D(50):=    0.184369284522256626E-012;

    D(51):=   -0.217535815466223768E-013;
    D(52):=    0.237023192465046574E-014;
    D(53):=   -0.218477801082521684E-015;
    D(54):=    0.146481789444569259E-016;
    D(55):=   -0.518208022232986191E-018;

    D(56):=    0.140548932455512415E-003;
    D(57):=   -0.452915480150940848E-004;
    D(58):=    0.693100354025415795E-005;
    D(59):=   -0.102351969771038658E-005;
    D(60):=    0.147227644402519850E-006;

    D(61):=   -0.207537155249882862E-007;
    D(62):=    0.287876987339816273E-008;
    D(63):=   -0.394077781101782018E-009;
    D(64):=    0.533343923174530136E-010;
    D(65):=   -0.713501356833401317E-011;

    D(66):=    0.938041898760630842E-012;
    D(67):=   -0.118823208969553820E-012;
    D(68):=    0.138435550860119163E-013;
    D(69):=   -0.135977095938542519E-014;
    D(70):=    0.968451999492137331E-016;

    D(71):=   -0.362824768103934656E-017;
    D(72):=    0.441568418143338155E-005;
    D(73):=   -0.158606941711927689E-005;
    D(74):=    0.268805158807486310E-006;
    D(75):=   -0.436322206017262056E-007;

    D(76):=    0.684939719910452141E-008;
    D(77):=   -0.104682361536778030E-008;
    D(78):=    0.156466072603692124E-009;
    D(79):=   -0.229122650076956052E-010;
    D(80):=    0.327324070519198133E-011;

    D(81):=   -0.447853420756558709E-012;
    D(82):=    0.560831487210665231E-013;
    D(83):=   -0.589667813699498405E-014;
    D(84):=    0.447886551807202722E-015;
    D(85):=   -0.178324075195023030E-016;

    D(86):=    0.163586969339542905E-006;
    D(87):=   -0.626359250225299106E-007;
    D(88):=    0.113343750494654797E-007;
    D(89):=   -0.196244073310274145E-008;
    D(90):=    0.327794308704771235E-009;

    D(91):=   -0.530586837100769928E-010;
    D(92):=    0.830439595632213275E-011;
    D(93):=   -0.123581485933800022E-011;
    D(94):=    0.167324109563134508E-012;
    D(95):=   -0.189271473107272035E-013;

    D(96):=    0.153993821616953699E-014;
    D(97):=   -0.654120828619607172E-016;
    D(98):=    0.668653115380598254E-008;
    D(99):=   -0.267045827599544460E-008;
    D(100):=   0.505616944454200397E-009;

    D(101):=   -0.915162759778108392E-010;
    D(102):=    0.158429307720013615E-010;
    D(103):=   -0.258483170627742805E-010;
    D(104):=    0.380947666251526417E-012;
    D(105):=   -0.466259898814247500E-013;

    D(106):=    0.408365445469242406E-014;
    D(107):=   -0.185866559520685867E-015;
    D(108):=    0.291630529962384718E-009;
    D(109):=   -0.119702801334653713E-009;
    D(110):=    0.231810283391865535E-010;

    D(111):=   -0.418527993466551614E-011;
    D(112):=    0.676636874219246717E-012;
    D(113):=   -0.901977627407920390E-013;
    D(114):=    0.855168134438476962E-014;
    D(115):=   -0.419102891528817993E-015;

    D(116):=    0.131439002736733006E-010;
    D(117):=   -0.531059160446456976E-011;
    D(118):=    0.950426601207490573E-012;
    D(119):=   -0.139035251649604742E-012;
    D(120):=    0.143615663695012141E-013;

    D(121):=   -0.762065976759714626E-015;
    D(122):=    0.531784274011477496E-012;
    D(123):=   -0.172266563961423343E-012;
    D(124):=    0.195312992445814170E-013;
    D(125):=   -0.112927963043820731E-014;

    D(126):=    0.108133294048250410E-013;
    D(127):=   -0.137259515463036970E-014;

    return D;
  end Input;

  subtype Poly is ScalMat(0 .. N,0 .. N);

  function MakePoly(V: Scalvec) return Poly is
    K,DX,DY: Integer;
    P: Poly;
  begin
    SetZero(P);
    K := NN;
    DY := N;
    DX := 0;
--    Put_Line("------------------------------");
--    Put_Line("P(" & DX'Img & "," & DY'Img & ") = V(" & K'Img & ")");
    P(DX,DY) := V(K);
    for I in 1 .. N loop
      K := K-1;
      DY := N-I;
      DX := I;
--      Put_Line("------------------------------");
--      Put_Line("P(" & DX'Img & "," & DY'Img & ") = V(" & K'Img & ")");
      P(DX,DY) := V(K);
      for J in 1 .. I loop
        K := K-1;
        DX := I-J;
--        Put_Line("P(" & DX'Img & "," & DY'Img & ") = V(" & K'Img & ")");
        P(DX,DY) := V(K);
      end loop;
    end loop;
    return P;
  end MakePoly;

  Alpha: constant Scalar := SHALF;
  Tau:   constant Scalar := Scalar(8)*SFIFTH;
  VPHI0: constant ScalVec := Input;

  VSIG,DxVSIG: ScalVec(1 .. 233);
  P: Poly;

begin

  UNPACK(VPHI0,SZER,SONE,VSIG);
  P := MakePoly(VSIG);
  Show(P);

--  DERX(VSIG,DxVSIG);
--  P := MakePoly(DxVSIG);
--  Show(P);

end Convert;
