with Reps, Scalars, Scalars.IO, Vectors, Vectors.Num, Taylors, Taylors.IO,Ada.Text_IO;
use Reps,Ada.Text_IO;

generic

   PDeg: in Natural;
   CutOff: in Natural;
   with package Numbers is new Scalars (<>);
   with package Numbers_IO is new Numbers.IO (<>);

package Doubling is

  use Numbers, Numbers_IO;

  package Tay is new Taylors (PDeg => PDeg,CutOff=>CutOff, Numbers => Numbers);
  package Tay_IO is new Tay.IO (Numbers_IO => Numbers_IO);
  use Tay;

  function CompZ(F1,F2,Z: Taylor) return Taylor;                        -- F1(Z,.)+F2(.,Z)
  function DCompZ(R: Radius; F1,F2,Z: Taylor) return Taylor;            -- D1F1(Z,.)+D2F2(.,Z)
  procedure CompZero(R: in Radius; F1,F2: in Taylor; Z: in out Taylor); -- solve F1(Z,.)+F2(.,Z)=0

  --procedures for symmetric functions
  function CompZ(S,X,Y,Z: Taylor) return Taylor;                        -- S(X,Z)+S(Y,Z)
  function DCompZ(R: Radius; S,X,Y,Z: Taylor) return Taylor;            -- D2S(X,Z)+D2S(Y,Z)
  procedure CompZero(R: in Radius; S,X,Y: in Taylor; Z: in out Taylor); -- solve S(X,Z)+S(Y,Z)=0
  procedure FindLambda(R: in Radius; S: in Taylor; A: in Scalar; Lambda: in out Scalar); -- solve S(Lambda-A,1-A)+S(-A,1-A)=0
  procedure RG(R: in Radius; A: in Scalar; Lambda,Mu: in out Scalar; S,Z: in out Taylor  ); --S->R(S), normalization a-la Denis
  procedure RenG(R: in Radius; A: in Scalar; Lambda,Mu: in out Scalar; S,Z: in out Taylor  ); --S->R(S), normalization a-la EKW
  procedure DLambdah(R: in Radius; A,Lambda: in Scalar; S,H: in Taylor; DL: out Scalar); -- DL:=Lambda'[S].H
  function DLambdah(R: Radius; A,Lambda: Scalar; S,H: Taylor) return Scalar;
  procedure DMuh(R: in Radius; A,Lambda,DL: in Scalar; S,H,Z,DZ: in  Taylor; DM: out Scalar); --DM:=Mu'[S].H
  function DMuh(R: Radius; A,Lambda,DL: Scalar; S,H,Z,DZ: Taylor) return Scalar;
  procedure DZh(R: in Radius; A,Lambda,DL: in Scalar; S,Z: in Taylor; H: in out Taylor); --H ->Z'[S].H
  procedure DRGh(R: in Radius; A: in Scalar; Lambda: in out Scalar;  S: in  Taylor; Z,H: in out Taylor); --H -> DR[S].H
  procedure FindS(Eps: in Radius; N: in Integer; R: in Radius; A: in Scalar; Lambda: in out Scalar; S,Z: in out Taylor  ); -- Finds S, s.t. P_N[R(S)-S]=0
  procedure DRGh(N: in Integer; R: in Radius; A: in Scalar; Lambda: in out Scalar; S: in  Taylor; Z,H: in out Taylor); --DR[S+delta[S]]H
  procedure Integrate(R: in Radius; S: in out Taylor); --finds H, such that H_2(x,y)4S(x,y),H_1(x,y)=S(y,x)

  --ANALYTICITY AND COMPACTNESS
  procedure Z0_Range(RhoX: in Radius; P: in Scalar; A,B,C: in Taylor; RhoZ: out Radius); --range of Z_0
  procedure Sigma0_Norms(Rho,RhoX,RhoZ: in Radius; P: in Scalar; S: in Taylor; Const1,Const2,S_Nrm1,S_Nrm2,S_Der1,S_Der2: out Scalar);--ranges of Sigma_0 and it derivative
  procedure N_h(Rho,RhoX,H_Delt,Z_Delt: in Radius; A_Tay,B_Tay,C_Tay: in Taylor; P,A,B: in Scalar; M,N_Der,N_Diff: out Scalar); --Newton map for the operator Z(x,y) -> tau^{-1}[sigma(x,Z(x,y))+sigma(y,Z(x,y))],
                                                                                                                                --N_Der is a bound its derivative in a H_Delt nbhd of S_0, N_Diff=N_h[0]
  procedure Compactness(Rho,RhoX,H_Delt,Zn_Delt: in Radius; A_Tay,B_Tay,C_Tay: in Taylor; P,A,B: in Scalar; Z_Range: out Scalar); --produces the range of Z_S for the input data
  procedure Lambda(Rho,H_Delt: in Radius;A_Tay,B_Tay,C_Tay: in Taylor; K: in Scalar; Lambda:--bound on Lambda
                                                                                       out Scalar);
  procedure Mu(Rho,H_Delt,Z_Delt: in Radius;A,B,C,S: Taylor; K,Lambda: in Scalar; Mu: out Scalar); --bound on Mu
  procedure FindPhiLambda(Phi,Eps: in Taylor; Lambda: out Scalar);
  procedure FindPhiLambda(Eta: in Scalar; Phi,Eps1,Eps2: in Taylor; Lambda: in out Scalar);
  procedure PhiRG(Eps: in Taylor;Phi: in out Taylor;  Lambda: out Scalar);
  procedure PhiRG(Eta: in Scalar; Eps1,Eps2: in Taylor;Phi: in out Taylor;  Lambda: in out Scalar);
  function PhiRG(Eps,Phi: Taylor) return Taylor;
  procedure PhiDRG(N: in Integer; Eps,Phi: in Taylor;Er,Ei: out Scalar;  EvT: out Taylor; M_File: in String);
  procedure PhiDRGE(Eta: in Scalar; N: in Integer; Eps1,Eps2: in Taylor; Phi: in out Taylor; Lambda: in out Scalar);
  procedure PhiNewton(N: in Integer; Eps: in Taylor; Phi: in out Taylor; Lambda: out Scalar);
   procedure PhiNewton(Eta: in Scalar; N: in Integer; Eps1,Eps2: in Taylor; Phi: in out Taylor; Lambda: in out Scalar);

   procedure FindZ(Lambda: in Scalar; Eps,InvPhi: in Taylor; Z: out Taylor);
   procedure FindZ(Eta,Lambda0,Lambda: in Scalar; Eps1,Eps2,InvPhi: in Taylor; Z: out Taylor);


   procedure Newton(R1: in Radius; P: in out Scalar; Tay: in Taylor; E: in Scalar);
   procedure Newton(Bs,Cs,Es,P: in out Scalar; Tay: in Taylor; E: in Scalar);
   procedure NewtonMap(N: in Integer; Lambda,Mu: in Scalar; X,U: in out Scalar; Phi,S: in Taylor);
   procedure EigenData(R1: in Radius; P: in out Scalar; Tay: in Taylor; Er1,Er2,Ei1,Ei2,Ev11,Ev12,Ev21,Ev22: out Scalar);
   procedure EigenData(Bs,Cs,Es,P: in out Scalar; Tay: in Taylor; Er1,Er2,Ei1,Ei2,Ev11,Ev12,Ev21,Ev22: out Scalar);
   procedure Map(Phi,S: in Taylor; Lambda,Mu: in Scalar; S1,S2: in out Scalar);
   procedure Map(Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar; S1,S2: in out Scalar);
   procedure Map(Phi,S: in Taylor; F1,F2: out Taylor);
   procedure LinMapDer(T11,T12,T21,T22: in out Scalar; F1,F2: in  Taylor);
   procedure Distortion(Co,Q,N:in Integer; R1,R2: in Radius;
                                           ME1,ME2,LsP0,LsN0,LuP0,LuN0,LsP1,LsN1,LuP1,LuN1,B,C,E,Lambda,Mu,Sx0,Su0,Sx1,Su1,CenX,CenY,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1:
                                           in Scalar; Phi,S,Sigma,T1,T2: in Taylor;
                                                      Cstar,Epsilon,Gamma,DF11,LocDist,MinExp,MaxExp,MinAng,MaxAng: out Scalar; Hyp4: out Boolean);
   procedure JS(F: in File_Type; J: in Integer);
   procedure DrawHS(DoLin: in Boolean; NameT,NameB,NameR,NameL: in String;Q,N: in Integer; LsP,LsN,LuP,LuN,Lambda,Mu,Bs,Cs,Es,P1,P2,CenX,CenY,Wx,Wu,Vx,Vu: in Scalar; Phi,S,T1,T2: in Taylor);
   procedure DrawManifolds(NameU,NameS,NameU2,NameS2: in String; K,Q,N: in Integer;
                                                                 B,C,E,Lambda,Mu,Sx,Su,Rad: in Scalar; Phi,Sigma: in Taylor);
   procedure Linearize1(R1: in Radius; Phi,S: in Taylor;P,Lambda,Mu,IT11,IT12,IT21,IT22,Ev1,Ev2: in Scalar; Y,U: in out Scalar);
   procedure Linearize2(N: in Integer; R1: in Radius; Phi,S: in Taylor;P,Lambda,Mu,A,B,C,D: in Scalar; Y,U: in out Scalar);
   procedure Linearize3(N: in Integer; R1: in Radius; Phi,S: in Taylor;P,Lambda,Mu,A,B,C,D: in Scalar; Y,U: in out Scalar);
   procedure LinOp(P,A,B,C,D: in Scalar; F1,F2: in Taylor; T1,T2: in out Taylor);
   procedure LinCoord(Deg: in Integer; P,A,B,C,D: in Scalar; F1,F2: in Taylor; T1,T2: in out Taylor);

   procedure Coordinate(Name01,Name02,Name11,Name12: in String;Q: in Integer; LsP0,LsN0,LuP0,LuN0,LsP1,LsN1,LuP1,LuN1: in Rep; X0,U0,X1,U1,CenX,CenY,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1: in Scalar; T1,T2,V1,V2: in Taylor; G1,G2: out Taylor);

   procedure Manifold(R: in Radius; X0,U0,X1,U1,X2,U2,X3,U3,CenX,CenY,Sc: in Scalar; G1,G2: out Taylor);

   procedure Domain(Name1,Name2,Name3,Name4: in String; N: in Integer; R: in Radius; Phi,S: in Taylor; Lambda, Mu: in Scalar);

   procedure IterateDomain(Name1,Name2,Name3,Name4: in String; N,M: in Integer; R: in Radius; Phi,S: in Taylor; Lambda, Mu: in Scalar);

    procedure DrawInclusion0(Name1,Name2: in String; K: in Integer; Phi,S: in Taylor; Lambda,Mu,LsP0,LsN0,LuP0,LuN0,LsP1,LsN1,LuP1,LuN1,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1,CenX,CenY,Sx0,Su0,Sx1,Su1,Sx,Su,Rad,Ecc: in Scalar);
    procedure DrawInclusion1(Name1,Name2: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu,CenX,CenY,S1,S2,Rad,Ecc: in Scalar);
    procedure DrawInclusion2(Name1,Name2: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu,CenX,CenY,S1,S2,Rad,Ecc: in Scalar);
    procedure DrawInclusion3(Name1,Name2: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu,CenX,CenY,S1,S2,Rad,Ecc: in Scalar);
    procedure DrawInclusion4(Name0,Name1,Name2: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu,CenX,CenY,S1,S2,Rad,Ecc: in Scalar);
    procedure DrawInclusion5(Name1,Name2: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu,CenX,CenY,S1,S2,Rad,Ecc: in Scalar);
    procedure DrawInclusion6(Name1,Name2,Name3: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu,CenX,CenY,Sx1,Su1,Sx2,Su2,Rad1,Ecc1,Rad2,Ecc2: in Scalar);

    function SetY(J,Q,N: in Integer; LsP0,LsN0,LuP0,LuN0,LsP1,LsN1,LuP1,LuN1: in Rep; Lambda,Mu,Px0,Pu0,Px1,Pu1,CenX,CenY,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1: in Scalar; Phi,S: in Taylor) return Scalar;
   function SetU(J,Q,N: in Integer; LsP0,LsN0,LuP0,LuN0,LsP1,LsN1,LuP1,LuN1: in Rep; Lambda,Mu,Px0,Pu0,Px1,Pu1,CenX,CenY,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1: in Scalar; Phi,S: in Taylor) return Scalar;


    procedure FindPoint(J: in Integer; D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3: in Scalar;  Check: in out Boolean;JJ1: out Integer);
    procedure  DerBounds(N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar;CenX1,CenY1,Sx1,Su1,Rad1,Ecc1: in Scalar; MaxD1,MaxD3,MaxT,MinD: out Scalar);
   procedure Containment0(N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu,Sx0,Su0,Sx1,Su1,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1,LsN0,LsP0,LuN0,LuP0,LsN1,LsP1,LuN1,LuP1: in Scalar; CenX1,CenY1,S1,C1,Rad1,Ecc1: in Scalar);
   procedure Containment1(N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar; CenX1,CenY1,S1,C1,Rad1,Ecc1:
in Scalar);
   procedure Containment2(N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar; CenX1,CenY1,S1,C1,Rad1,Ecc1:
in Scalar);
   procedure Containment3(N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar; CenX1,CenY1,S1,C1,Rad1,Ecc1:
in Scalar);
   procedure Containment4(N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar; CenX1,CenY1,S1,C1,Rad1,Ecc1:
                                                                                       in Scalar);
   procedure Containment5(Name: in String; Q,N: in Integer; LsP,LsN,LuP,LuN: in Rep; Lambda,Mu,P1,P2,CenX,CenY,Wx,Wu,Vx,Vu,S1,C1,S2,C2,Ecc1,Ecc2,Rad1,Rad2: in Scalar; Phi,S: in Taylor);
   procedure Gaps(DoLin: in Boolean;NameP,NameCen,NameCenn: in String; Q,N: in Integer;K0bx,K0bu,K0tx,K0tu,K1bx,K1bu,K1tx,K1tu,Sc,Lip,B,C,E,Lambda,Mu,Px,Pu,Px0,Pu0,Px1,Pu1,CenX,CenY,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1,V1,V2: in Scalar;
                                                                                       Phi,S,Sigma,T1,T2: in Taylor;
                                                                                       MinL0,MinL1,MaxL01: out Scalar);

private

   ScalZero:  constant Scalar := Scal(Zero);
   ScalOne:  constant Scalar := Scal(One);
   ScalTwo:  constant Scalar := Scal(Two);
   ScalHalf:  constant Scalar := Scal(Half);
   ScalPi:  constant Scalar := Scal(3.14159265358979323846,3.14159265358979323847);
   Trunc: constant Boolean := IsNumeric(ScalZero);

   package Vec is new Vectors (Scalar => Scalar, Zero_Scalar => ScalZero);
   package Vec_Num is new Vec.Num;
   use Vec, Vec_Num;

   package TV is new Vectors(Scalar => Taylor, Zero_Scalar => ZeroFun(Radius'Last));

   subtype TayVector is TV.Vector;
   subtype TayMatrix is TV.Matrix;

   function DD(N: Integer) return Integer;
   procedure PrepNewt(N: in Integer; R: in Radius; A: in Scalar; Lambda: in out Scalar; S: in  Taylor; Z: in out Taylor; D: out Matrix; V: out Vector);
   --D is derivative of R(S)-S, and V is the vector S-R(S) in the symmetric basis.
   --Everything truncated to order N

end Doubling;
