with Messages,Ada.Text_IO,Reps,Reps.IO;
use Messages;

package body Doubling is

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
--   package Txt_IO is new Ada.Text_IO;
--   use Txt_IO;


   function IMin(I1,I2: Integer) return Integer renames Integer'Min;
   function IMax(I1,I2: Integer) return Integer renames Integer'Max;
   function CompZ(F1,F2,Z: Taylor) return Taylor is
   begin
      return Comp1(Z,F1)+Comp2(Z,F2);
   end CompZ;

   function DCompZ(R: Radius; F1,F2,Z: Taylor) return Taylor is
   begin
      return Comp1(Z,Der1(R,F1))+Comp2(Z,Der2(R,F2));
   end DCompZ;

   procedure CompZero(R: in Radius; F1,F2: in Taylor; Z: in out Taylor) is
      Iter: constant Integer := 50;
      Eps: constant Radius := 1.0E-12;
      H: Taylor;
   begin
      for I in 1 .. Iter loop
         H := Inv(DCompZ(R,F1,F2,Z))*CompZ(F1,F2,Z);
         Sub(H,Z);
         exit when Sup(Norm(R,H))<Eps;
      end loop;
      --- error estimates ...
   end CompZero;


   --procedures for symmetric functions
   function CompZ(S,X,Y,Z: Taylor) return Taylor is
   begin
      return Comp(X,Z,S)+Comp(Y,Z,S);
   end CompZ;

   function DCompZ(R: Radius; S,X,Y,Z: Taylor) return Taylor is
      Der: Taylor;
   begin
      Der:=Der2(R,S);
      return Comp(X,Z,Der)+Comp(Y,Z,Der);
   end DCompZ;


   procedure CompZero(R: in Radius; S,X,Y: in Taylor; Z: in out Taylor) is
      Iter: constant Integer := 50;
      Eps: constant Radius := 1.0E-12;
      H: Taylor;
   begin
      for I in 1 .. Iter loop
         H := Inv(DCompZ(R,S,X,Y,Z))*CompZ(S,X,Y,Z);
         Sub(H,Z);
         exit when Sup(Norm(R,H))<Eps;
      end loop;
      --- error estimates ...
   end CompZero;


   procedure FindPhiLambda(Phi,Eps: in Taylor; Lambda: out Scalar) is
      Zero_Taylor: Taylor;
   begin
      SetZero(Rho(Eps),Zero_Taylor);
      Lambda:=Two*Val(Scal(Zero),Scal(One),Phi);
      if Eps=Zero_Taylor then return; end if;
      for I in 1..200 loop
         Lambda:=Two*Val(Scal(Zero),Scal(One),Phi)-Val(Lambda-Scal(One),Scal(One),Eps)-Val(-Scal(One),Scal(One),Eps);
      end loop;
   end FindPhiLambda;

   procedure FindPhiLambda(Eta: in Scalar; Phi,Eps1,Eps2: in Taylor; Lambda: in out Scalar) is
      use Ada.Text_IO;

      Zero_Taylor: Taylor;
      R: Rep:=Rho(Eps1);
      Lambda0: constant Scalar:=Lambda;
      Lambdaa,Nr: Scalar;
   begin
      SetZero(R,Zero_Taylor);
      Lambda:=(Scal(One)+Sqr(Eta))*Val(Scal(Zero),Scal(One),Phi);
      if (Eps1=Zero_Taylor and Eps2=Zero_Taylor) then return; end if;
      Nr:=Scal(One);
      while Sup(Nr)>Rep(1.0E-18) loop
         Lambdaa:=Lambda;
         Lambda:=(Scal(One)+Sqr(Eta))*Val(Scal(Zero),Scal(One),Phi)-Val(Scal(Zero),Lambda-Lambda0,Eps1)-Sqr(Eta)*Val(Scal(Zero),Scal(Zero),Eps2);
         Nr:=Abs(Lambda-Lambdaa);
         -- Put("Lambda[i+1]-Lambda[i]="); Show(Lambda-Lambdaa);
      end loop;
   end FindPhiLambda;


   procedure PhiRG(Eps: in Taylor;Phi: in out Taylor; Lambda: out  Scalar) is
      G: Taylor;
      R: constant Radius:=Rho(Phi);
   begin
      FindPhiLambda(Phi,Eps,Lambda);
      G:=Comp2(Monom(R,0,1,Lambda),Phi);
      Phi:=(Scal(Two)/Lambda)*(Comp2(G,Phi)-Scal(Half)*Comp(Lambda*Phi-Monom(R,0,0,Scal(One)),G,Eps)-Scal(Half)*Comp(Monom(R,0,1,Lambda)-Monom(R,0,0,Scal(One)),G,Eps))-Monom(Rho(Phi),0,1,Scal(One));
   end PhiRG;



   procedure PhiRG(Eta: in Scalar; Eps1,Eps2: in Taylor;Phi: in out Taylor; Lambda: in out Scalar) is
      G: Taylor;
      R: constant Radius:=Rho(Phi);
      Lambda0: constant Scalar:=Lambda;
   begin
      FindPhiLambda(Eta,Phi,Eps1,Eps2,Lambda);
      G:=Comp2(Monom(R,0,1,Lambda),Phi);
      Phi:=((ScalOne+Sqr(Eta)) /Lambda)*Comp2(G,Phi)
        -Monom(Rho(Phi),0,1,Sqr(Eta))
        -Inv(Lambda)*Comp(G-Monom(R,0,0,Scal(One)),Lambda*Phi-Monom(R,0,0,Lambda0),Eps1)
        -(Sqr(Eta)/Lambda)*Comp(Monom(R,0,1,Lambda),G-Monom(R,0,0,Scal(One)),Eps2);
   end PhiRG;


   procedure PhiDRG(N: in Integer; Eps,Phi: in Taylor; Er,Ei: out Scalar;EvT: out Taylor; M_File: in String) is
         use Ada.Text_IO;

         Phi_Vec: TayVector(1..N);
         Smal: Scalar:=Scal(0.001);
         R: constant Radius:=Rho(Phi);
         Lambdaa: Scalar;
         D: Matrix(1..N,1..N);
         EgR,EgI:  Vector(1..N);
         Ev,Ev1:  Vector(1..N);
         F: File_Type;
   begin
      Create(F,Out_File,M_File);
      for I in 1..N loop
         Phi_Vec(I):=Phi+Monom(R,0,I,Smal);
         PhiRG(Eps,Phi_Vec(I),Lambdaa);
         Phi_Vec(I):=Inv(Smal)*(Phi_Vec(I)-Phi);
         for J in 1..N loop
            D(I,J):=Scal(R)**(I-J)*Component(0,J,Phi_Vec(I));
            Put(F,D(I,J));
            New_Line(F);
         end loop;
      end loop;
      Er:=ScalZero;
      Ei:=ScalZero;
      Eigenvalues(D,EgR,EgI);
      Er:=EgR(1);
      Ei:=EgI(1);
--      New_Line;
--      Put("Re(E)[1]=");Show(EgR(1));
--      Put("Im(E)[1]=");Show(EgI(1));
      for I  in 2..N loop
         if (Sqr(EgR(I))+Sqr(EgI(I)))>(Sqr(Er)+Sqr(Ei)) then
            Er:=EgR(I);
            Ei:=EgI(I);
         end  if;
--         New_Line;
--         Put("Re(E)[");Put("]=");Show(EgR(I));
--         Put("Im(E)[");Put("]=");Show(EgI(I));
      end loop;
      Eigenvector(D,Er,Ev,Scal(1.0E-19));
      SetZero(R,EvT);
      New_Line(F);
      for I in 1..N loop
         Put(F,Ev(I));
         New_Line(F);
         EvT:=EvT+(Scal(R)**(-I))*Monom(R,0,I,Ev(I));
      end loop;
      Close(F);
      New_Line;
      Put("Checkng the eigenvector:");
      New_Line;
      Ev1:=D*Ev-Er*Ev;
      for I in 1..N loop
         Put(Current_Output,Ev1(I));
         New_Line;
      end loop;
   end PhiDRG;



   procedure PhiDRGE(Eta: in Scalar; N: in Integer; Eps1,Eps2: in Taylor; Phi: in out Taylor; Lambda: in out Scalar) is
      use Ada.Text_IO;
      Phi_Vec: TayVector(1..N+1);
      Smal: Scalar:=Scal(0.00001);
      R: constant Radius:=Rho(Phi);
      D: Matrix(1..N+1,1..N+1);
      Lambda0: constant Scalar:=Lambda;
      Lambda1: Scalar:=Lambda;
   begin
      for I in 1..N loop
         Phi_Vec(I):=Phi+Monom(R,0,I,Smal);
         Lambda:=Lambda0;
         PhiRG(Sqrt(Eta),Eps1,Eps2,Phi_Vec(I),Lambda);
         Phi_Vec(I):=Inv(Smal)*(Phi_Vec(I)-Phi);
         for J in 1..N loop
            D(J,I):=Component(0,J,Phi_Vec(I));
            Show(D(J,I));
         end loop;
         D(N+1,I):=ScalZero;
         Show(D(N+1,I));
      end loop;
      Phi_Vec(N+1):=Phi;
      Lambda:=Lambda0;
      PhiRG(Sqrt(Eta+Smal),Eps1,Eps2,Phi_Vec(N+1),Lambda);
      Phi_Vec(N+1):=Inv(Smal)*(Phi_Vec(N+1)-Phi);
      for J in 1..N loop
         D(J,N+1):=Component(0,J,Phi_Vec(N+1));
         Show(D(J,N+1));
      end loop;
      D(N+1,N+1):=(Sqr(Eta+Smal)-Smal)/Smal;
      Show(D(N+1,N+1));
        --      Eigenvalues(De,EgR,EgI);


        --      Decomp(D,Cond,Pvt,Work);
        --      Solve(D,V,Pvt);
        --      Lambda:=Lambda1;
        --      for I in 1..N loop
        --         Phi:=Phi+Monom(R,0,I,V(I));--Monom(R,0,I,Component(0,I,R_Phi));
        --      end loop;
        --      Phi:=Phi-Monom(R,0,0,Component(0,0,Phi)-Scal(One));

   end PhiDRGE;


   procedure PhiNewton(N: in Integer; Eps: in Taylor; Phi: in out Taylor; Lambda: out Scalar) is
      use Ada.Text_IO;

      Pvt: Pivot(1..N);
      R_Phi,R_Phii: Taylor;
      Phi_Vec: TayVector(1..N);
      Smal: Scalar:=Scal(0.01);
      R: constant Radius:=Rho(Phi);
      Cond,Lambdaa: Scalar;
      D: Matrix(1..N,1..N);
      V,Work:  Vector(1..N);
   begin
      R_Phi:=Phi;
      PhiRG(Eps,R_Phi,Lambdaa);
      R_Phii:=R_Phi;
      R_Phi:=R_Phi-Phi;
      for I in 1..N loop
         Phi_Vec(I):=Phi+Monom(R,0,I,Smal);
         PhiRG(Eps,Phi_Vec(I),Lambda);
         Phi_Vec(I):=Inv(Smal)*(Phi_Vec(I)-Phi);
         for J in 1..N loop
            D(J,I):=-Component(0,J,Phi_Vec(I));
         end loop;
         D(I,I):=ScalOne+D(I,I);
         V(I):=Component(0,I,R_Phi);
      end loop;
      Decomp(D,Cond,Pvt,Work);
      Solve(D,V,Pvt);
      Lambda:=Lambdaa;
      --      Phi:=R_Phii;
      for I in 1..N loop
         Phi:=Phi+Monom(R,0,I,V(I));-- -Monom(R,0,I,Component(0,I,R_Phi));
      end loop;
      Phi:=Phi-Monom(R,0,0,Component(0,0,Phi)-Scal(One));
   end PhiNewton;


   procedure PhiNewton(Eta: in Scalar; N: in Integer; Eps1,Eps2: in Taylor; Phi: in out Taylor; Lambda: in out Scalar) is
      use Ada.Text_IO;
      Pvt: Pivot(1..N);
      R_Phi,R_Phii: Taylor;
      Phi_Vec: TayVector(1..N);
      Smal: Scalar:=Scal(0.01);
      R: constant Radius:=Rho(Phi);
      Cond: Scalar;
      D: Matrix(1..N,1..N);
      V,Work:  Vector(1..N);
      Lambda0: constant Scalar:=Lambda;
      Lambda1: Scalar:=Lambda;
   begin
      R_Phi:=Phi;
      PhiRG(Eta,Eps1,Eps2,R_Phi,Lambda1);
      R_Phii:=R_Phi;
      R_Phi:=R_Phi-Phi;
      for I in 1..N loop
         Phi_Vec(I):=Phi+Monom(R,0,I,Smal);
         Lambda:=Lambda0;
         PhiRG(Eta,Eps1,Eps2,Phi_Vec(I),Lambda);
         Phi_Vec(I):=Inv(Smal)*(Phi_Vec(I)-Phi);
         for J in 1..N loop
            D(J,I):=-Component(0,J,Phi_Vec(I));
         end loop;
         D(I,I):=ScalOne+D(I,I);
         V(I):=Component(0,I,R_Phi);
      end loop;
      Decomp(D,Cond,Pvt,Work);
      Solve(D,V,Pvt);
      Lambda:=Lambda1;
      -- Phi:=R_Phii;
      for I in 1..N loop
         Phi:=Phi+Monom(R,0,I,V(I));--Monom(R,0,I,Component(0,I,R_Phi));
      end loop;
      Phi:=Phi-Monom(R,0,0,Component(0,0,Phi)-Scal(One));
   end PhiNewton;





   procedure FindZ(Lambda: in Scalar; Eps,InvPhi: in Taylor; Z: out Taylor) is
      R: constant Radius:=Rho(InvPhi);
      Zz: Taylor;
      Nr: Scalar;
   begin
      Z:=Monom(R,0,0,Scal(One));
      Nr:=Scal(One);
      while Sup(Nr)>Rep(1.0E-18) loop
         Zz:=Z;
         Z:=Comp2(Monom(R,1,0,Scal(One))+Monom(R,0,1,Scal(One))+Inv(Lambda)*Comp(Monom(R,1,0,Lambda)+Monom(R,0,0,Lambda-Scal(One)),Z,Eps)+
                  Inv(Lambda)*Comp(Monom(R,0,1,Lambda)-Monom(R,0,0,Scal(One)),Z,Eps),InvPhi);
         Nr:=Norm(R,Zz-Z);
      end loop;
      Z:=Z+Monom(R,0,0,Scal(One)-Component(0,0,Z));
   end FindZ;


   procedure FindZ(Eta,Lambda0,Lambda: in Scalar; Eps1,Eps2,InvPhi: in Taylor; Z: out Taylor) is
      use Ada.Text_IO;
      R: constant Radius:=Rho(InvPhi);
      Zz,Zero_Taylor: Taylor;
      Nr: Scalar:=Scal(One);
      C: Scalar;
   begin
      C:=( Lambda/(Scal(One)+Sqr(Eta))+(Sqr(Eta)*Val(Scal(Zero),Scal(Zero),Eps2)+Val(Scal(Zero),Lambda-Lambda0,Eps1))/(Scal(One)+Sqr(Eta)));
      Z:=Monom(R,0,0,Scal(One));
      SetZero(R,Zero_Taylor);
      Z:=Comp2(Monom(R,1,0,Sqr(Eta))+Monom(R,0,1,Scal(One)),InvPhi);
      Z:=Z+Monom(R,0,0,Scal(One)-Component(0,0,Z));
      if (Eps1=Zero_Taylor and Eps2=Zero_Taylor) then return; end if;
      while Sup(Nr)>Rep(1.0E-18) loop
         Zz:=Z;
         Z:=Comp2(Inv(C)*(( Lambda/(Scal(One)+Sqr(Eta)))*(Monom(R,1,0,Sqr(Eta))+Monom(R,0,1,Scal(One))+Monom(R,0,0,Scal(One)))
                           +(Sqr(Eta)/(Scal(One)+Sqr(Eta)))*Comp(Monom(R,1,0,Lambda),Z-Monom(R,0,0,Scal(One)),Eps2)
                           + Inv(Scal(One)+Sqr(Eta))*Comp(Z-Monom(R,0,0,Scal(One)),Monom(R,0,1,Lambda) -Monom(R,0,0,Lambda0-Lambda),Eps1))-Monom(R,0,0,Scal(One)),InvPhi);
         --         Put("Convergence in FindZ: ");Show(Norm(R,Zz-Z),True);
         Nr:=Norm(R,Zz-Z);
      end loop;
      Z:=Z+Monom(R,0,0,Scal(One)-Component(0,0,Z));
   end FindZ;



   function PhiRG(Eps, Phi: Taylor) return Taylor is
      RGPhi: Taylor;
      Lambda: Scalar;
   begin
      RGPhi:=Phi;
      PhiRG(Eps,RGPhi,Lambda);
      return RGPhi;
   end PhiRG;


   procedure FindLambda(R: in Radius; S: in Taylor; A: in Scalar; Lambda: in out Scalar) is
      Iter: constant Integer := 50;
      Eps: constant Radius := 1.0E-12;
      D,C: Scalar;
   begin
      for I in 1 .. Iter loop
         D:=Val(Lambda-A,ScalOne-A,Der1(R,S));
         C:=(Val(Lambda-A,ScalOne-A,S)+Val(ScalZero-A,ScalOne-A,S))/D;
         Lambda:=Lambda-C;
         exit when SupAbs(C)<Eps or Iter>100;
      end loop;
      --- error estimates ...
   end FindLambda;



   procedure RenG(R: in Radius; A: in Scalar; Lambda,Mu: in out Scalar; S,Z: in out Taylor  ) is
      X,Y,H,RS: Taylor;
   begin

      FindLambda(R,S,A,Lambda);

      SetZero(R,X);
      SetZero(R,Y);
      SetZero(R,H);
      X:=X+Monom(R,0,0,A*Lambda-A)+Monom(R,1,0,Lambda);
      Y:=Y+Monom(R,0,0,A*Lambda-A)+Monom(R,0,1,Lambda);
      H:=H+Monom(R,0,0,A*Lambda-A)+Monom(R,0,1,Lambda);

      CompZero(R,S,X,Y,Z);
      RS:=Comp(Z,H,S);

      --      Mu:=Component(0,0,RS);
      Mu:=Scal(5.0)*Val(Scal(One),Scal(Zero),Der2(R,RS));
      S:=Inv(Mu)*RS;
   end RenG;


   procedure RG(R: in Radius; A: in Scalar; Lambda,Mu: in out Scalar; S,Z: in out Taylor  ) is
      X,Y,H,RS: Taylor;
   begin

      FindLambda(R,S,A,Lambda);

      SetZero(R,X);
      SetZero(R,Y);
      SetZero(R,H);
      X:=X+Monom(R,0,0,A*Lambda-A)+Monom(R,1,0,Lambda);
      Y:=Y+Monom(R,0,0,A*Lambda-A)+Monom(R,0,1,Lambda);
      H:=H+Monom(R,0,0,A*Lambda-A)+Monom(R,0,1,Lambda);

      CompZero(R,S,X,Y,Z);
      RS:=Comp(Z,H,S);

      Mu:=Component(0,0,RS);
      S:=Inv(Mu)*RS;
   end RG;


   procedure DLambdah(R: in Radius; A,Lambda: in Scalar; S,H: in Taylor; DL: out Scalar)  is
   begin
      DL:=-(Val(Lambda-A,ScalOne-A,H)+ Val(-A,ScalOne-A,H))/Val(Lambda-A,ScalOne-A,Der1(R,S));
   end DLambdah;


   function DLambdah(R: Radius; A,Lambda: Scalar; S,H: Taylor) return Scalar is
   begin
      return -(Val(Lambda-A,ScalOne-A,H)+ Val(-A,ScalOne-A,H))/Val(Lambda-A,ScalOne-A,Der1(R,S));
   end DLambdah;


   procedure DMuh(R: in Radius; A,Lambda,DL: in Scalar; S,H,Z,DZ: in  Taylor; DM: out Scalar) is
      S1,S2,Y: Taylor;
   begin

      S1:=Der1(R,S);
      S2:=Der2(R,S);

      SetZero(R,Y);
      Y:=Y+Monom(R,0,0,A*Lambda-A)+Monom(R,0,1,Lambda);

      DM:=Val(ScalZero,ScalZero,Comp(Z,Y,S1)*DZ)+A*DL*Val(ScalZero,ScalZero,Comp(Z,Y,S2))+Val(ScalZero,ScalZero,Comp(Z,Y,H));
   end DMuh;

   function DMuh(R: Radius; A,Lambda,DL: Scalar; S,H,Z,DZ: Taylor) return Scalar is
      S1,S2,Y: Taylor;
   begin
      S1:=Der1(R,S);
      S2:=Der2(R,S);
      SetZero(R,Y);
      Y:=Y+Monom(R,0,0,A*Lambda-A)+Monom(R,0,1,Lambda);

      return Val(ScalZero,ScalZero,Comp(Z,Y,S1)*DZ)+A*DL*Val(ScalZero,ScalZero,Comp(Z,Y,S2))+Val(ScalZero,ScalZero,Comp(Z,Y,H));
   end DMuh;


   procedure DZh(R: in Radius; A,Lambda,DL: in Scalar; S,Z: in Taylor; H: in out Taylor) is
      X,Y,X1,Y1,S1,S2: Taylor;
   begin

      SetZero(R,X1);
      SetZero(R,Y1);
      X1:=X1+Monom(R,0,0,A)+Monom(R,1,0,ScalOne);
      Y1:=Y1+Monom(R,0,0,A)+Monom(R,0,1,ScalOne);

      SetZero(R,X);
      SetZero(R,Y);
      X:=X+Monom(R,0,0,A*(Lambda-ScalOne))+Monom(R,1,0,Lambda);
      Y:=Y+Monom(R,0,0,A*Lambda-A)+Monom(R,0,1,Lambda);

      S1:=Der1(R,S);
      S2:=Der2(R,S);
      H:=-( Comp(X,Z,H)+Comp(Y,Z,H)+DL*( Comp(X,Z,S1)*X1 +Comp(Y,Z,S1)*Y1 ) ) / ( Comp(X,Z,S2) + Comp(Y,Z,S2) );
   end DZh;


   procedure DRGh(R: in Radius; A: in Scalar; Lambda: in out Scalar; S: in Taylor; Z,H: in out Taylor) is
      RS,DZ,S1,S2,E,P: Taylor;
      DL,DM,Mu: Scalar;
   begin
      Mu:=ScalZero;
      RS:=S;
      RG(R,A,Lambda,Mu,RS,Z);
      DLambdah(R,A,Lambda,S,H,DL);
      DZ:=H;
      DZh(R,A,Lambda,DL,S,Z,DZ);
      DMuh(R,A,Lambda,DL,S,H,Z,DZ,DM);

      ----------
      SetZero(R,E);
      E:=E+Monom(R,0,0,A*(Lambda-ScalOne))+Monom(R,0,1,Lambda);
      SetZero(R,P);
      P:=P+Monom(R,0,0,A)+Monom(R,0,1,ScalOne);
      S1:=Der1(R,S);
      S2:=Der2(R,S);

      H:=Inv(Mu)*( (-DM)*RS+DZ*Comp(Z,E,S1)+DL*P*Comp(Z,E,S2)+Comp(Z,E,H) );
   end DRGh;


  function DD(N: Integer) return Integer is
  begin
    if N>2 then return (N-N/2)*(N/2+1)+N; else return 2*N; end if;
  end DD;

  procedure PrepNewt(N: in Integer; R: in Radius; A: in Scalar; Lambda: in out Scalar; S: in Taylor; Z: in out Taylor; D: out Matrix; V:   out Vector) is
     use Ada.Text_IO, Reps.IO;
     Dim: constant Integer := DD(N);
     K: Integer;
     Mu: Scalar:=ScalZero;
     F: Taylor;
     Nrm: Vector(1..Dim);
     H,Clmn: TayVector(1..Dim);
  begin

     --Normalized basis vectors (H) and matrix columns (Clmn)
     --Right hand side vector of the linear algebra in Newton (V)
     F:=S;
     RG(R,A,Lambda,Mu,F,Z);
     F:=F-S;
     K:=0;
     for I in -1..(N-1) loop
        for J in IMax(0,I)..(N-1-I) loop
           if K>0 then
              if I=-1 then
                 H(K):=Monom(R,I+1,J,ScalOne); --vector (xi+eta)
              else
                 H(K):=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne); --vector (xi+eta)
              end if;
              Nrm(K):=Norm(R,H(K));
              H(K):=Inv(Nrm(K))*H(K); --normalized vector (xi+eta)
              Clmn(K):=H(K);
              DRGh(R,A,Lambda,S,Z,Clmn(K));
              Clmn(K):=Clmn(K)-H(K);
              V(K):=(-Nrm(K))*Component(I+1,J,F);
              F:=F+V(K)*H(K);
           end if;
           K:=K+1;
        end loop;
     end loop;

     --Components of columns along the basis vectors
     for M in Clmn'Range loop
        K:=0;
        for I in -1..(N-1) loop
           for J in IMax(0,I)..(N-1-I) loop
              if K>0 then
                 D(K,M):=Nrm(K)*Component(I+1,J,Clmn(M));
                 Clmn(M):=Clmn(M)-D(K,M)*H(K);
              end if;
              K:=K+1;
           end loop;
        end loop;
     end loop;
  end PrepNewt;


  procedure FindS(Eps: in Radius; N: in Integer; R: in Radius; A: in Scalar; Lambda: in out Scalar; S,Z: in out Taylor  ) is
     Dim: constant Integer := DD(N);
     Iter: constant Integer := 5000;
     K: Integer;
     Cond: Scalar;
     DS: Taylor;
     Pvt: Pivot(1..Dim);
     V,Work: Vector(1..Dim);
     D: Matrix(1..Dim,1..Dim);
     H: TayVector(1..Dim); --2N is the number of symmetric basis vectors of order <=N
  begin
     -- New_Line;
     -- Put("Newton in ");Put(Dim,2,10);Put(" variables");

     --Normalized basis vectors
     K:=0;
     for I in -1..(N-1) loop
        for J in IMax(0,I)..(N-1-I) loop
           if K>0 then
              if I=-1 then
                 H(K):=Monom(R,I+1,J,ScalOne);
              else
                 H(K):=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne);
              end if;
              H(K):=Inv(Norm(R,H(K)))*H(K);
           end if;
           K:=K+1;
        end loop;
     end loop;

     --Newton for F(DS):=P[R(S+DS)-(S+DS)]=0. Numerical for now
     for I in 1..Iter loop
        PrepNewt(N,R,A,Lambda,S,Z,D,V);
        --solve D*X=V
        Decomp(D,Cond,Pvt,Work);
        Solve(D,V,Pvt);
        SetZero(R,DS);
        for J in H'Range loop
           DS:=DS+V(J)*H(J);
        end loop;
        S:=S+DS;
        exit when Sup(Norm(R,DS))<Eps;
        -- Show("Convergence criterion = ",Sup(Norm(R,DS)));
     end loop;
  end FindS;


  procedure DRGh(N: in Integer; R: in Radius; A: in Scalar; Lambda: in out Scalar; S: in  Taylor; Z,H: in out Taylor) is
     Dim: constant Integer := DD(N);
     D: Matrix(1..Dim,1..Dim);
     K: Integer;
     DRh,DDeltah: Taylor;
     Cond: Scalar;
     Pvt: Pivot(1..Dim);
     Nrm,V,Work: Vector(1..Dim);
     X,Clmn: TayVector(1..Dim);
  begin

     DRh:=H;
     DRGh(R,A,Lambda,S,Z,DRh);
     DRh:=H-DRh;
     --Normalized basis vectors (X) and matrix columns (Clmn)
     --Right hand side vector of the linear algebra in Newton (V)
     K:=0;
     for I in -1..(N-1) loop
        for J in IMax(0,I)..(N-1-I) loop
           if K>0 then
              if I=-1 then
                 X(K):=Monom(R,I+1,J,ScalOne); --vector (xi+eta)
              else
                 X(K):=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne); --vector (xi+eta)
              end if;
              Nrm(K):=Norm(R,X(K));
              X(K):=Inv(Nrm(K))*X(K); --normalized vector (xi+eta)
              Clmn(K):=X(K);
              DRGh(R,A,Lambda,S,Z,Clmn(K));
              Clmn(K):=Clmn(K)-X(K);
              V(K):=Nrm(K)*Component(I+1,J,DRh);
           end if;
           K:=K+1;
        end loop;
     end loop;

     --Components of columns along the basis vectors
     for M in Clmn'Range loop
        K:=0;
        for I in -1..(N-1) loop
           for J in IMax(0,I)..(N-1-I) loop
              if K>0 then
                 D(K,M):=Nrm(K)*Component(I+1,J,Clmn(M));
                 Clmn(M):=Clmn(M)-D(K,M)*X(K);
              end if;
              K:=K+1;
           end loop;
        end loop;
     end loop;

     Decomp(D,Cond,Pvt,Work);
     Solve(D,V,Pvt);
     SetZero(R,DDeltah);

     for J in X'Range loop
        DDeltah:=DDeltah+V(J)*X(J);
     end loop;

     H:=H+DDeltah;
     DRGh(R,A,Lambda,S,Z,H);
  end DRGh;


  procedure Integrate(R: in Radius; S: in out Taylor) is
     --finds the generating function H, such that H_2(x,y)=S(x,y),H_1(x,y)=S(y,x)
     H: Taylor;
  begin
     SetZero(R,H);
     for I in 1..PDeg loop
        H:=H+Monom(R,0,I,Component(0,I-1,S)/Scal(I))+Monom(R,I,0,Component(0,I-1,S)/Scal(I));
        for J in 1..Pdeg loop
           H:=H+Monom(R,I,J,Component(I,J-1,S)/Scal(J));
        end loop;
     end loop;
     S:=H;
     --error estimates...
  end Integrate;


  procedure Z0_Range(RhoX: in Radius; P: in Scalar; A,B,C: in Taylor; RhoZ: out Radius) is
     use Ada.Text_IO, Reps.IO;
     --Z_0=(B(X)+B(Y))/2(A(X)+A(y))*[sqrt[1+4(A(X)+A(Y))(C(X)+C(Y))/(B(X)+B(Y))^2]-1]
     Ax,Ay,Bx,By,Cx,Cy,X,Y,F,F1,B1,A1: Taylor;
     F0,B0,A0: Scalar;
     R: Radius;
  begin
     X:=Monom(RhoX,0,0,P)+Monom(RhoX,1,0,Scal(One));
     Y:=Monom(RhoX,0,0,P)+Monom(RhoX,0,1,Scal(One));
     Ax:=Comp1(X,A);
     Ay:=Comp1(Y,A);
     Bx:=Comp1(X,B);
     By:=Comp1(Y,B);
     Cx:=Comp1(X,C);
     Cy:=Comp1(Y,C);
     F:=Scal(4.0)*(Ax+Ay)*(Cx+Cy);

     F0:=Component(0,0,F);
     F1:=F-Monom(RhoX,0,0,F0);
     R:=Sup(Norm(RhoX,RhoX,F1));
     F0:=F0+Scal(-R,R);

     B0:=Component(0,0,Bx+By);
     B1:=Bx+By-Monom(RhoX,0,0,B0);
     R:=Sup(Norm(RhoX,RhoX,B1));
     B0:=B0+Scal(-R,R);


     A0:=Component(0,0,Ax+Ay);
     A1:=Ax+Ay-Monom(RhoX,0,0,A0);
     R:=Sup(Norm(RhoX,RhoX,A1));
     A0:=A0+Scal(-R,R);

     RhoZ:=Sup( abs( Half*B0* ( Sqrt( Scal(One)-F0/Sqr(B0) )-Scal(One) ) / A0 ));
  end Z0_Range;


   procedure Sigma0_Norms(Rho,RhoX,RhoZ: in Radius; P: in Scalar; S: in Taylor; Const1,Const2,S_Nrm1,S_Nrm2,S_Der1,S_Der2: out Scalar) is
     use Ada.Text_IO;
     Sx,X,Sx_Der: Taylor;
   begin
      X:=Monom(RhoX,0,0,P)+Monom(RhoX,1,0,Scal(One));
      Sx:=Comp1(X,S);

      --S_Nrm1 is the norm of sigma_0 -const terms in the domain (RhoX,Rho)
      Const1:=Component(0,0,Sx);
      S_Nrm1:=Norm(RhoX,Rho,Sx-Monom(Rho,0,0,Const1));

      --S_Nrm2 is the norm of sigma_0 -const terms in the domain (RhoX,RhoZ)
      S_Nrm2:=Norm(RhoX,RhoZ,Sx-Monom(Rho,0,0,Const1));

      Sx_Der:=Der2(Rho,Sx);
      Const2:=Component(0,0,Sx_Der);
      --S_Der1 is the norm of the D_2 sigma_0 -const terms in the domain in the domain (RhoX,Rho)
      S_Der1:=Norm(RhoX,Rho,Sx_Der-Monom(Rho,0,0,Const2) );

      --S_Der2 is the norm of the D_2 sigma_0 -const terms in the domain (RhoX,RhoZ)
      S_Der2:=Norm(RhoX,RhoZ,Sx_Der-Monom(Rho,0,0,Const2) );

   end Sigma0_Norms;


   procedure N_h(Rho,RhoX,H_Delt,Z_Delt: in Radius; A_Tay,B_Tay,C_Tay: in Taylor; P,A,B: in Scalar;M,N_Der,N_Diff: out Scalar) is
      --Right now the
      use Ada.Text_IO, Reps.IO;
      RhoZ: Radius;
      S_Rho: Scalar:=Scal(Rho);
      S_RhoX: Scalar:=Scal(RhoX);
      S_H_Delt:  Scalar:=Scal(H_Delt);
      S_Z_Delt:  Scalar:=Scal(Z_Delt);
      AbsA: Scalar:=abs(A);
      Max_RhoZ,N,Maxx,D,Fa,Const,Const1,Const2,ThetaRho,ThetaS,DThetaRho,DThetaS,Eps,I1,I2,I3: Scalar;
      S: Taylor;
   begin
      Z0_Range(RhoX,P,A_Tay,B_Tay,C_Tay,RhoZ);
      SetZero(Rho,S);
      S:=S+A_Tay*Monom(Rho,0,2,Scal(One))+B_Tay*Monom(Rho,0,1,Scal(One))+C_Tay+
            A*Monom(Rho,0,2,Scal(One))+B*Monom(Rho,0,1,Scal(One));

      --Norm of I-DC_0(Z_0)
      Sigma0_Norms(Rho,RhoX,RhoZ,P,S,Const1,Const2,ThetaRho,ThetaS,DThetaRho,DThetaS); --NrmRho and NrmS are function norms on rho and s, DNrmRho and DNrmS are similar derivative norms
      Const:=Sqr(B)+Scal(4.0)*A*Const1;
      M:=Inv(Scal(One)-(Const2+DThetaS)/Sqrt(Const-Two*AbsA*ThetaS));
      Max_RhoZ:=Scal(RhoZ)+Z_Delt*abs(M);
      if Sup(Max_RhoZ)>Rho then
         New_Line;
         Message("Doubling.N_h",Domain_Error);
      end if;

      --Norm C_h(Z_0)-Z_0
      Fa:=Scal(Rep'Max(Sup(S_RhoX+abs(P)/S_Rho),Sup(RhoZ*Inv(S_Rho))));
      N_Diff:=Two*Fa*S_H_Delt*Sqrt(Two*Const)/(Const-Scal(4.0)*AbsA*(ThetaS+Fa*S_H_Delt));

      Sigma0_Norms(Rho,RhoX,Sup(Max_RhoZ),P,S,Const1,Const2,ThetaRho,ThetaS,DThetaRho,DThetaS);

      --Norm DC_h(Z_0+M*z)-DC_0(Z_0)=I1+I2+I3
      Fa:=Scal( Rep'Max( Sup( S_RhoX+abs(P)/Rho) , Sup( Max_RhoZ/Rho ) ) );
      N:=ScalOne/Log(S_Rho/Max_RhoZ);

      Maxx:=N*Exp((N-Scal(One))*Log(Max_RhoZ))/Exp(N*Log(S_Rho));

      I1:=Maxx*S_H_Delt/Sqrt(Const-Scal(4.0)*AbsA*(ThetaS+Fa*S_H_Delt));

      D:=( Const-Scal(4.0)*AbsA*(ThetaS+Fa*S_H_Delt) ) / Scal(3.0);
      if Inf(D)<=Zero then
         New_Line;
         Message("Doubling.N_h",Division_By_Zero);
      end if;
      I2:=Scal(4.0)*AbsA*S_H_Delt*(Const2+DThetaS)/( Two*D*Sqrt(D) );

      Eps:=M*S_Z_Delt/(S_Rho-Max_RhoZ);
      I3:=Eps*(Const2+DThetaRho)/Sqrt(Const-Scal(4.0)*AbsA*ThetaRho);

      N_Der:=(I1+I2+I3)*M;
   end N_h;


   procedure Compactness(Rho,RhoX,H_Delt,Zn_Delt: in Radius; A_Tay,B_Tay,C_Tay: in Taylor; P,A,B: in Scalar; Z_Range: out Scalar) is
      use Ada.Text_IO, Reps.IO;
      RhoZ: Radius;
      S_Rho: Scalar:=Scal(Rho);
      S_RhoX: Scalar:=Scal(RhoX);
      S_H_Delt:  Scalar:=Scal(H_Delt);
      Zn_Delta: Rep:=Zn_Delt;
      AbsA: Scalar:=abs(A);
      Theta,Del_Zn,C1,C2,M,N,Maxx,Max_RhoZ,D,Fa,Const,Const1,Const2,ThetaRho,ThetaS,DThetaRho,DThetaS,Eps,I1,I2,I3: Scalar;
      S: Taylor;
   begin

      <<AGAIN>> Z0_Range(RhoX,P,A_Tay,B_Tay,C_Tay,RhoZ);
      SetZero(Rho,S);
      S:=S+A_Tay*Monom(Rho,0,2,Scal(One))+B_Tay*Monom(Rho,0,1,Scal(One))+C_Tay+
        A*Monom(Rho,0,2,Scal(One))+B*Monom(Rho,0,1,Scal(One));

      --Norm of I-DC_0(Z_0)
      Sigma0_Norms(Rho,RhoX,RhoZ,P,S,Const1,Const2,ThetaRho,ThetaS,DThetaRho,DThetaS);
      Const:=Sqr(B)+Scal(4.0)*A*Const1;
      M:=Inv(Scal(One)-(Const2+DThetaS)/Sqrt(Const-Two*AbsA*ThetaS));
      Max_RhoZ:=Scal(RhoZ)+Zn_Delta*abs(M);

      --Max_RhoZ an apriori bouns s' in the notes
      if Sup(Max_RhoZ)>Rho then
         New_Line;
         Message("Doublng.Compactness",Domain_Error);
      end if;

      --Norm C_h(Z_0)-Z_0
      Fa:=Scal(Rep'Max(Sup((S_RhoX+abs(P))/S_Rho),Sup(RhoZ*Inv(S_Rho))));
      Del_Zn:=Two*Fa*S_H_Delt*Sqrt(Two*Const)/(Const-Scal(4.0)*AbsA*(ThetaS+Fa*S_H_Delt));

      Sigma0_Norms(Rho,RhoX,Sup(Max_RhoZ),P,S,Const1,Const2,ThetaRho,ThetaS,DThetaRho,DThetaS);

      --Norm DC_h(Z_0+M*z)-DC_0(Z_0)=I1+I2+I3
      Fa:=Scal( Rep'Max( Sup( (S_RhoX+abs(P))/Rho) , Sup( Max_RhoZ*Inv(S_Rho) ) ) );
      N:=ScalOne/Log(S_Rho/Max_RhoZ);

      Maxx:=N*Exp((N-Scal(One))*Log(Max_RhoZ))/Exp(N*Log(S_Rho));

      I1:=Maxx*S_H_Delt/Sqrt(Const-Scal(4.0)*AbsA*(ThetaS+Fa*S_H_Delt));

      D:=( Const-Scal(4.0)*AbsA*(ThetaS+Fa*S_H_Delt) ) / Scal(3.0);
      if Inf(D)<=Zero then
         New_Line;
         Message("Doubling.Compactness",Division_By_Zero);
      end if;
      I2:=Scal(4.0)*AbsA*S_H_Delt*(Const2+DThetaS)/( Two*D*Sqrt(D) );

      Eps:=Zn_Delta*M/(S_Rho-Max_RhoZ);
      I3:=Eps*(Const2+DThetaRho)/Sqrt(Const-Scal(4.0)*AbsA*ThetaRho);

      C2:=(I1+I2+I3)*M;

      --Norm C_h(Z_0+M*z+M*delta z)-C_h(Z_0+M*z)-DC_h(Z_0+M*z)M=F1+F2
      --Serves as a factor in front of [delta z]^2
      Eps:=Sqr(M)/(S_Rho-Max_RhoZ);
      C1:=Half*Eps*Sqrt(Const+Scal(4.0)*abs(A)*(ThetaS+Fa*S_H_Delt))/abs(A);

      Z_Range:=Scal(Sup(Del_Zn));
      for I in 1..100 loop
         Del_Zn:=Scal(Sup(C1*Sqr(Del_Zn)/(S_Rho-Max_RhoZ+M*Del_Zn)+C2*Del_Zn));
         Z_Range:=Z_Range+Del_Zn;
      end loop;
      Theta:=(C1*Del_Zn/(S_Rho-Max_RhoZ)+C2);
      if Sup(Theta)>One then
         New_Line;
         Message("Theta in Doublng.Compactness > 1",Domain_Error);
      end if;

      Z_Range:=Scal(RhoZ)+M*(Z_Range+ Del_Zn/(Scal(One)-Theta));

      --Max_RhoZ has to be larger than |Z_S|, then increase the apriori bound Zn_Delta and restart
      if Inf(Max_RhoZ)<Sup(Z_Range) then
         Zn_Delta:=Rep(1.01)*Zn_Delta;
         goto AGAIN;
      end if;
   end Compactness;



   procedure Lambda(Rho,H_Delt: in Radius;A_Tay,B_Tay,C_Tay: in Taylor; K: in Scalar; Lambda:out Scalar) is
      use Ada.Text_IO, Reps.IO;
      A,B,C,DLambda: Scalar;
      S: Scalar:=ScalOne-K;
      R1,R2: Rep;
   begin
      A:=Val(Scal(Zero),S,A_Tay);
      B:=Val(Scal(Zero),S,B_Tay);
      C:=Val(Scal(Zero),S,C_Tay);
      C:=Sqr(K)*A-K*B+Two*C;

      Lambda:=( Sqrt(Scal(One)-Scal(4.0)*A*C/Sqr(B))-Scal(One) )*B/(Two*A);
      B:=Two*A*Lambda+B;
      R1:=Rep'Max( Sup(abs(K)),Sup(abs(ScalOne-K)) );
      R2:=Rep'Max( Sup(Scal(1.1)*abs(Lambda)),Sup(abs(ScalOne-K)) );--this 1.1 has to fixed later - too rigid
      C:=R1*Scal(-H_Delt,H_Delt)/Rho+R2*Scal(-H_Delt,H_Delt)/Rho;
      DLambda:=( Sqrt(Scal(One)-Scal(4.0)*A*C/Sqr(B))-Scal(One) )*B/(Two*A);
      Lambda:=K+Lambda+Scal(-Sup(DLambda),Sup(DLambda));
      if Sup(abs(Lambda-K))>R2 then
         Message("Doubling.Lambda",Domain_Error);
      end if;
   end Lambda;


   procedure Mu(Rho,H_Delt,Z_Delt: in Radius; A,B,C,S: Taylor; K,Lambda: in Scalar; Mu: out Scalar) is
      use Ada.Text_IO, Reps.IO;
      X: Scalar:=K*(Lambda-ScalOne);
      AA,BB,CC,Z0: Scalar;
      R: Rep;
   begin
      AA:=Val(X,Scal(Zero),A);
      BB:=Val(X,Scal(Zero),B);
      CC:=Val(X,Scal(Zero),C);
      Z0:=Half*BB* ( Sqrt( Scal(One)-Scal(4.0)*AA*CC/Sqr(BB) )-Scal(One) ) / AA ;
      R:=Rep'Max( Sup(abs(Z0+Scal(-Z_Delt,Z_Delt))),Sup(K*(Lambda-ScalOne)));
      Mu:=Val(Z0+Scal(-Z_Delt,Z_Delt),K*(Lambda-Scal(One)),S)+R*Scal(-H_Delt,H_Delt)/Rho;
   end Mu;




   procedure Newton(R1: in Radius; P: in out Scalar; Tay: in Taylor; E: in Scalar) is --FINE
                                                                                      --solves Tay(Z)=E

      use Ada.Text_IO;
      Eps: Rep;
      DeV,C1,C2,Epsilon0,M,M0,C: Scalar;
      Err: Rep:=Rep(1.0E-16);
      R: Radius;
      I: Integer:=1;
      Tayy,Tayy1,Tayy2,Tay1,Tay2,Tay11,Tay12,Tay22: Taylor;
   begin
      R:=Rho(Tay);
      --Copy(Tay,Tayy);
      SetZero(R,Tayy);
      --      ResetErrComp(Tayy);
      for I in 0..PDeg loop
         for J in 0..PDeg loop
            Tayy:=Tayy+Monom(R,I,J,Component(I,J,Tay));
         end loop;
      end loop;

      Tayy1:=Der1(R1,Tayy);
      Tayy2:=Der2(R1,Tayy);

     -- Put("HERE1");

      <<AGAIN>> Eps:=Rep(1.0E+100);
      --Approximation
      while Eps > Err loop
         C:=Center(P);
       --  Put("HERE2");
         --R:=SupRad(C1,Tayy);
         C2:=Center(Val(C,C,Tayy1)+Val(C,C,Tayy2));
         C1:=Center(Val(C,C,Tayy));
         C1:=(Center(E)-C1)/C2;
         Eps:=SupAbs(Center(C1));
      --   New_Line;
      --   Show(Scal(Eps));
         P:=Center(P+C1);
      end loop;
      if Trunc then return; end if;
      Tay1:=Der1(R1,Tay);
      Tay2:=Der2(R1,Tay);
      Tay11:=DerDer11(R1,Tay);
      Tay22:=DerDer22(R1,Tay);
      Tay12:=DerDer12(R1,Tay);
      --Bound on Epsilon0
      --     R:=SupRad(P,Tay);
      DeV:=Val(P,P,Tay1)+Val(P,P,Tay2);
      Epsilon0:=(E-Val(P,P,Tay))/DeV;
      M0:=Epsilon0/DeV;
      --Bound on the solution
      P:=P+Epsilon0;
      P:=P+Ball(Epsilon0);
      --Kantorovich inequality
      --R:=SupRad(P,Tay);
      M:=Val(P,P,Tay11)+Val(P,P,Tay22)+ScalTwo*Val(P,P,Tay12);
      if SupAbs(M0*M)>Half then
         New_Line;
         Put("Kantorovich criterion=");Show(Abs(M0*M));
         Err:=Rep(0.1)*Err;
         if I>100 then
            Message("Taylors.Newton",Giving_Up);
         end if;
         I:=I+1;
         New_Line;Put("Continuing the Newton...");
         goto AGAIN;
      end if;
   end Newton;



   procedure Newton(Bs,Cs,Es,P: in out Scalar; Tay: in Taylor; E: in Scalar) is --FINE
                                                                                      --solves Tay(Z)=E

      use Ada.Text_IO;
      Eps: Rep;
      Der,C,C1,C2,Epsilon,Beta,Fact: Scalar;
      Err: Rep:=Rep(1.0E-20);
      R: Radius;
      I: Integer:=1;
   begin
      R:=Rho(Tay);

      <<AGAIN>> Eps:=Rep(1.0E+100);
      --Approximation
      while Eps > Err loop
         C:=Center(P);
         C2:=Center(DerVal1(Bs,Cs,Es,C,C,Tay)+DerVal2(Bs,Cs,Es,C,C,Tay));
         C1:=Center(Val(Bs,Cs,Es,C,C,Tay));
         C1:=(Center(E)-C1)/C2;
         Eps:=SupAbs(Center(C1));
         P:=Center(P+C1);
      end loop;
      C:=P;
      if Trunc then return; end if;
      Der:=(DerVal1(Bs,Cs,Es,P,P,Tay)+DerVal2(Bs,Cs,Es,P,P,Tay));
      Fact:=Center(Scal(One)/abs(Der));
      Der:=Scal(Sup(abs(Scal(One)-Fact*(DerVal1(Bs,Cs,Es,P,P,Tay)+DerVal2(Bs,Cs,Es,P,P,Tay)))));
      Epsilon:=abs((E-Fact*Val(Bs,Cs,Es,P,P,Tay)));
      Beta:=Scal(Sup(Epsilon/(Scal(One)-Abs(Der))));
      P:=C+Ball(Beta);
      Der:=Scal(Sup(abs(Scal(One)-Fact*(DerVal1(Bs,Cs,Es,P,P,Tay)+DerVal2(Bs,Cs,Es,P,P,Tay)))));
       while Epsilon>(Scal(One)-Abs(Der))*Beta loop
         Beta:=Rep(1.001)*Beta;
         P:=C+Ball(Beta);
          Der:=Scal(Sup(abs(Scal(One)-Fact*(DerVal1(Bs,Cs,Es,P,P,Tay)+DerVal2(Bs,Cs,Es,P,P,Tay)))));
      end loop;
   end Newton;



   procedure NewtonMap(N: in Integer; Lambda,Mu: in Scalar; X,U: in out Scalar; Phi,S: in Taylor) is --FINE
                                                                                      --solves Tay(Z)=E

      use Ada.Text_IO;
      Err: Rep;
      Eps: Rep:=Rep(1.0E-15);
      Xn,Un,Xn1,Un1,XDel,UDel,Det: Scalar;
      Delt: Scalar:=Scal(Rep(0.000001));
      IDF,DF: Matrix(1..2,1..2);
   begin

      Xn:=X;
      Un:=U;
      Err:=Rep(1.0E+20);
      while Err>Eps loop
         Xn1:=Xn;
         Un1:=Un;
         --Value
         for I in 1..N loop
            Map(Phi,S,Lambda,Mu,Xn,Un);
         end loop;
         --Derivative
         X:=Xn1+Delt;U:=Un1;
         for I in 1..N loop
            Map(Phi,S,Lambda,Mu,X,U);
         end loop;
         DF(1,1):=Scal(One)-(X-Xn1)/Delt;
         DF(2,1):=-(U-Un1)/Delt;
         X:=Xn1;U:=Un1+Delt;
         for I in 1..N loop
            Map(Phi,S,Lambda,Mu,X,U);
         end loop;
         DF(1,2):=-(X-Xn1)/Delt;
         DF(2,2):=Scal(One)-(U-Un1)/Delt;

         Det:=DF(1,1)*DF(2,2)-DF(1,2)*DF(2,1);
         IDF(1,1):=DF(2,2)/Det;IDF(1,2):=-DF(1,2)/Det;
         IDF(2,1):=-DF(2,1)/Det;IDF(2,2):=DF(1,1)/Det;


         XDel:=IDF(1,1)*(Xn-Xn1)+IDF(1,2)*(Un-Un1);
         UDel:=IDF(2,1)*(Xn-Xn1)+IDF(2,2)*(Un-Un1);

         Xn:=Xn1+XDel;
         Un:=Un1+UDel;
         Err:=Sup(Max(abs(XDel),abs(UDel)));
--         Show(Scal(Err));
      end loop;
      U:=Un;
      X:=Xn;

   end NewtonMap;


   procedure EigenData(R1: in Radius; P: in out Scalar; Tay: in Taylor; Er1,Er2,Ei1,Ei2,Ev11,Ev12,Ev21,Ev22: out Scalar) is --FINE
      use Ada.Text_IO;

      A,CB,B,Det,C: Scalar;
      Tay1,Tay2: Taylor;
      D,DD,T,IT: Matrix(1..2,1..2);
   begin
      Tay1:=Der1(R1,Tay);
      Tay2:=Der2(R1,Tay);
      CB:=Sqr(Val(P,P,Tay2)/Val(P,P,Tay1))-ScalOne;
      B:=-Scal(One)/Val(P,P,Tay1);
      C:=Val(P,P,Tay1)-Sqr(Val(P,P,Tay2))/Val(P,P,Tay1);

      if Inf(CB)>Zero then
         Ei2:=ScalZero;
         Ei1:=ScalZero;
         A:=-(Val(P,P,Tay2)/Val(P,P,Tay1));
         Er2:=A+Sqrt(CB);
         Er1:=A-Sqrt(CB);
      elsif Sup(CB)<Zero then
         Ei2:=Sqrt(-CB);
         Ei1:=-Sqrt(-CB);
         A:=-(Val(P,P,Tay2)/Val(P,P,Tay1));
         Er2:=A;
         Er1:=A;
      else
         Ei2:= Sqrt(Scal(Zero,-Inf(CB)));
         Ei1:=-Sqrt(Scal(Zero,-Inf(CB)));
         A:=-(Val(P,P,Tay2)/Val(P,P,Tay1));
         Er2:=A+Sqrt(Scal(Zero,Sup(CB)));
         Er1:=A-Sqrt(Scal(Zero,Sup(CB)));
      end if;

      --Diagonalization
      Ev11:=Scal(One);
      Ev12:=(Er1-A)/B*Ev11;
      Ev21:=Scal(One);
      EV22:=(Er2-A)/B*Ev21;
      T(1,1):=Ev11;T(1,2):=Ev21;
      T(2,1):=Ev12;T(2,2):=Ev22;

      Det:=T(1,1)*T(2,2)-T(1,2)*T(2,1);
      IT(1,1):=T(2,2)/Det;IT(1,2):=-T(1,2)/Det;
      IT(2,1):=-T(2,1)/Det;IT(2,2):=T(1,1)/Det;

      D(1,1):=A;D(1,2):=B;
      D(2,1):=C;D(2,2):=A;

      New_Line;
      Put("Matrix");
      New_Line;
      Put("D(1,1)=");Show(D(1,1));
      Put("D(1,2)=");Show(D(1,2));
      Put("D(2,1)=");Show(D(2,1));
      Put("D(2,2)=");Show(D(2,2));

      Mult(IT,D);
      DD:=T;
      Mult(D,DD);
      New_Line;
      Put("Diagonalized matrix");
      New_Line;
      Put("D(1,1)=");Show(DD(1,1));
      Put("D(1,2)=");Show(DD(1,2));
      Put("D(2,1)=");Show(DD(2,1));
      Put("D(2,2)=");Show(DD(2,2));

   end EigenData;


   procedure EigenData( Bs,Cs,Es,P: in out Scalar; Tay: in Taylor; Er1,Er2,Ei1,Ei2,Ev11,Ev12,Ev21,Ev22:
out Scalar) is --FINE
      use Ada.Text_IO;

      A,CB,Det,B,C: Scalar;
      D,DD,T,IT: Matrix(1..2,1..2);
   begin
      CB:=Sqr(DerVal2(Bs,Cs,Es,P,P,Tay)/DerVal1(Bs,Cs,Es,P,P,Tay))-ScalOne;
      B:=-Scal(One)/DerVal1(Bs,Cs,Es,P,P,Tay);
      C:=DerVal1(Bs,Cs,Es,P,P,Tay)-Sqr(DerVal2(Bs,Cs,Es,P,P,Tay))/DerVal1(Bs,Cs,Es,P,P,Tay);

      if Inf(CB)>Zero then
         Ei2:=ScalZero;
         Ei1:=ScalZero;
         A:=-(DerVal2(Bs,Cs,Es,P,P,Tay)/DerVal1(Bs,Cs,Es,P,P,Tay));
         Er2:=A+Sqrt(CB);
         Er1:=A-Sqrt(CB);
      elsif Sup(CB)<Zero then
         Ei2:=Sqrt(-CB);
         Ei1:=-Sqrt(-CB);
         A:=-(DerVal2(Bs,Cs,Es,P,P,Tay)/DerVal1(Bs,Cs,Es,P,P,Tay));
         Er2:=A;
         Er1:=A;
      else
         Ei2:= Sqrt(Scal(Zero,-Inf(CB)));
         Ei1:=-Sqrt(Scal(Zero,-Inf(CB)));
         A:=-(DerVal2(Bs,Cs,Es,P,P,Tay)/DerVal1(Bs,Cs,Es,P,P,Tay));
         Er2:=A+Sqrt(Scal(Zero,Sup(CB)));
         Er1:=A-Sqrt(Scal(Zero,Sup(CB)));
      end if;

      --Diagonalization
      Ev11:=Scal(One);
      Ev12:=(Er1-A)/B*Ev11;
      Ev21:=Scal(One);
      EV22:=(Er2-A)/B*Ev21;
      T(1,1):=Ev11;T(1,2):=Ev21;
      T(2,1):=Ev12;T(2,2):=Ev22;

      Det:=T(1,1)*T(2,2)-T(1,2)*T(2,1);
      IT(1,1):=T(2,2)/Det;IT(1,2):=-T(1,2)/Det;
      IT(2,1):=-T(2,1)/Det;IT(2,2):=T(1,1)/Det;

      D(1,1):=A;D(1,2):=B;
      D(2,1):=C;D(2,2):=A;

      New_Line;
      Put("Matrix");
      New_Line;
      Put("D(1,1)=");Show(D(1,1));
      Put("D(1,2)=");Show(D(1,2));
      Put("D(2,1)=");Show(D(2,1));
      Put("D(2,2)=");Show(D(2,2));

      Mult(IT,D);
      DD:=T;
      Mult(D,DD);
      New_Line;
      Put("Diagonalized matrix");
      New_Line;
      Put("D(1,1)=");Show(DD(1,1));
      Put("D(1,2)=");Show(DD(1,2));
      Put("D(2,1)=");Show(DD(2,1));
      Put("D(2,2)=");Show(DD(2,2));

   end EigenData;



--   procedure Deriv(R1: in Radius; Phi,S: in Taylor; Lambda,Mu: in Scalar; S1,S2: in out Scalar) is
--
--   begin
--      X:=S1;
--      U:=S2;

--      S1:=Der1(R1,S);
--      S2:=Der2(R1,S);

--      for K in 1..N loop
--         X:=Y;
--         Map(Phi,S,Lambda,Mu,Y,U);
--         VS1:=Val(Y,X,S1);
--         VS2:=Val(Y,X,S2);

--         Y1:=-VS2/VS1;
--         Y2:=-Scal(One)/VS1;

--         SS1:=Val(X,Y,S1);
--         SS2:=Val(X,Y,S2);

--         DF(1,1):=Y1;
--         DF(1,2):=Y2;
--         DF(2,1):=SS1+SS2*Y1;
--         DF(2,2):=SS2*Y2;

--      end loop;
--   end Deriv;


   procedure Map(Phi,S: in Taylor; Lambda,Mu: in Scalar; S1,S2: in out Scalar) is
      --assuming that S1 and S2 are in the coord plane with the origin at (1/2,1/2)
      use Ada.Text_IO;
      Y,Del: Scalar;
      Eps: constant Rep:=Rep(1.0E-15);
      Err: Rep;
      R: Radius;
      N: Integer;
   begin

      S1:=S1+Scal(Half);
      S2:=S2+Scal(Half);

      R:=Rho(S);
      N:=0;
      while SupAbs(S1-Scal(Half))>R loop
         N:=N+1;
         Put("WARNING: RESCALING IN MAP: N=");Show(Scal(N));
         S1:=Lambda*S1;
         S2:=Mu*S2;
         --      Put("S1=");Show(S1,False);Put(",  S2="); Show(S2);
         if N>50 then
            New_Line;
            Message("Doubling.Map",Domain_Error);
         end if;
      end loop;

      <<AGAIN>> null;
      for I in 1..(2**N) loop
           Y:=Val(Scal(Zero),S1-Scal(Half),Phi)-S2;
           Err:=Rep(1.0E+20);
           while Err>Eps loop
              --Put("Y=");Show(Y);
              --if SupAbs(Y-Scal(Half))>R then
              --New_Line;
              --Message("Doubling.Map",Domain_Error);
              --end if;
              Del:=S2+Val(Y-Scal(Half),S1-Scal(Half),S);
              Y:=Y-Del;
              Err:=SupAbs(Del);
              --Put("Err in Map=");Show(Scal(Err));
           end loop;
           --Put("Y=");Show(Y);
--           if SupAbs( Y-Scal(Half) )>R then
--              Put("WARNING: RETURNING");
--              S1:=Lambda*S1;
--              S2:=Mu*S2;
--              N:=N+1;
--              New_Line;
--              goto AGAIN;
--           end if;
           S2:=Val(S1-Scal(Half),Y-Scal(Half),S);
           S1:=Y;
      end loop;

      for I in 1..N loop
         S1:=S1/Lambda;
         S2:=S2/Mu;
      end loop;

      S1:=S1-Scal(Half);
      S2:=S2-Scal(Half);
   end Map;


   procedure Map(Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar; S1,S2: in out Scalar) is
      --assuming that S1 and S2 are in the coord plane with the origin at (1/2,1/2)
      use Ada.Text_IO;
      Y,Del,Beta,Epsilon,Der,Fact: Scalar;
      Eps: constant Rep:=Rep(1.0E-15);
      Err: Rep;
      R: Radius;
      N: Integer;
   begin
      S1:=S1+Scal(Half);
      S2:=S2+Scal(Half);

      R:=Rho(S);
      N:=0;
      while SupAbs(S1-Scal(Half))>R loop
         N:=N+1;
         Put("WARNING: RESCALING IN MAP: N=");Show(Scal(N));
         S1:=Lambda*S1;
         S2:=Mu*S2;
         --      Put("S1=");Show(S1,False);Put(",  S2="); Show(S2);
         if N>50 then
            New_Line;
            Message("Doubling.Map",Domain_Error);
         end if;
      end loop;

      <<AGAIN>> null;
      for I in 1..(2**N) loop
           Y:=Val(B,C,E,Scal(Zero),S1-Scal(Half),Phi)-S2;
           Err:=Rep(1.0E+20);
           while Err>Eps loop
              Del:=Center(S2+Val(B,C,E,Y-Scal(Half),S1-Scal(Half),S));
              Y:=Center(Y-Del);
              Err:=SupAbs(Del);
           end loop;
           if not Trunc then
              Der:=DerVal1(B,C,E,Y-Scal(Half),S1-Scal(Half),S);
              Fact:=Center(Scal(One)/abs(Der));
              Der:=Scal(One)-Fact*DerVal1(B,C,E,Y-Scal(Half),S1-Scal(Half),S);
              Epsilon:=Fact*Abs(S2+Val(B,C,E,Y-Scal(Half),S1-Scal(Half),S));
              Beta:=Scal(Sup(abs(Epsilon/(Scal(One)-Abs(Der)))));
              Y:=Center(Y)+Ball(Beta);
              Der:=Scal(One)-DerVal1(B,C,E,Y-Scal(Half),S1-Scal(Half),S);
              while  Epsilon>(Scal(One)-Abs(Der))*Beta loop
                Beta:=Rep(1.001)*Beta;
                Y:=Center(Y)+Ball(Beta);
                Der:=Scal(One)-Fact*DerVal1(B,C,E,Y-Scal(Half),S1-Scal(Half),S);
              end loop;
           end if;
           S2:=Val(B,C,E,S1-Scal(Half),Y-Scal(Half),S);
           S1:=Y;
      end loop;

      for I in 1..N loop
         S1:=S1/Lambda;
         S2:=S2/Mu;
      end loop;

      S1:=S1-Scal(Half);
      S2:=S2-Scal(Half);
   end Map;



   procedure Map(Phi,S: in Taylor; F1,F2: out Taylor) is
      use Ada.Text_IO;
      Y: Taylor;
      Eps: constant Rep:=Rep(1.0E-15);
      R: Radius;
   begin
      R:=Rho(S);
      SetZero(R,Y);
      Y:=Y+Comp2(Monom(R,1,0,Scal(One)),Phi)-(Monom(R,0,1,Scal(One))+Monom(R,0,0,Scal(Half)));

      for I in 1..50 loop
         Y:=Y-(Monom(R,0,1,Scal(One))+Monom(R,0,0,Scal(Half)))-Comp(Y,Monom(R,1,0,Scal(One)),S);
      end loop;

      SetZero(R,F1);
      SetZero(R,F2);
      F1:=F1+Y;
      F2:=F2+Comp2(Y,S)-Monom(R,0,0,Scal(Half));
   end Map;



   procedure LinOp(P,A,B,C,D: in Scalar; F1,F2: in Taylor; T1,T2: in out Taylor) is
      use Ada.Text_IO;
      R: Radius;
      T11: Taylor;
   begin
      R:=Rho(T1);

--      FF1:=Comp2(Monom(R,0,1,Scal(-One)),F1);
--      FF2:=-Comp2(Monom(R,0,1,Scal(-One)),F2);
--      for I in 1..3 loop
         T1:=Comp(F1,F2,T1);
         T2:=Comp(F1,F2,T2);
--      end loop;

--      for I in 1..3 loop
         T11:=T1;
         T1:=A*T1+B*T2;
         T2:=C*T11+D*T2;
--      end loop;



--      T11:=T1;
--      T1:=A*T1+B*T2;
--      T2:=C*T11+D*T2;
--      T11:=T1;
--      T1:=Comp(T1,T2,FF1);
--      T2:=Comp(T11,T2,FF2);

   end LinOp;


   procedure LinCoord(Deg: in Integer;P,A,B,C,D: in Scalar; F1,F2: in Taylor; T1,T2: in out Taylor) is
      use Ada.Text_IO;
      TDelt1,TDelt2,TT1,TT2,TTT1,TTT2,H: Taylor;
      Eps: Rep:=Rep(1.0E-17);
      ID: Matrix(1..2*Deg*Deg,1..2*Deg*Deg);
      Delt: Scalar:=Scal(Rep(1.0E-7));
      N,M: Integer;
      Pvt: Pivot(1..2*Deg*Deg);
      Cond: Scalar;
      V,Work:  Vector(1..2*Deg*Deg);
      R: Radius;
      Err: Rep;
   begin
      R:=Rep(0.5);

      Err:=Rep(1.0E+20);
--      while Err>Eps loop
      for KK in 1..30 loop

      SetZero(R,TT1);
      SetZero(R,TT2);
      SetZero(R,TDelt1);
      SetZero(R,TDelt2);

      TT1:=T1;
      TT2:=T2;
      LinOp(P,A,B,C,D,F1,F2,TT1,TT2);
      N:=1;
      for I in 1..Deg loop
         for J in 1..Deg loop
            M:=1;
            H:=Monom(R,I,J,Scal(One));
            TTT1:=T1+Delt*H;
            TTT2:=T2;
            LinOp(P,A,B,C,D,F1,F2,TTT1,TTT2);
            for L in 1..Deg loop
               for K in 1..Deg loop
                  ID(M,N):=(Scal(-One)/Delt)*Component(L,K,(TTT1-TT1));
                  ID(M+Deg*Deg,N):=(Scal(-One)/Delt)*Component(L,K,(TTT2-TT2));
                  M:=M+1;
               end loop;
            end loop;

            M:=1;
            TTT1:=T1;
            TTT2:=T2+Delt*H;
            LinOp(P,A,B,C,D,F1,F2,TTT1,TTT2);
            for L in 1..Deg loop
               for K in 1..Deg loop
                  ID(M,N+Deg*Deg):=(Scal(-One)/Delt)*Component(L,K,(TTT1-TT1));
                  ID(M+Deg*Deg,N+Deg*Deg):=(Scal(-One)/Delt)*Component(L,K,(TTT2-TT2));
                  M:=M+1;
               end loop;
            end loop;


            N:=N+1;
         end loop;
      end loop;

      for I in ID'Range(1) loop
         ID(I,I):=Scal(One)+ID(I,I);
      end loop;


        --Derivative
--      for KK in 1..1 loop

      TT1:=T1;
      TT2:=T2;
      LinOp(P,A,B,C,D,F1,F2,T1,T2);

         N:=1;
         for I in 1..Deg loop
            for J in 1..Deg loop
               V(N):=Component(I,J,T1-TT1);
               V(N+Deg*Deg):=Component(I,J,T2-TT2);
               N:=N+1;
            end loop;
         end loop;

         Decomp(ID,Cond,Pvt,Work);
         Solve(ID,V,Pvt);

         N:=1;
         SetZero(R,TDelt1);
         SetZero(R,TDelt2);
         for I in 1..Deg loop
            for J in 1..Deg loop
               TDelt1:=TDelt1+Monom(R,I,J,V(N));
               TDelt2:=TDelt2+Monom(R,I,J,V(N+Deg*Deg));
               N:=N+1;
            end loop;
         end loop;

         T1:=TT1+TDelt1;
         T2:=TT2+TDelt2;
         Err:=Sup(Norm(R,TDelt1)+Norm(R,TDelt2));
         New_Line;
         Put("Err in LinCoord=");Show(Scal(Err),False);
      end loop;
   end LinCoord;


   procedure LinMapDer(T11,T12,T21,T22: in out Scalar; F1,F2: in Taylor) is
      --assuming that S1 and S2 are in the coord plane centered at (1/2,1/2)
      use Ada.Text_IO;
      T,IT,D,DD: Matrix(1..2,1..2);
      R1: Radius;
      Det,Ev11,Ev12,Ev21,Ev22,Er2,Er1,Er11: Scalar;
   begin
      R1:=Rho(F1);
      Det:=T11*T22-T12*T21;
      IT(1,1):=T22/Det; IT(1,2):=-T12/Det;
      IT(2,1):=-T21/Det;IT(2,2):=T11/Det;

      T(1,1):=T11;
      T(1,2):=T12;
      T(2,1):=T21;
      T(2,2):=T22;

      D(1,1):=(Val(Scal(Zero),Scal(Zero),Der1(R1,F1)));
      D(1,2):=(Val(Scal(Zero),Scal(Zero),Der2(R1,F1)));
      D(2,1):=(Val(Scal(Zero),Scal(Zero),Der1(R1,F2)));
      D(2,2):=(Val(Scal(Zero),Scal(Zero),Der2(R1,F2)));
      D:=IT*(D*T);
      New_Line;
      Show(D(1,1));
      Show(D(1,2));
      Show(D(2,1));
      Show(D(2,2));

      Er2:=D(1,1)+Sqrt(D(1,2)*D(2,1));
      Er1:=D(1,1)-Sqrt(D(1,2)*D(2,1));
      if abs(Er2)>abs(Er1) then
         Er11:=Er1;
         Er1:=Er2;
         Er2:=Er11;
      end if;

      --Diagonalization
      Ev11:=Scal(One);
      Ev12:=(Er1-D(1,1))/D(1,2)*Ev11;
      Ev21:=Scal(One);
      EV22:=(Er2-D(1,1))/D(1,2)*Ev21;
      T(1,1):=Ev11;T(1,2):=Ev21;
      T(2,1):=Ev12;T(2,2):=Ev22;

      Det:=T(1,1)*T(2,2)-T(1,2)*T(2,1);
      IT(1,1):=T(2,2)/Det;IT(1,2):=-T(1,2)/Det;
      IT(2,1):=-T(2,1)/Det;IT(2,2):=T(1,1)/Det;


      Mult(IT,D);
      DD:=T;
      Mult(D,DD);
      New_Line;
      Put("Diagonalized matrix LinMapDer");
      New_Line;
      Put("D(1,1)=");Show(DD(1,1));
      Put("D(1,2)=");Show(DD(1,2));
      Put("D(2,1)=");Show(DD(2,1));
      Put("D(2,2)=");Show(DD(2,2));

      T11:=IT(1,1);
      T12:=IT(1,2);
      T21:=IT(2,1);
      T22:=IT(2,2);

   end LinMapDer;



   procedure Distortion(Co,Q,N:in Integer; R1,R2: in Radius;
ME1,ME2,LsP0,LsN0,LuP0,LuN0,LsP1,LsN1,LuP1,LuN1,B,C,E,Lambda,Mu,Sx0,Su0,Sx1,Su1,CenX,CenY,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1:
in Scalar; Phi,S,Sigma,T1,T2: in Taylor;
Cstar,Epsilon,Gamma,DF11,LocDist,MinExp,MaxExp,MinAng,MaxAng: out Scalar; Hyp4: out Boolean) is --FINE
      --R1 is the domain for derivatives
      --R2 is the domain for second derivatives
      --Ls,Lu are the sizes of the rectangle
      --N is the # of iterates of the map
      --Wx,Wu are the coordinates of the vector "along the stable manifold"

      use Ada.Text_IO;

      AD,Det,X,Y,U,EDel,XDel,Xi,Eta,XX,YY,UU,VS1,VS2,VS11,VS12,VS21,VS22,SS1,SS2,SS11,SS12,SS21,SS22,XXDelt,UUDelt,XXX,UUU,Y1,Y2,Y11,Y22,Y12,Y21,VDelta: Scalar;
      Pr1,Pr2,Pt1,Pt2,Er1,Er2,Ei1,Ei2,Gama,Eps,GradLeng,GradLen,TV,LocEpsilon,LocDistor,LocGamma,LocCstar,Expn,Angl: Scalar;
      S1,S2,S11,S12,S21,S22: Taylor;
      DF,DG,Inv,Tr,ITr: Matrix(1..2,1..2);
      DDF,DDG: Tensor(1..2,1..2,1..2);
      MaxEw,MaxEv,MinEw,MinEv,Vec,Vecc: Vector(1..2);
      II,MI: Integer;
   begin

      Inv(1,1):=Scal(One);
      Inv(1,2):=Scal(Zero);
      Inv(2,1):=Scal(Zero);
      Inv(2,2):=-Scal(One);

      S1:=Der1(R1,S);
      S2:=Der2(R1,S);
      S11:=Der1(R2,Der1(R1,S));
      S22:=Der2(R2,Der2(R1,S));
      S21:=Der2(R2,Der1(R1,S));
      S12:=S21;


      --Hypothesis 2) and 3): definitions of eps and gamma
      New_Line;
      Put("DUARTE DISTORTION THEOREM: computing constants eps and gamma...");

      XDel:=((LsP0+LsN0))/Scal(Q);
      EDel:=((LuP0+LuN0))/Scal(Q);

      Epsilon:=Scal(Zero);
      Gamma:=Scal(Zero);
      DF11:=Scal(Zero);
      GradLeng:=Scal(Zero);
      LocDist:=Scal(Zero);


      MaxAng:=Scal((-10000000.0));
      MinAng:=Scal((10000000.0));
      MinExp:=Scal(10000.0);
      MaxExp:=Scal(Zero);
      Xi:=-(LsN0);
      II:=0;
      for I in 1..Q loop
         Eta:=-(LuN0);
         for J in 1..Q loop
            if Scal(100*II)/Scal(Q*Q)=Scal(100*II/(Q*Q)) then
               JS(Current_Output,100*II/(Q*Q));
            end if;

            MI:=1;


<<JUMP>>            XXX:=Center(Eta*Wx0+Xi*Vx0+Sx0);
            UUU:=Center(Eta*Wu0+Xi*Vu0+Su0);

            XXDelt:=(Scal(Sup(XXX))-Scal(Inf(XXX)))/Scal(MI);
            UUDelt:=(Scal(Sup(UUU))-Scal(Inf(UUU)))/Scal(MI);
            X:=Center(Scal(Inf(XXX),Sup(Scal(Inf(XXX))+XXDelt)));
            U:=Center(Scal(Inf(UUU),Sup(Scal(Inf(UUU))+UUDelt)));
            for III in 1..MI loop

            -----------------------------------------

            DG(1,1):=Scal(One);
            DG(2,2):=Scal(One);
            DG(2,1):=Scal(Zero);
            DG(1,2):=Scal(Zero);

            YY:=Center(Lambda*(X+Scal(Half))-Scal(Half));
            UU:=Center(Mu*(U+Scal(Half))-Scal(Half));
            for K in 1..N loop
               Map(Phi,Sigma,B,C,E,Lambda,Mu,YY,UU);
            end loop;
        --    YY:=Center(YY);
        --    UU:=Center(UU);
            for K in 1..N loop
               XX:=YY;
               Map(Phi,Sigma,B,C,E,Lambda,Mu,YY,UU);
               VS1:=DerVal1(B,C,E,YY,XX,Sigma);
               VS2:=DerVal2(B,C,E,YY,XX,Sigma);

               Y1:=-VS2/VS1;
               Y2:=-Scal(One)/VS1;

               SS1:=DerVal1(B,C,E,XX,YY,Sigma);
               SS2:=DerVal2(B,C,E,XX,YY,Sigma);

               DF(1,1):=Y1;
               DF(1,2):=Y2;
               DF(2,1):=SS1+SS2*Y1;
               DF(2,2):=SS2*Y2;

               Mult(DF,DG);
            end loop;
            AD:=DG(1,1)+DG(2,2);
            Det:=Sqr(AD)-Scal(4.0);
            if Inf(Det)>Zero then
               Ei2:=ScalZero;
               Ei1:=ScalZero;
               Er2:=Half*(AD+Sqrt(Det));
               Er1:=Half*(AD-Sqrt(Det));
            elsif Sup(Det)<Zero then
               Ei2:=Half*Sqrt(-Det);
               Ei1:=-Half*Sqrt(-Det);
               Er2:=Half*AD;
               Er1:=Half*AD;
            else
         Ei2:= Half*Sqrt(Scal(Zero,-Inf(Det)));
         Ei1:=-Half*Sqrt(Scal(Zero,-Inf(Det)));
         Er2:=Half*(AD+Sqrt(Scal(Zero,Sup(Det))));
         Er1:=Half*(AD-Sqrt(Scal(Zero,Sup(Det))));
      end if;

      if Sup(abs(Ei1))/=Zero or  Sup(abs(Ei2))/=Zero then
         New_Line;
         Put("E1=");Show(Er1,False);Put("+I*");Show(Ei1);
         Put("E2=");Show(Er2,False);Put("+I*");Show(Ei2);
      end if;

      ----------------------------
      Y:=X;

      ITr(1,1):=(Val(X-CenX,U-CenY,Der1(R1,T1)));
      ITr(1,2):=(Val(X-CenX,U-CenY,Der2(R1,T1)));
      ITr(2,1):=(Val(X-CenX,U-CenY,Der1(R1,T2)));
      ITr(2,2):=(Val(X-CenX,U-CenY,Der2(R1,T2)));
      Det:=ITr(1,1)*ITr(2,2)-ITr(1,2)*ITr(2,1);
      Tr(1,1):=ITr(2,2)/Det; Tr(1,2):=-ITr(1,2)/Det;
      Tr(2,1):=-ITr(2,1)/Det;Tr(2,2):=ITr(1,1)/Det;

      DG(1,1):=Scal(One);
      DG(2,2):=Scal(One);
      DG(2,1):=Scal(Zero);
      DG(1,2):=Scal(Zero);


      for K in 1..N loop
         X:=Y;
         Map(Phi,Sigma,B,C,E,Lambda,Mu,Y,U);
         VS1:=DerVal1(B,C,E,Y,X,Sigma);
         VS2:=DerVal2(B,C,E,Y,X,Sigma);
         VS12:=Val(Y,X,S12);
         VS21:=VS12;
         VS11:=Val(Y,X,S11);
         VS22:=Val(Y,X,S22);

         Y1:=-VS2/VS1;
         Y2:=-Scal(One)/VS1;
         Y11:=-(VS21*Y1+VS22)/VS1+VS2*(VS11*Y1+VS12)/Sqr(VS1);
         Y22:= VS11*Y2/Sqr(VS1);
         Y21:= (VS11*Y1+VS12)/Sqr(VS1);
         Y12:= (VS11*Y1+VS12)/Sqr(VS1);


         SS1:=DerVal1(B,C,E,X,Y,Sigma);
         SS2:=DerVal2(B,C,E,X,Y,Sigma);
         SS11:=Val(X,Y,S11);
         SS22:=Val(X,Y,S22);
         SS21:=Val(X,Y,S21);
         SS12:=SS21;

         DF(1,1):=Y1;
         DF(1,2):=Y2;
         DF(2,1):=SS1+SS2*Y1;
         DF(2,2):=SS2*Y2;

         --Second derivatives
         DDF(1,1,1):=Y11;
         DDF(1,1,2):=Y12;
         DDF(1,2,1):=Y21;
         DDF(1,2,2):=Y22;
         DDF(2,1,1):=SS11+SS12*Y1+SS21*Y1+SS22*Sqr(Y1)+SS2*Y11;
         DDF(2,1,2):=SS12*Y2+SS22*(Y2*Y1)+SS2*Y12;
         DDF(2,2,1):=(SS21+SS22*Y1)*Y2+SS2*Y21;
         DDF(2,2,2):=SS22*Sqr(Y2)+SS2*Y22;

         if K=1 then
            DDG:=DDF;
         else
            DDG:=DF*DDG+Mult(DDF,DG,DG);
         end if;
         Mult(DF,DG);

      end loop;

      ITr(1,1):=(Val(Y-CenX,U-CenY,Der1(R1,T1)));
      ITr(1,2):=(Val(Y-CenX,U-CenY,Der2(R1,T1)));
      ITr(2,1):=(Val(Y-CenX,U-CenY,Der1(R1,T2)));
      ITr(2,2):=(Val(Y-CenX,U-CenY,Der2(R1,T2)));

      DG:=ITr*(DG*Tr);

      --          goto AHA;

      MaxEv(1):=Scal(One);
      MaxEv(2):=ME1;
      MinEv(1):=Scal(One);
      MinEv(2):=ME2;
      MaxEw:=MaxEv;
      MinEw:=MinEv;

      Mult(Tr,MaxEw);
      Mult(Tr,MinEw);
      Angl:=MaxEw(2)/MaxEw(1);
      if Inf(Angl)<Inf(MinAng) then
         MinAng:=Angl;
      end if;
      if Sup(Angl)>Sup(MaxAng) then
         MaxAng:=Angl;
      end if;
      Angl:=MinEw(2)/MinEw(1);
      if Inf(Angl)<Inf(MinAng) then
         MinAng:=Angl;
      end if;
      if Sup(Angl)>Sup(MaxAng) then
         MaxAng:=Angl;
      end if;


--      New_Line;
--      Show("Max Vec=");
--      Show(MaxEw(1),False);
--      Show(MaxEw(2));
--      Show("Min Vec=");
--      Show(MinEw(1),False);
--      Show(MinEw(2));

      Pr1:=((Y-Sx0)*Wx0+(U-Su0)*Wu0);
      Pr2:=((Y-Sx0)*Vx0+(U-Su0)*Vu0);
      Pt1:=((Y-Sx1)*Wx1+(U-Su1)*Wu1);
      Pt2:=((Y-Sx1)*Vx1+(U-Su1)*Vu1);

      if  ((Scal(Sup(Pr1))>=-LsN0 and Scal(Inf(Pr1))<=LsP0) and (Scal(Sup(Pr2))>=-LuN0 and Scal(Inf(Pr2))<=LuP0) ) or
        ((Scal(Sup(Pt1))>=-LsN1 and Scal(Inf(Pt1))<=LsP1) and (Scal(Sup(Pt2))>=-LuN1 and Scal(Inf(Pt2))<=LuP1) ) then

--         Er1:=Half*( (DG(1,1)+DG(2,2)) + Sqrt( Sqr( DG(1,1)+DG(2,2) )-Scal(4.0) ) );
--         Er2:=Half*( (DG(1,1)+DG(2,2)) - Sqrt( Sqr( DG(1,1)+DG(2,2) )-Scal(4.0) ) );
--         if Abs(Er1)>Abs(Er2) then
--            Er2:=Er1;
--         end if;

--         Ev1:=Scal(One);
--         Ev2:=(Er1-DG(1,1))/DG(1,2)*Ev1;
--         Det:=Sqrt(Sqr(Ev1)+Sqr(Ev2));
--         Ev11:=Ev1;
--       Ev1:=Ev1/Det;
--       Ev2:=Ev2/Det;





         VDelta:=Scal(One)/Scal(50.0);
         for JJJ in 1..50 loop
            Vec:=Scal(Inf(Scal(JJJ-1)*VDelta),Sup(Scal(JJJ)*VDelta))*MaxEv+
              (Scal(One)-Scal(Inf(Scal(JJJ-1)*VDelta),Sup(Scal(JJJ)*VDelta)))*MinEv;
            Vecc:=Vec;
            Mult(DG,Vec);

            if ( Sqrt(Sqr(Vec(1))+Sqr(Vec(2)))<=
                   Sqrt(Sqr(Vecc(1))+Sqr(Vecc(2))) ) then
               Put("WARNING: UNSTABLE CONES ARE NOT EXPANDED");
            end if;

            Expn:=Sqrt(Sqr(Vec(1))+Sqr(Vec(2)))/Sqrt(Sqr(Vecc(1))+Sqr(Vecc(2)));
            if Sup(Expn)>Sup(MaxExp) then
               MaxExp:=Expn;
            end if;
            if Inf(Expn)<Inf(MinExp) then
               MinExp:=Expn;
            end if;


            -- Det:=Sqrt(Sqr(MaxEw(1))+Sqr(MaxEw(2)));
            -- MaxEw(1):=MaxEw(1)/Sup(Det);
            -- MaxEw(2):=MaxEw(2)/Sup(Det);
            Vec(2):=Vec(2)/Inf(Vec(1));

            if Sup(Vec(2))>Sup(ME1) or Inf(Vec(2))<Inf(ME2)  then
              Put("WARNING: UNSTABLE CONES ARE NOT INVARIANT");
            end if;
         end loop;

         end if;

--      <<AHA>> null;


          Mult(ITr,DDG);
          Mult(DDG,Tr,Tr);

          if abs(DG(1,1))>DF11 then  DF11:=abs(DG(1,1)); end if;

          GradLen:=Sqrt(Sqr(DDG(1,1,1))+Sqr(DDG(1,1,2)) )/Abs(DG(1,1));

          if GradLen>GradLeng then GradLeng:=GradLen; end if;

          if abs(DG(1,1))<Scal(One) then
--             Put("PROBLEM in DIRECT MAP: |DG(1,1)| is less then 1");
--             New_Line;
MI:=5*MI;---Show(Scal(MI));New_Line;
 goto JUMP;
          end if;
          if abs(DG(2,2))>Scal(One) then
--             Put("PROBLEM in DIRECT MAP: |DG(2,2)| is larger then 1");
--             New_Line;
MI:=5*MI;---Show(Scal(MI));New_Line;
 goto JUMP;
          end if;

          Eps:=abs(DG(1,2))/(Abs(DG(1,1))-Scal(One));
          if Eps>Epsilon then Epsilon:=Eps; end if;
          LocEpsilon:=Eps;
          Eps:=abs(DG(2,1))/(Abs(DG(1,1))-Scal(One));
          if Eps>Epsilon then Epsilon:=Eps; end if;
          if Eps>LocEpsilon then LocEpsilon:=Eps; end if;


          Gama:=abs(DDG(1,1,1))/(abs(DG(1,1))*(abs(DG(1,1))-Scal(One)));
          if Gama>Gamma then Gamma:=Gama; end if;
          LocGamma:=Gama;
          Gama:=abs(DDG(2,2,2))/(abs(DG(1,1))*(abs(DG(1,1))-Scal(One)));
          if Gama>Gamma then Gamma:=Gama; end if;
          if Gama>LocGamma then LocGamma:=Gama; end if;
          Gama:=abs(DDG(1,1,2))/(abs(DG(1,1))-Scal(One));
          if Gama>Gamma then Gamma:=Gama; end if;
          if Gama>LocGamma then LocGamma:=Gama; end if;
          Gama:=abs(DDG(1,2,2))/(abs(DG(1,1))-Scal(One));
          if Gama>Gamma then Gamma:=Gama; end if;
          if Gama>LocGamma then LocGamma:=Gama; end if;
          Gama:=abs(DDG(2,1,1))/(abs(DG(1,1))-Scal(One));
          if Gama>Gamma then Gamma:=Gama; end if;
          if Gama>LocGamma then LocGamma:=Gama; end if;
          Gama:=abs(DDG(2,1,2))/(abs(DG(1,1))-Scal(One));
          if Gama>Gamma then Gamma:=Gama; end if;
          if Gama>LocGamma then LocGamma:=Gama; end if;

          LocCstar:=LocEpsilon*abs(DG(1,1));


            LocDistor:=Scal(4.0)*(LocCstar+Scal(3.0))*LocGamma+Two*LocEpsilon;
            if LocDistor>LocDist then
               LocDist:=LocDistor;
            end if;

            X:=Scal(Sup(X),Sup(Scal(Sup(X))+XXDelt));
            U:=Scal(Sup(U),Sup(Scal(Sup(U))+UUDelt));
end loop;

            II:=II+1;

            Eta:=Scal(Sup(Eta),Sup(Scal(Sup(Eta))+Scal(Sup(EDel))));
         end loop;
         Xi:=Scal(Sup(Xi),Sup(Scal(Sup(Xi))+Scal(Sup(XDel))));
      end loop;


       Cstar:=DF11*Epsilon;

      --Hypothesis 4)
      TV:=(LsN0+LsP0)*(LuN0+LuP0)*((Wx0*Vx0+Wu0*Vu0)/(Sqrt(Wx0*Wx0+Wu0*Wu0)*Sqrt(Vx0*Vx0+Vu0*Vu0)))*GradLeng;

      Put("done");

      if ( TV<Gamma*(Scal(One)-Scal(One)/abs(DF11)) ) then
         Hyp4:=True;
          New_Line;
        Put("Hypothesis 4 is verified");
      else
         Hyp4:=False;
         New_Line;
        Put("Hypothesis 4 is NOT verified");
      end if;


   end Distortion;




   procedure JS(F: in File_Type; J: in Integer) is

   begin
      if    J=0 then  New_Line; Put(F,"0%");
      elsif J=1 then  New_Line; Put(F,"1%");
      elsif J=2 then  New_Line; Put(F,"2%");
      elsif J=3 then  New_Line; Put(F,"3%");
      elsif J=4 then  New_Line; Put(F,"4%");
      elsif J=5 then  New_Line; Put(F,"5%");
      elsif J=6 then  New_Line; Put(F,"6%");
      elsif J=7 then  New_Line; Put(F,"7%");
      elsif J=8 then  New_Line; Put(F,"8%");
      elsif J=9 then  New_Line; Put(F,"9%");

      elsif J=10 then  New_Line; Put(F,"10%");
      elsif J=11 then  New_Line; Put(F,"11%");
      elsif J=12 then  New_Line; Put(F,"12%");
      elsif J=13 then  New_Line; Put(F,"13%");
      elsif J=14 then  New_Line; Put(F,"14%");
      elsif J=15 then  New_Line; Put(F,"15%");
      elsif J=16 then  New_Line; Put(F,"16%");
      elsif J=17 then  New_Line; Put(F,"17%");
      elsif J=18 then  New_Line; Put(F,"18%");
      elsif J=19 then  New_Line; Put(F,"19%");

      elsif J=20 then  New_Line; Put(F,"20%");
      elsif J=21 then  New_Line; Put(F,"21%");
      elsif J=22 then  New_Line; Put(F,"22%");
      elsif J=23 then  New_Line; Put(F,"23%");
      elsif J=24 then  New_Line; Put(F,"24%");
      elsif J=25 then  New_Line; Put(F,"25%");
      elsif J=26 then  New_Line; Put(F,"26%");
      elsif J=27 then  New_Line; Put(F,"27%");
      elsif J=28 then  New_Line; Put(F,"28%");
      elsif J=29 then  New_Line; Put(F,"29%");

      elsif J=30 then  New_Line; Put(F,"30%");
      elsif J=31 then  New_Line; Put(F,"31%");
      elsif J=32 then  New_Line; Put(F,"32%");
      elsif J=33 then  New_Line; Put(F,"33%");
      elsif J=34 then  New_Line; Put(F,"34%");
      elsif J=35 then  New_Line; Put(F,"35%");
      elsif J=36 then  New_Line; Put(F,"36%");
      elsif J=37 then  New_Line; Put(F,"37%");
      elsif J=38 then  New_Line; Put(F,"38%");
      elsif J=39 then  New_Line; Put(F,"39%");

      elsif J=40 then  New_Line; Put(F,"40%");
      elsif J=41 then  New_Line; Put(F,"41%");
      elsif J=42 then  New_Line; Put(F,"42%");
      elsif J=43 then  New_Line; Put(F,"43%");
      elsif J=44 then  New_Line; Put(F,"44%");
      elsif J=45 then  New_Line; Put(F,"45%");
      elsif J=46 then  New_Line; Put(F,"46%");
      elsif J=47 then  New_Line; Put(F,"47%");
      elsif J=48 then  New_Line; Put(F,"48%");
      elsif J=49 then  New_Line; Put(F,"49%");

      elsif J=50 then  New_Line; Put(F,"50%");
      elsif J=51 then  New_Line; Put(F,"51%");
      elsif J=52 then  New_Line; Put(F,"52%");
      elsif J=53 then  New_Line; Put(F,"53%");
      elsif J=54 then  New_Line; Put(F,"54%");
      elsif J=55 then  New_Line; Put(F,"55%");
      elsif J=56 then  New_Line; Put(F,"56%");
      elsif J=57 then  New_Line; Put(F,"57%");
      elsif J=58 then  New_Line; Put(F,"58%");
      elsif J=59 then  New_Line; Put(F,"59%");

      elsif J=60 then  New_Line; Put(F,"60%");
      elsif J=61 then  New_Line; Put(F,"61%");
      elsif J=62 then  New_Line; Put(F,"62%");
      elsif J=63 then  New_Line; Put(F,"63%");
      elsif J=64 then  New_Line; Put(F,"64%");
      elsif J=65 then  New_Line; Put(F,"65%");
      elsif J=66 then  New_Line; Put(F,"66%");
      elsif J=67 then  New_Line; Put(F,"67%");
      elsif J=68 then  New_Line; Put(F,"68%");
      elsif J=69 then  New_Line; Put(F,"69%");

      elsif J=70 then  New_Line; Put(F,"70%");
      elsif J=71 then  New_Line; Put(F,"71%");
      elsif J=72 then  New_Line; Put(F,"72%");
      elsif J=73 then  New_Line; Put(F,"73%");
      elsif J=74 then  New_Line; Put(F,"74%");
      elsif J=75 then  New_Line; Put(F,"75%");
      elsif J=76 then  New_Line; Put(F,"76%");
      elsif J=77 then  New_Line; Put(F,"77%");
      elsif J=78 then  New_Line; Put(F,"78%");
      elsif J=79 then  New_Line; Put(F,"79%");

      elsif J=80 then  New_Line; Put(F,"80%");
      elsif J=81 then  New_Line; Put(F,"81%");
      elsif J=82 then  New_Line; Put(F,"82%");
      elsif J=83 then  New_Line; Put(F,"83%");
      elsif J=84 then  New_Line; Put(F,"84%");
      elsif J=85 then  New_Line; Put(F,"85%");
      elsif J=86 then  New_Line; Put(F,"86%");
      elsif J=87 then  New_Line; Put(F,"87%");
      elsif J=88 then  New_Line; Put(F,"88%");
      elsif J=89 then  New_Line; Put(F,"89%");

      elsif J=90 then  New_Line; Put(F,"90%");
      elsif J=91 then  New_Line; Put(F,"91%");
      elsif J=92 then  New_Line; Put(F,"92%");
      elsif J=93 then  New_Line; Put(F,"93%");
      elsif J=94 then  New_Line; Put(F,"94%");
      elsif J=95 then  New_Line; Put(F,"95%");
      elsif J=96 then  New_Line; Put(F,"96%");
      elsif J=97 then  New_Line; Put(F,"97%");
      elsif J=98 then  New_Line; Put(F,"98%");
      elsif J=99 then  New_Line; Put(F,"99%");

      end if;
   end JS;


   procedure DrawHS(DoLin: in Boolean; NameT,NameB,NameR,NameL: in String;Q,N: in Integer; LsP,LsN,LuP,LuN,Lambda,Mu,Bs,Cs,Es,P1,P2,CenX,CenY,Wx,Wu,Vx,Vu: in Scalar; Phi,S,T1,T2: in Taylor) is --FINE

      use Reps.IO;
      use Ada.Text_IO;

      X,Y,U,Xi,Eta,U1,XDel,EDel,Y1,UU,YY: Scalar;
      F,FF: File_Type;
      R1: Radius;
      Shift:Rep:=Rep(0.1);
      Fact: Scalar:=Scal(Rep(0.1));
   begin
      R1:=Rep(0.8)*Rho(S);

      XDel:=Rep(1.0)*((LsP+LsN))/Rep(Q);
      EDel:=Rep(1.0)*((LuP+LuN))/Rep(Q);

      --Upper side
      Create(F,Out_File,NameT);
      Eta:=(LuP);
      Xi:=-(LsN);
      for I in 1..Q loop
         X:=Eta*Wx+Xi*Vx;
         U:=Eta*Wu+Xi*Vu;
         Y:=X+P1;
         U:=U+P2;

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;

         Txt_Put(F,Approx(Y1+CenX)+Half);
         Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);

         YY:=Y;
         UU:=U;
         for K in 1..N loop
            Map(Phi,S,Bs,Cs,Es,Lambda,Mu,Y,U);
         end loop;

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;

         Txt_Put(F,Approx(Y1+CenX)+Half);
         Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);

         Y:=YY;
         U:=-(UU+Scal(Half))-Scal(Half);
         for K in 1..N loop
            Map(Phi,S,Bs,Cs,Es,Lambda,Mu,Y,U);
         end loop;
         U:=-(U+Scal(Half))-Scal(Half);

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;

         Txt_Put(F,Approx(Y1+CenX)+Half);
         Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);


         Xi:=Xi+XDel;
      end loop;
      Close(F);


      --Lower side
      Create(F,Out_File,NameB);
      Eta:=-(LuN);
      Xi:=-(LsN);
      for I in 1..Q loop
         X:=Eta*Wx+Xi*Vx;
         U:=Eta*Wu+Xi*Vu;

         Y:=X+P1;
         U:=U+P2;

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;

         Txt_Put(F,Approx(Y1+CenX)+Half);
        Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);

         YY:=Y;
         UU:=U;
         for K in 1..N loop
            Map(Phi,S,Bs,Cs,Es,Lambda,Mu,Y,U);
         end loop;

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;

         Txt_Put(F,Approx(Y1+CenX)+Half);
         Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);


         Y:=YY;
         U:=-(UU+Scal(Half))-Scal(Half);
         for K in 1..N loop
            Map(Phi,S,Bs,Cs,Es,Lambda,Mu,Y,U);
         end loop;
         U:=-(U+Scal(Half))-Scal(Half);

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;

         Txt_Put(F,Approx(Y1+CenX)+Half);
         Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);
         Xi:=Xi+XDel;
      end loop;
      Close(F);


      --Right side
      Create(F,Out_File,NameR);
      Eta:=-(LuN);
      Xi:= (LsP);
      for I in 1..Q loop
         X:=Eta*Wx+Xi*Vx;
         U:=Eta*Wu+Xi*Vu;
         Y:=X+P1;
         U:=U+P2;

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;

         Txt_Put(F,Approx(Y1+CenX)+Half);
         Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);
         YY:=Y;
         UU:=U;

         for K in 1..N loop
            Map(Phi,S,Bs,Cs,Es,Lambda,Mu,Y,U);
         end loop;

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;

         Txt_Put(F,Approx(Y1+CenX)+Half);
         Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);


         Y:=YY;
         U:=-(UU+Scal(Half))-Scal(Half);
         for K in 1..N loop
            Map(Phi,S,Bs,Cs,Es,Lambda,Mu,Y,U);
         end loop;
         U:=-(U+Scal(Half))-Scal(Half);

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;

         Txt_Put(F,Approx(Y1+CenX)+Half);
         Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);
         Eta:=Eta+EDel;
      end loop;
      Close(F);


      --Left side
      Create(F,Out_File,NameL);
      Eta:=-(LuN);
      Xi:= -(LsN);
      for I in 1..Q loop
         X:=Eta*Wx+Xi*Vx;
         U:=Eta*Wu+Xi*Vu;
         Y:=X+P1;
         U:=U+P2;

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;
         YY:=Y;
         UU:=U;

         Txt_Put(F,Approx(Y1+CenX)+Half);
         Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);

         for K in 1..N loop
            Map(Phi,S,Bs,Cs,Es,Lambda,Mu,Y,U);
         end loop;

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;

         Txt_Put(F,Approx(Y1+CenX)+Half);
         Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);


         Y:=YY;
         U:=-(UU+Scal(Half))-Scal(Half);
         for K in 1..N loop
            Map(Phi,S,Bs,Cs,Es,Lambda,Mu,Y,U);
         end loop;
         U:=-(U+Scal(Half))-Scal(Half);

         if DoLin then
            Y1:=Val(Y-CenX,U-CenY,T1);
            U1:=Val(Y-CenX,U-CenY,T2);
         else
            Y1:=Y-CenX;
            U1:=U-CenY;
         end if;

         Txt_Put(F,Approx(Y1+CenX)+Half);
         Put(F,"  ");
         Txt_Put(F,Approx(U1+CenY)+Half);
         New_Line(F);
         Eta:=Eta+EDel;
      end loop;
      Close(F);
   end DrawHS;


   procedure Gaps(DoLin: in Boolean;NameP,NameCen,NameCenn: in String; Q,N: in Integer;K0bx,K0bu,K0tx,K0tu,K1bx,K1bu,K1tx,K1tu,Sc,Lip,B,C,E,Lambda,Mu,Px,Pu,Px0,Pu0,Px1,Pu1,CenX,CenY,
Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1,V1,V2:in Scalar; Phi,S,Sigma,T1,T2: in Taylor;

                                                 MinL0,MinL1,MaxL01: out Scalar) is --FINE

      use Reps.IO;
      use Ada.Text_IO;

      Xl,Xr,Ul,Ur: Rep;
      F,FF: File_Type;
      R1: Radius;
      Shift:Rep:=Rep(0.1);
      Fact: Scalar:=Scal(Rep(0.1));
      DCor: Matrix(1..2,1..2);
      Vs,Vu: Vector(1..2);
      XCen,UCen: Vector(1..2400);
      V11: Scalar;
      T11,T12,T21,T22: Taylor;
   begin
      Create(F,Out_File,NameP);
      Txt_Put(F,Approx(K0bx+CenX)+Half);Put(F,"  ");
      Txt_Put(F,Approx(K0bu+CenY)+Half);
      New_Line(F);
      Txt_Put(F,Approx(K0tx+CenX)+Half);Put(F,"  ");
      Txt_Put(F,Approx(K0tu+CenY)+Half);
      New_Line(F);
      Txt_Put(F,Approx(K1bx+CenX)+Half);Put(F,"  ");
      Txt_Put(F,Approx(K1bu+CenY)+Half);
      New_Line(F);
      Txt_Put(F,Approx(K1tx+CenX)+Half);Put(F,"  ");
      Txt_Put(F,Approx(K1tu+CenY)+Half);
      Close(F);
      R1:=Rep(0.95)*Rho(S);

      Open(F,In_File,NameCen);
      Create(FF,Out_File,NameCenn);
      for I in XCen'Range loop
         Txt_Get(F,Xl);
         Txt_Get(F,Xr);
         XCen(I):=Sc*Scal(Xl,Xr);
         Txt_Get(F,Ul);
         Txt_Get(F,Ur);
         UCen(I):=Sc*Scal(Ul,Ur);
         Txt_Put(FF,Approx(XCen(I)+CenX)+Half); --  +CenX));
         Put(FF,"  ");
         Txt_Put(FF,Approx(UCen(I)+CenY)+Half); --  +CenY));
         New_Line(FF);
      end loop;
      Close(FF);
      Close(F);

      T11:= Der1(R1,T1);
      T12:=Der2(R1,T1);
      T21:=Der1(R1,T2);
      T22:=Der2(R1,T2);

      DCor(1,1):=Val(Px-Cenx,Pu-CenY,T11);
      DCor(1,2):=Val(Px-Cenx,Pu-CenY,T12);
      DCor(2,1):=Val(Px-Cenx,Pu-CenY,T21);
      DCor(2,2):=Val(Px-Cenx,Pu-CenY,T22);

      Vs(1):=V1;
      Vs(2):=V2;
      Vu(1):=V1;
      Vu(2):=-V2;

      Mult(DCor,Vs);
      V11:=Vs(1);
      Vs(1):=Vs(1)/Sqrt(Sqr(Vs(1))+Sqr(Vs(2)));
      Vs(2):=Vs(2)/Sqrt(Sqr(V11)+Sqr(Vs(2)));

      Mult(DCor,Vu);
      V11:=Vu(1);
      Vu(1):=Vu(1)/Sqrt(Sqr(Vu(1))+Sqr(Vu(2)));
      Vu(2):=Vu(2)/Sqrt(Sqr(V11)+Sqr(Vu(2)));

      --      Vs(1):=Sc*Scal(0.546769451300831);
      --      Vs(2):=Sc*Scal(-0.837283206044519);

      --      Vu(1):=Sc*Scal(-0.999597186245105);
      --      Vu(2):=Sc*Scal(-0.028380719703166);


      --      New_Line;
      --      Put("'Linearized' eigenvectors at the fixed point:");
      --      New_Line;
      --      Put("Stable:");
      --      New_Line;
      --      Put("Vx=");Show(Vs(1));
      --      Put("Vu=");Show(Vs(2));
      --      Put("Unstable:");
      --      New_Line;
      --      Put("Vx=");Show(Vu(1));
      --      Put("Vu=");Show(Vu(2));

      MinL0:=Sqrt(Sqr(K0bx-K0tx)+Sqr(K0bu-K0tu));
      MinL1:=Sqrt(Sqr(K1bx-K1tx)+Sqr(K1bu-K1tu));
      MaxL01:=Sqrt(Sqr(K0bx-K1tx)+Sqr(K0bu-K1tu));

      Vs(1):=(K0bx-K1tx)/Sqrt(Sqr(K0bx-K1tx)+Sqr(K0bu-K1tu));
      Vs(2):=(K0bu-K1tu)/Sqrt(Sqr(K0bx-K1tx)+Sqr(K0bu-K1tu));

      --      New_Line;
      --      Put("'Stable' directions between the end points of the gap:");
      --      New_Line;
      --      Put("Vx=");Show(Vs(1));
      --      Put("Vu=");Show(Vs(2));

      MaxL01:=MaxL01*Sqrt(Scal(One)+Sqr(Lip));
   end Gaps;




   procedure Manifold(R: in Radius; X0,U0,X1,U1,X2,U2,X3,U3,CenX,CenY,Sc: in Scalar; G1,G2: out Taylor) is

      use Reps.IO;
      use Ada.Text_IO;
      Q: constant Integer:=4;
      Z: Matrix(1..Q,1..Q);
      U,X,Work: Vector(1..Q);
      Pvt: Pivot(1..Q);
      Cond: Scalar;
      R1: Radius:=Rep(0.2)*R;
   begin
      X(1):=X0;
      X(2):=X1;
      X(3):=X2;
      X(4):=X3;

      U(1):=U0;
      U(2):=U1;
      U(3):=U2;
      U(4):=U3;


      for I in 1..Q loop
         for J in 1..Q loop
            Z(I,J):=X(I)**(J-1);
         end loop;
      end loop;

      Decomp(Z,Cond,Pvt,Work);
      X:=U;
      Solve(Z,X,Pvt);

      SetZero(R,G1);
      SetZero(R,G2);
      G1:=G1+Monom(R,1,0,Scal(One));
      G2:=G2+Monom(R,0,1,Scal(One));
      for I in 1..Q loop
         G2:=G2-Monom(R,I-1,0,X(I));
      end loop;
      G1:=Sc*Comp(Monom(R1,1,0,Scal(One))+Monom(R1,0,0,CenX),Monom(R1,0,1,Scal(One))+Monom(R1,0,0,CenY),G1);
      G2:=Sc*Comp(Monom(R1,1,0,Scal(One))+Monom(R1,0,0,CenX),Monom(R1,0,1,Scal(One))+Monom(R1,0,0,CenY),G2);
   end Manifold;


   procedure Coordinate(Name01,Name02,Name11,Name12: in String;Q: in Integer; LsP0,LsN0,LuP0,LuN0,LsP1,LsN1,LuP1,LuN1: in Rep;
                                                                              X0,U0,X1,U1,CenX,CenY,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1: in Scalar; T1,T2,V1,V2: in Taylor; G1,G2: out Taylor) is

      use Reps.IO;
      use Ada.Text_IO;
      QQ: constant Integer:=2*Q*Q;
      Pvt: Pivot(1..QQ);
      Cond: Scalar;
      Work: Vector(1..QQ);
      K,L,M: Integer;
      EDel,XDel,Xi,Eta: Scalar;
      F1,F2: File_Type;
      R: Radius;
      X,U,Y,V,Co1,Co2: Vector(1..QQ);
      Z: Matrix(1..QQ,1..QQ);
   begin
      R:=Rho(T1);
      K:=1;

--      goto SKIP;

      XDel:=Rep(1.0)*(Scal(LsP0+LsN0))/Rep(Q);
      EDel:=Rep(1.0)*(Scal(LuP0+LuN0))/Rep(Q);

      Create(F1,Out_File,Name01);
      Create(F2,Out_File,Name02);

      Xi:=-Scal(LsN0);
      for I in 1..Q loop
         Eta:=-Scal(LuN0);
         for J in 1..Q loop
            X(K):=Eta*Wx0+Xi*Vx0+X0;
            U(K):=Eta*Wu0+Xi*Vu0+U0;

            Txt_Put(F1,Approx(X(K)));
            Put(F1,"  ");
            Txt_Put(F1,Approx(U(K)));
            New_Line(F1);

            Y(K):=Val(X(K)-X0,U(K)-U0,T1)+X0;
            V(K):=Val(X(K)-X0,U(K)-U0,T2)+U0;

            X(K):=X(K)-CenX;
            U(K):=U(K)-CenY;

            Txt_Put(F2,Approx(Y(K)));
            Put(F2,"  ");
            Txt_Put(F2,Approx(V(K)));
            New_Line(F2);

            Y(K):=Y(K)-CenX;
            V(K):=V(K)-CenY;


            K:=K+1;
            Eta:=Eta+EDel;
         end loop;
         Xi:=Xi+XDel;
      end loop;
      Close(F1);
      Close(F2);

--      <<SKIP>> null;
--      goto SKIP;

      XDel:=Rep(1.0)*(Scal(LsP1+LsN1))/Rep(Q);
      EDel:=Rep(1.0)*(Scal(LuP1+LuN1))/Rep(Q);

      Create(F1,Out_File,Name11);
      Create(F2,Out_File,Name12);

      Xi:=-Scal(LsN1);
      for I in 1..Q loop
         Eta:=-Scal(LuN1);
         for J in 1..Q loop
            X(K):=Eta*Wx1+Xi*Vx1+X1;
            U(K):=Eta*Wu1+Xi*Vu1+U1;

            Txt_Put(F1,Approx(X(K)));
            Put(F1,"  ");
            Txt_Put(F1,Approx(U(K)));
            New_Line(F1);

            Y(K):=Val(X(K)-X1,U(K)-U1,V1)+X1;
            V(K):=Val(X(K)-X1,U(K)-U1,V2)+U1;

            X(K):=X(K)-CenX;
            U(K):=U(K)-CenY;

            Txt_Put(F2,Approx(Y(K)));
            Put(F2,"  ");
            Txt_Put(F2,Approx(V(K)));
            New_Line(F2);

            Y(K):=Y(K)-CenX;
            V(K):=V(K)-CenY;

            K:=K+1;
            Eta:=Eta+EDel;
         end loop;
         Xi:=Xi+XDel;
      end loop;
      Close(F1);
      Close(F2);

--      <<SKIP>> null;

      for J in 1..QQ loop
         L:=0;
         K:=0;
         while (L<QQ and K<PDeg) loop
            M:=0;
            while (L<QQ and M<PDeg) loop
               if (K+M)>0 then
                  L:=L+1;
                  Z(J,L):=(X(J)**K)*(U(J)**M);
--                  Show(Scal(J),False);Show(Scal(L),False);Show(Z(J,L));
               end if;
               M:=M+1;
            end loop;
            K:=K+1;
         end loop;
      end loop;


      New_Line;
      Put("K=");Show(Scal(K));
      Put("M=");Show(Scal(M));


      Decomp(Z,Cond,Pvt,Work);
      Co1:=Y;
      Co2:=V;
      Solve(Z,Co1,Pvt);
      Solve(Z,Co2,Pvt);

--      for I in 1..10 loop
--         DCo1:=Y-Z*Co1;
--         DCo2:=V-Z*Co2;
--         Solve(Z,DCo1,Pvt);
--         Solve(Z,DCo2,Pvt);
--         Co1:=Co1+DCo1;
--         Co2:=Co2+DCo2;
--      end loop;

      SetZero(R,G1);
      SetZero(R,G2);


      L:=0;
      K:=0;
      while (L<QQ and K<PDeg) loop
         M:=0;
         while (L<QQ and M<PDeg) loop
            if (K+M>0) then
               L:=L+1;
               G1:=G1+Monom(R,K,M,Co1(L));
               G2:=G2+Monom(R,K,M,Co2(L));
            end if;
            M:=M+1;
         end loop;
         K:=K+1;
      end loop;

      Co1:=Z*Co1;

      for I in 1..QQ loop
         Show( Y(I));
      end loop;

   end Coordinate;




    procedure DrawManifolds(NameU,NameS,NameU2,NameS2: in String; K,Q,N: in Integer;
B,C,E,Lambda,Mu,Sx,Su,Rad: in Scalar; Phi,Sigma: in Taylor) is

      use Reps.IO;
      use Ada.Text_IO;

      X,U: Scalar;
      F1,F2,F3,F4: File_Type;
   begin


      Create(F1,Out_File,NameU);
      Create(F2,Out_File,NameS);
      Create(F3,Out_File,NameU2);
      Create(F4,Out_File,NameS2);
      for J in 1..K loop
         X:=Sx+Cos(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
         U:=Su+Sin(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;

         for II in 1..Q loop
            for I in 1..N loop
                 Map(Phi,Sigma,B,C,E,Lambda,Mu,X,U);
            end loop;
         end loop;
         ---Unstable
         Txt_Put(F1,Approx(X));
         Put(F1,"  ");
         Txt_Put(F1,Approx(U));
         New_Line(F1);
         --Stable
         U:=-(U+Scal(Half))-Scal(Half);
         Txt_Put(F2,Approx(X));
         Put(F2,"  ");
         Txt_Put(F2,Approx(U));
         New_Line(F2);

         --Stable
         X:=Lambda*(X+Scal(Half))-Scal(Half);
         U:=Mu*(U+Scal(Half))-Scal(Half);

         Txt_Put(F4,Approx(X));
         Put(F4,"  ");
         Txt_Put(F4,Approx(U));
         New_Line(F4);

         --Unstable
         U:=-(U+Scal(Half))-Scal(Half);
         Txt_Put(F3,Approx(X));
         Put(F3,"  ");
         Txt_Put(F3,Approx(U));
         New_Line(F3);

      end loop;
      Close(F1);
      Close(F2);
      Close(F3);
      Close(F4);

   end DrawManifolds;



   procedure Linearize1(R1: in Radius; Phi,S: in Taylor;P,Lambda,Mu,IT11,IT12,IT21,IT22,Ev1,Ev2: in Scalar; Y,U: in out Scalar) is
      use Reps.IO;
      use Ada.Text_IO;
      A,B,C,D,CB,Dett,Det,T11,T12,T21,T22,Y1,U1,Un,Yn,YDel,UDel,Yn1,Un1,Ynn: Scalar;
      Tay1,Tay2: Taylor;
      N: Integer:=3;
      K: Integer;
      Err:  Rep;
      Eps: constant Rep:=(1.0E-8);
      IL,Inv: Matrix(1..2,1..2);
   begin

      Tay1:=Der1(R1,S);
      Tay2:=Der2(R1,S);
      CB:=Sqr(Val(P,P,Tay2)/Val(P,P,Tay1))-ScalOne;
      B:=-Scal(One)/Val(P,P,Tay1);
      C:=Val(P,P,Tay1)-Sqr(Val(P,P,Tay2))/Val(P,P,Tay1);
      A:=-(Val(P,P,Tay2)/Val(P,P,Tay1));
      D:=A;

      Dett:=A*D-CB;
      IL(1,1):=D/Dett;IL(1,2):=-B/Dett;
      IL(2,1):=-C/Dett;IL(2,2):=A/Dett;


      Det:=IT11*IT22-IT12*IT21;
      T11:=IT22/Det;T12:=-IT12/Det;
      T21:=-IT21/Det;T22:=IT11/Det;

      U1:=U;
      Y1:=Y;
      Yn:=Y1;
      Un:=U1;

      Inv:=IL;
      for I in 2..N loop
         Mult(IL,Inv);
      end loop;

      K:=0;
      Err:=Rep(1.0E+20);
      while Err>Eps loop
         Yn1:=Yn;
         Un1:=Un;

         Un:=Un-Scal(Half);
         Yn:=Yn+P;

         for I in 1..N loop
            Map(Phi,S,Lambda,Mu,Yn,Un);
         end loop;
         Un:=Un+Scal(Half);
         Yn:=Yn-P;

         Ynn:=Yn;
         Yn:=Inv(1,1)*Yn+Inv(1,2)*Un;
         Un:=Inv(2,1)*Ynn+Inv(2,2)*Un;

         YDel:=Y1-Yn;
         Yn:=Yn1+YDel;
         UDel:=U1-Un;
         Un:=Un1+UDel;

         K:=K+1;
         if K>500 then
            Message("Doubling.Linearize",Giving_Up);
         end if;
         Err:=Sup(Max(Abs(UDel),Abs(YDel)));
         New_Line;
         Put("K=");Show(Scal(K),False);
         Put("Eps=");Show(Scal(Err));
      end loop;
      U:=Un;
      Y:=Yn;
      --error terms

   end Linearize1;



     procedure Linearize2(N: in Integer; R1: in Radius; Phi,S: in Taylor;P,Lambda,Mu,A,B,C,D: in Scalar; Y,U: in out Scalar) is
      use Reps.IO;
      use Ada.Text_IO;
      Y1: Scalar;
      Eps: constant Rep:=(1.0E-8);
      L,Inv: Matrix(1..2,1..2);
   begin

      L(1,1):=A;
      L(1,2):=B;
      L(2,1):=C;
      L(2,2):=D;

--      Det:=A*D-B*C;
--      L(1,1):=D/Det;
--      L(1,2):=-B/Det;
--      L(2,1):=-C/Det;
--      L(2,2):=A/Det;

      Inv(1,1):=Scal(One);
      Inv(1,2):=Scal(Zero);
      Inv(2,1):=Scal(Zero);
      Inv(2,2):=Scal(One);

      for I in 1..N loop
         Mult(L,Inv);
      end loop;

--      Y1:=Y;
--      Y:=Inv(1,1)*Y+Inv(1,2)*U;
--      U:=Inv(2,1)*Y1+Inv(2,2)*U;


      U:=U-Scal(Half);
      Y:=Y+P;

      U:=-(U+Scal(Half))-Scal(Half);
      for I in 1..N loop
         Map(Phi,S,Lambda,Mu,Y,U);
      end loop;
      U:=-(U+Scal(Half))-Scal(Half);

      U:=U+Scal(Half);
      Y:=Y-P;

      Y1:=Y;
      Y:=Inv(1,1)*Y+Inv(1,2)*U;
      U:=Inv(2,1)*Y1+Inv(2,2)*U;

      --error terms
   end Linearize2;


   procedure Linearize3(N: in Integer; R1: in Radius; Phi,S: in Taylor;P,Lambda,Mu,A,B,C,D: in Scalar; Y,U: in out Scalar) is
      use Reps.IO;
      use Ada.Text_IO;
      Y1,U1,Un,Yn,YDel,UDel,Delt,Det: Scalar;
      Eps: constant Rep:=(1.0E-10);
      De,ID: Matrix(1..2,1..2);
      Err: Rep;
   begin
      Delt:=Scal(Rep(0.0001));
      Err:=Rep(1.0E+20);
      while Err>Eps loop

         --Derivative
         Y1:=Y;
         U1:=U;

         Linearize2(N,R1,Phi,S,P,Lambda,Mu,A,B,C,D,Y1,U1);
         Yn:=Y+Delt;
         Un:=U;
         Linearize2(N,R1,Phi,S,P,Lambda,Mu,A,B,C,D,Yn,Un);
         De(1,1):=Scal(One)-(Yn-Y1)/Delt;
         De(2,1):=-(Un-U1)/Delt;
         Yn:=Y;
         Un:=U+Delt;
         Linearize2(N,R1,Phi,S,P,Lambda,Mu,A,B,C,D,Yn,Un);
         De(1,2):=-(Yn-Y1)/Delt;
         De(2,2):=Scal(One)-(Un-U1)/Delt;

         Det:=De(1,1)*De(2,2)-De(1,2)*De(2,1);
         ID(1,1):=De(2,2)/Det;
         ID(1,2):=-De(1,2)/Det;
         ID(2,1):=-De(2,1)/Det;
         ID(2,2):=De(1,1)/Det;


         Y1:=Y;
         U1:=U;
         Linearize2(N,R1,Phi,S,P,Lambda,Mu,A,B,C,D,Y,U);
         YDel:=ID(1,1)*(Y-Y1)+ ID(1,2)*(U-U1);
         UDel:=ID(2,1)*(Y-Y1)+ ID(2,2)*(U-U1);
         Y:=Y1+YDel;
         U:=U1+UDel;
         Err:=Sup(Max(abs(YDel),abs(UDel)));
      end loop;
      -- Y:=Yn;
      -- U:=Un;
      --error terms
   end Linearize3;


   procedure Domain(Name1,Name2,Name3,Name4: in String; N: in Integer; R: in Radius; Phi,S: in Taylor; Lambda, Mu: in Scalar) is
      use Reps.IO;
      use Ada.Text_IO;
      UMinn,UMaxx,UMax,UMin,U,Y,X,YDel,XDel,UDel: Scalar;
      F1,F2,F3,F4: File_Type;
      K: Integer;
   begin
      YDel:=Two*Scal(R)/Scal(N);
      XDel:=Two*Scal(R)/Scal(N);

      Create(F1,Out_File,Name1);
      Create(F2,Out_File,Name2);
      Create(F3,Out_File,Name3);
      Create(F4,Out_File,Name4);

      X:=Scal(-R);
      K:=1;
      while Sup(X)<R loop
         Y:=Scal(-R);
         UMax:=-Scal(Rep(1.0E+20));
         UMin:=Scal(Rep(1.0E+20));
         while Sup(Y)<R loop
            U:=-Val(Y,X,S);
            if U>UMax then UMax:=U; end if;
            if U<UMin then UMin:=U; end if;
            Y:=Y+Ydel;
         end loop;
         UMaxx:=UMax;
         UMinn:=UMin;


         if K=1 then
            UDel:=(UMaxx-UMinn)/Scal(N);
            U:=Uminn;
            while U<=UMaxx loop
               Txt_Put(F2,Approx(X)+Half );
               Put(F2,"  ");
               Txt_Put(F2,Inf(U));
               New_Line(F2);
               U:=U+UDel;
            end loop;
         end if;


         if K=N then
            UDel:=(UMaxx-UMinn)/Scal(N);
            U:=Uminn;
            while U<=UMaxx loop
               Txt_Put(F4,Approx(X)+Half );
               Put(F4,"  ");
               Txt_Put(F4,Inf(U));
               New_Line(F4);
               U:=U+UDel;
            end loop;
         end if;


         Txt_Put(F1,Approx(X)+Half );
         Put(F1,"  ");
         Txt_Put(F1,Inf(UMax));
         New_Line(F1);

         Txt_Put(F3,Approx(X)+Half );
         Put(F3,"  ");
         Txt_Put(F3,Sup(UMin));
         New_Line(F3);

         X:=X+Xdel;
         K:=K+1;
      end loop;
      Close(F1);
      Close(F2);
      Close(F3);
      Close(F4);
   end Domain;


   procedure IterateDomain(Name1,Name2,Name3,Name4: in String; N,M: in Integer; R: in Radius; Phi,S: in Taylor; Lambda, Mu: in Scalar) is
      use Reps.IO;
      use Ada.Text_IO;
      XX,UMinn,UMaxx,UMax,UMin,U,Y,X,YDel,XDel,UDel,UU: Scalar;
      F1,F2,F3,F4: File_Type;
      K: Integer;
      X1,X2,X3,X4,U1,U2,U3,U4: Vector(1..N);
   begin
      YDel:=Two*Scal(R)/Scal(N);
      XDel:=Two*Scal(R)/Scal(N);

      Create(F1,Out_File,Name1);
      Create(F2,Out_File,Name2);
      Create(F3,Out_File,Name3);
      Create(F4,Out_File,Name4);

      X:=Scal(-R);
      K:=1;
      while Sup(X)<R loop
         Y:=Scal(-R);
         UMax:=-Scal(Rep(1.0E+20));
         UMin:=Scal(Rep(1.0E+20));
         while Sup(Y)<R loop
            U:=-Val(Y,X,S);
            if U>UMax then UMax:=U; end if;
            if U<UMin then UMin:=U; end if;
            Y:=Y+Ydel;
         end loop;
         UMaxx:=UMax;
         UMinn:=UMin;


         if K=1 then
            UDel:=(UMaxx-UMinn)/Scal(N);
            U:=Uminn;
            while U<=UMaxx loop
               UU:=-U;
               if UU>UMaxx then
                  UU:=UMaxx;
               elsif UU<UMinn then
                  UU:=UMinn;
               end if;

               UU:=UU-Scal(Half);
               XX:=X;
               Map(Phi,S,Lambda,Mu,XX,UU);
               XX:=(XX+Scal(Half)/Lambda);
               UU:=(-(UU+Scal(Half)))/Mu;
               X2(K):=XX;
               U2(K):=UU;

               Txt_Put(F2,Approx(XX) );
               Put(F2,"  ");
               Txt_Put(F2,Inf(UU));
               New_Line(F2);
               U:=U+UDel;
            end loop;
         end if;


         if K=N then
            UDel:=(UMaxx-UMinn)/Scal(N);
            U:=Uminn;
            while U<=UMaxx loop
               UU:=-U;
               if UU>UMaxx then
                  UU:=UMaxx;
               elsif UU<UMinn then
                  UU:=UMinn;
               end if;
               UU:=UU-Scal(Half);
               XX:=X;
               Map(Phi,S,Lambda,Mu,XX,UU);
               XX:=(XX+Scal(Half)/Lambda);
               UU:=(-(UU+Scal(Half)))/Mu;
               X4(K):=XX;
               U4(K):=UU;

               Txt_Put(F4,Approx(XX) );
               Put(F4,"  ");
               Txt_Put(F4,Inf(UU));
               New_Line(F4);
               U:=U+UDel;
            end loop;
         end if;



         UMax:=-UMax;
         if UMax<UMinn then
            UMax:=UMinn;
         elsif UMax>UMaxx then
            UMax:=UMaxx;
         end if;

         UMax:=UMax-Scal(Half);
         XX:=X;
         Map(Phi,S,Lambda,Mu,XX,UMax);
         XX:=(XX+Scal(Half)/Lambda);
         UMax:=(-(UMax+Scal(Half)))/Mu;
         X1(K):=XX;
         U1(K):=UMax;

         Txt_Put(F1,Approx(XX) );
         Put(F1,"  ");
         Txt_Put(F1,Inf(UMax));
         New_Line(F1);

         UMin:=-UMin;
         if UMin<UMinn then
            UMin:=UMinn;
         elsif UMin>UMaxx then
            UMin:=UMaxx;
         end if;
         UMin:=UMin-Scal(Half);
         XX:=X;
         Map(Phi,S,Lambda,Mu,XX,UMin);
         XX:=(XX+Scal(Half)/Lambda);
         UMin:=(-(UMin+Scal(Half)))/Mu;
         X3(K):=XX;
         U3(K):=UMin;

         Txt_Put(F3,Approx(XX) );
         Put(F3,"  ");
         Txt_Put(F3,Sup(UMin));
         New_Line(F3);

         X:=X+Xdel;
         K:=K+1;
      end loop;


--      for I in 1..M loop
--         for J in 1..N loop
--            UU:=U2(J);
--            XX:=X2(J);
--            UU:=-UU;
--            XX:=XX;

--          for JJ in 1..N loop
--             X:=X1(JJ);
--             U:=U1(JJ);
--             if X=XX then
--               if U>
--          end loop;


--            Map(Phi,S,Lambda,Mu,XX,UU);
--            XX:=(XX+Scal(Half)/Lambda);
--            UU:=(-(UU+Scal(Half)))/Mu;
--            X2(K):=XX;
--            U2(K):=UU;

--            Txt_Put(F2,Approx(XX) );
--            Put(F2,"  ");
--            Txt_Put(F2,Inf(UU));
--            New_Line(F2);
--            U:=U+UDel;
--         end loop;
--      end loop;


      Close(F1);
      Close(F2);
      Close(F3);
      Close(F4);
   end IterateDomain;


   procedure DrawInclusion0(Name1,Name2: in String; K: in Integer; Phi,S: in Taylor; Lambda,Mu,LsP0,LsN0,LuP0,LuN0,LsP1,LsN1,LuP1,LuN1,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1,CenX,CenY,Sx0,Su0,Sx1,Su1,Sx,Su,Rad,Ecc: in Scalar) is
      use Reps.IO;
      -- use Reps_Ops;
      use Ada.Text_IO;
      X,U,X0,U0,X1,U1,Xi0,Xi1,Eta0,Eta1,XDel0,XDel1,EDel0,EDel1:  Scalar;
      F,FF: File_Type;
      B: Scalar;

      function Signum(S: Scalar) return Scalar is
      begin
         if Inf(S)>Zero then
            return Scal(One);
         elsif Sup(S)<Zero then
            return Scal(-One);
         else
            return Scal(Zero);
         end if;

      end  Signum;

   begin
      B:=Sqrt(Scal(One)-Sqr(Ecc));

      Create(F,Out_File,Name1);
      for J in 1..K loop
         X:=Sx+Cos(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
         U:=Su+B*Sin(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
         Txt_Put(F,Approx(X +Scal(Half)));
         Put(F,"  ");
         Txt_Put(F,Approx(U+Scal(Half)));
         New_Line(F);
      end loop;
      Close(F);

      Create(FF,Out_File,Name2);

      XDel0:=(LsP0+LsN0)/Scal(K);
      EDel0:=(LuP0+LuN0)/Scal(K);
      XDel1:=(LsP1+LsN1)/Scal(K);
      EDel1:=(LuP1+LuN1)/Scal(K);

      Xi0:=-(LsN0);
      Eta0:=-(LuN0);
      Xi1:=-(LsN1);
      Eta1:=-(LuN1);
      for J in 0..K-1 loop
         X0:=Eta0*Wx0+Xi0*Vx0+Sx0;
         U0:=Eta0*Wu0+Xi0*Vu0+Su0;
         X1:=Eta1*Wx1+Xi1*Vx1+Sx1;
         U1:=Eta1*Wu1+Xi1*Vu1+Su1;

         Txt_Put(FF,Approx(Lambda*(X0+Scal(Half)) ));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Mu*(U0+Scal(Half)) ));
         New_Line(FF);
         Txt_Put(FF,Approx(Lambda*(X1+Scal(Half)) ));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Mu*(U1+Scal(Half)) ));
         New_Line(FF);

         Eta0:=Scal(Sup(Eta0),Sup(Scal(Sup(EDel0))+Scal(Sup(Eta0))));
         Eta1:=Scal(Sup(Eta1),Sup(Scal(Sup(EDel1))+Scal(Sup(Eta1))));
      end loop;


      Xi0:=-(LsN0);
      Eta0:=(LuP0);
      Xi1:=-(LsN1);
      Eta1:=(LuP1);
      for J in K-1..2*K-1 loop
         X0:=Eta0*Wx0+Xi0*Vx0+Sx0;
         U0:=Eta0*Wu0+Xi0*Vu0+Su0;
         X1:=Eta1*Wx1+Xi1*Vx1+Sx1;
         U1:=Eta1*Wu1+Xi1*Vu1+Su1;

         Txt_Put(FF,Approx(Lambda*(X0+Scal(Half)) ));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Mu*(U0+Scal(Half)) ));
         New_Line(FF);
         Txt_Put(FF,Approx(Lambda*(X1+Scal(Half)) ));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Mu*(U1+Scal(Half))));
         New_Line(FF);

         Xi0:=Scal(Sup(Xi0),Sup(Scal(Sup(XDel0))+Scal(Sup(Xi0))));
         Xi1:=Scal(Sup(Xi1),Sup(Scal(Sup(XDel1))+Scal(Sup(Xi1))));
      end loop;


      Xi0:=(LsP0);
      Eta0:=(LuP0);
      Xi1:=(LsP1);
      Eta1:=(LuP1);
      for J in 2*K-1..3*K-1 loop
         X0:=Eta0*Wx0+Xi0*Vx0+Sx0;
         U0:=Eta0*Wu0+Xi0*Vu0+Su0;
         X1:=Eta1*Wx1+Xi1*Vx1+Sx1;
         U1:=Eta1*Wu1+Xi1*Vu1+Su1;

         Txt_Put(FF,Approx(Lambda*(X0+Scal(Half)) ));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Mu*(U0+Scal(Half)) ));
         New_Line(FF);
         Txt_Put(FF,Approx(Lambda*(X1+Scal(Half)) ));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Mu*(U1+Scal(Half)) ));
         New_Line(FF);

         Eta0:=Scal(Inf(-Scal(Sup(EDel0))+Scal(Inf(Eta0))),Inf(Eta0));
         Eta1:=Scal(Inf(-Scal(Sup(EDel1))+Scal(Inf(Eta1))), Inf(Eta1));
      end loop;


      Xi0:=(LsP0);
      Eta0:=-(LuN0);
      Xi1:=(LsP1);
      Eta1:=-(LuN1);
      for J in 3*K-1..4*K-1 loop
         X0:=Eta0*Wx0+Xi0*Vx0+Sx0;
         U0:=Eta0*Wu0+Xi0*Vu0+Su0;
         X1:=Eta1*Wx1+Xi1*Vx1+Sx1;
         U1:=Eta1*Wu1+Xi1*Vu1+Su1;

         Txt_Put(FF,Approx(Lambda*(X0+Scal(Half))));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Mu*(U0+Scal(Half)) ));
         New_Line(FF);
         Txt_Put(FF,Approx(Lambda*(X1+Scal(Half)) ));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Mu*(U1+Scal(Half)) ));
         New_Line(FF);

         Xi0:=Scal(Inf(-Scal(Sup(XDel0))+Scal(Inf(Xi0))),Inf(Xi0));
         Xi1:=Scal(Inf(-Scal(Sup(XDel1))+Scal(Inf(Xi1))),Inf(Xi1));
      end loop;
      Close(FF);

   end DrawInclusion0;



   procedure DrawInclusion1(Name1,Name2: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu,CenX,CenY,S1,S2,Rad,Ecc: in Scalar) is
      use Reps.IO;
      -- use Reps_Ops;
      use Ada.Text_IO;
      --      K: constant Integer:=600;
      X,U:  Vector(1..K);
      F,FF: File_Type;
      B1: Scalar;

      function Signum(S: Scalar) return Scalar is
      begin
         if Inf(S)>Zero then
            return Scal(One);
         elsif Sup(S)<Zero then
            return Scal(-One);
         else
            return Scal(Zero);
         end if;

      end  Signum;

   begin
      B1:=Sqrt(Scal(One)-Sqr(Ecc));


      Create(F,Out_File,Name1);
      Create(FF,Out_File,Name2);
      for J in X'Range loop
         X(J):=S1+Cos(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
         U(J):=S2+B1*Sin(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
         Txt_Put(F,Approx(X(J)+Scal(Half)));
         Put(F,"  ");
         Txt_Put(F,Approx(U(J)+Scal(Half)));
         New_Line(F);
      end loop;

      for J in X'Range loop
         X(J):=Center(X(J));
         U(J):=Center(U(J));
         Map(Phi,S,B,C,E,Lambda,Mu,X(J),U(J));
         Txt_Put(FF,Approx(Sqr(Lambda)*(X(J)+Scal(Half))));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Sqr(Mu)*(U(J)+Scal(Half))));
         New_Line(FF);
         Txt_Put(FF,Approx(Sqr(Lambda)*(X(J)+Scal(Half))));
         Put(FF,"  ");
         Txt_Put(FF,Approx(-Sqr(Mu)*(U(J)+Scal(Half))));
         New_Line(FF);
      end loop;
      Close(F);
      Close(FF);
   end DrawInclusion1;


   procedure DrawInclusion2(Name1,Name2: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu,CenX,CenY,S1,S2,Rad,Ecc: in Scalar) is
      use Reps.IO;
      use Ada.Text_IO;
      X,U:  Vector(1..K);
      F,FF: File_Type;
      B1: Scalar;

      function Signum(S: Scalar) return Scalar is
      begin
         if Inf(S)>Zero then
            return Scal(One);
         elsif Sup(S)<Zero then
            return Scal(-One);
         else
            return Scal(Zero);
         end if;

      end  Signum;


   begin
      B1:=Sqrt(Scal(One)-Sqr(Ecc));


      Create(F,Out_File,Name1);
      Create(FF,Out_File,Name2);
      for J in X'Range loop
         X(J):=S1+Cos(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
         U(J):=S2+B1*Sin(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
      end loop;

      for I in 1..2 loop
         for J in X'Range loop
            for II in 1..N loop
               Map(Phi,S,B,C,E,Lambda,Mu,X(J),U(J));
            end loop;
            if I=2 then
               Txt_Put(F,Approx(X(J)+Scal(Half)));
               Put(F,"  ");
               Txt_Put(F,Approx(U(J)+Scal(Half)));
               New_Line(F);
            end if;

            if I=1 then
               Txt_Put(FF,Approx(Lambda*(X(J)+Scal(Half))));
               Put(FF,"  ");
               Txt_Put(FF,Approx(Mu*(U(J)+Scal(Half))));
               New_Line(FF);

               Txt_Put(FF,Approx(Lambda*(X(J)+Scal(Half))));
               Put(FF,"  ");
               Txt_Put(FF,Approx(-Mu*(U(J)+Scal(Half))));
               New_Line(FF);
            end if;

         end loop;
      end loop;
      Close(F);
      Close(FF);

   end DrawInclusion2;



   procedure DrawInclusion3(Name1,Name2: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar; CenX,CenY,S1,S2,Rad,Ecc: in Scalar) is
      use Reps.IO;
      use Ada.Text_IO;
      X,U:  Vector(1..K);
      F,FF: File_Type;
      B1: Scalar;

      function Signum(S: Scalar) return Scalar is
      begin
         if Inf(S)>Zero then
            return Scal(One);
         elsif Sup(S)<Zero then
            return Scal(-One);
         else
            return Scal(Zero);
         end if;

      end  Signum;


   begin
      B1:=Sqrt(Scal(One)-Sqr(Ecc));


      Create(F,Out_File,Name1);
      Create(FF,Out_File,Name2);
      for J in X'Range loop
         X(J):=S1+Cos(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
         U(J):=S2+B1*Sin(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
         Txt_Put(F,Approx(X(J)+Scal(Half)));
         Put(F,"  ");
         Txt_Put(F,Approx(U(J)+Scal(Half)));
         New_Line(F);

         Txt_Put(FF,Approx(Lambda*(X(J)+Scal(Half))));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Mu*(U(J)+Scal(Half))));
         New_Line(FF);
      end loop;
      Close(F);
      Close(FF);

   end DrawInclusion3;


   procedure DrawInclusion4(Name0,Name1,Name2: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar; CenX,CenY,S1,S2,Rad,Ecc: in Scalar) is
      use Reps.IO;
      use Ada.Text_IO;
      X,U,X1,U1,X2,U2:  Vector(1..K);
      F0,F,FF: File_Type;
      B1: Scalar;

      function Signum(S: Scalar) return Scalar is
      begin
         if Inf(S)>Zero then
            return Scal(One);
         elsif Sup(S)<Zero then
            return Scal(-One);
         else
            return Scal(Zero);
         end if;

      end  Signum;


   begin
      B1:=Sqrt(Scal(One)-Sqr(Ecc));

      Create(F0,Out_File,Name0);
      Create(F,Out_File,Name1);
      Create(FF,Out_File,Name2);
      for J in X'Range loop
         X(J):=S1+Cos(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
         U(J):=S2+B1*Sin(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
      end loop;

      for J in X'Range loop
         for II in 1..N loop
            Map(Phi,S,B,C,E,Lambda,Mu,X(J),U(J));
         end loop;

         Txt_Put(F,Approx(X(J)+Scal(Half)));
         Put(F,"  ");
         Txt_Put(F,Approx(-(U(J)+Scal(Half))));
         New_Line(F);

         Txt_Put(F0,Approx(X(J)+Scal(Half)));
         Put(F0,"  ");
         Txt_Put(F0,Approx(U(J)+Scal(Half)));
         New_Line(F0);

         X1(J):=Center(Lambda*(X(J)+Scal(Half))-Scal(Half));
         U1(J):=Center(Mu*(U(J)+Scal(Half))-Scal(Half));
         X2(J):=Center(Lambda*(X(J)+Scal(Half))-Scal(Half));
         U2(J):=Center(-Mu*(U(J)+Scal(Half))-Scal(Half));
         for II in 1..N loop
            Map(Phi,S,B,C,E,Lambda,Mu,X1(J),U1(J));
            Map(Phi,S,B,C,E,Lambda,Mu,X2(J),U2(J));
         end loop;

         Txt_Put(FF,Approx((X1(J)+Scal(Half))));
         Put(FF,"  ");
         Txt_Put(FF,Approx((U1(J)+Scal(Half))));
         New_Line(FF);

         Txt_Put(FF,Approx((X2(J)+Scal(Half))));
         Put(FF,"  ");
         Txt_Put(FF,Approx((U2(J)+Scal(Half))));
         New_Line(FF);


      end loop;
      Close(FF);
      Close(F);
      Close(F0);
   end DrawInclusion4;



   procedure DrawInclusion5(Name1,Name2: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar; CenX,CenY,S1,S2,Rad,Ecc: in Scalar) is
      use Reps.IO;
      use Ada.Text_IO;
      X,U,X1,U1,X2,U2:  Vector(1..K);
      F,FF: File_Type;
      B1: Scalar;

   begin
      B1:=Sqrt(Scal(One)-Sqr(Ecc));

      Create(F,Out_File,Name1);
      Create(FF,Out_File,Name2);
      for J in X'Range loop
         X(J):=S1+Cos(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
         U(J):=S2+B1*Sin(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad;
      end loop;

      for J in X'Range loop
         for II in 1..N loop
            Map(Phi,S,B,C,E,Lambda,Mu,X(J),U(J));
         end loop;

         X1(J):=Center(Lambda*(X(J)+Scal(Half))-Scal(Half));
         U1(J):=Center(Mu*(U(J)+Scal(Half))-Scal(Half));
         X2(J):=Center(Lambda*(X(J)+Scal(Half))-Scal(Half));
         U2(J):=Center(-Mu*(U(J)+Scal(Half))-Scal(Half));

         for II in 1..N loop
            Map(Phi,S,B,C,E,Lambda,Mu,X1(J),U1(J));
            Map(Phi,S,B,C,E,Lambda,Mu,X2(J),U2(J));
         end loop;

         X1(J):=Center(Lambda*(X1(J)+Scal(Half))-Scal(Half));
         U1(J):=Center(Mu*(U1(J)+Scal(Half))-Scal(Half));
         X2(J):=Center(Lambda*(X2(J)+Scal(Half))-Scal(Half));
         U2(J):=Center(Mu*(U2(J)+Scal(Half))-Scal(Half));

         Txt_Put(F,Approx(X1(J)));
         Put(F,"  ");
         Txt_Put(F,Approx(U1(J)));
         New_Line(F);
         Txt_Put(F,Approx(X2(J)));
         Put(F,"  ");
         Txt_Put(F,Approx(U2(J)));
         New_Line(F);

         for II in 1..N loop
            Map(Phi,S,B,C,E,Lambda,Mu,X1(J),U1(J));
            Map(Phi,S,B,C,E,Lambda,Mu,X2(J),U2(J));
         end loop;

         X1(J):=Center(Lambda*(X1(J)+Scal(Half))-Scal(Half));
         U1(J):=Center(Mu*(U1(J)+Scal(Half))-Scal(Half));
         X2(J):=Center(Lambda*(X2(J)+Scal(Half))-Scal(Half));
         U2(J):=Center(Mu*(U2(J)+Scal(Half))-Scal(Half));


         Txt_Put(FF,Approx(X1(J)));
         Put(FF,"  ");
         Txt_Put(FF,Approx(U1(J)));
         New_Line(FF);
         Txt_Put(FF,Approx(X2(J)));
         Put(FF,"  ");
         Txt_Put(FF,Approx(U2(J)));
         New_Line(FF);
      end loop;
      Close(F);
      Close(FF);
   end DrawInclusion5;




   procedure DrawInclusion6(Name1,Name2,Name3: in String; K,N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu:
in Scalar; CenX,CenY,Sx1,Su1,Sx2,Su2,Rad1,Ecc1,Rad2,Ecc2: in Scalar) is
      use Reps.IO;
      use Ada.Text_IO;
      X,U,XX,UU,XXX,UUU:  Vector(1..K);
      F,FF,FFF: File_Type;
      B1,B2: Scalar;

   begin
      B1:=Sqrt(Scal(One)-Sqr(Ecc1));
      B2:=Sqrt(Scal(One)-Sqr(Ecc2));


      Create(F,Out_File,Name1);
      Create(FF,Out_File,Name2);
      Create(FFF,Out_File,Name3);
      for J in X'Range loop
         X(J):=Sx1+Cos(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad1;
         U(J):=Su1+B1*Sin(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad1;
         XXX(J):=Sx2+Cos(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad2;
         UUU(J):=Su2+B2*Sin(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad2;

         Txt_Put(FFF,Approx(XXX(J)));
         Put(FFF,"  ");
         Txt_Put(FFF,Approx(UUU(J)));
         New_Line(FFF);

         for II in 1..N loop
            Map(Phi,S,B,C,E,Lambda,Mu,X(J),U(J));
         end loop;

         XX(J):=X(J);
         UU(J):=-(U(J)+Scal(Half))-Scal(Half);

         X(J):=Center(Sqr(Lambda)*(X(J)+Scal(Half))-Scal(Half));
         U(J):=Center(Sqr(Mu)*(U(J)+Scal(Half))-Scal(Half));

         XX(J):=Center(Sqr(Lambda)*(XX(J)+Scal(Half))-Scal(Half));
         UU(J):=Center(Sqr(Mu)*(UU(J)+Scal(Half))-Scal(Half));

         Txt_Put(F,Approx(X(J)));
         Put(F,"  ");
         Txt_Put(F,Approx(U(J)));
         New_Line(F);

         Txt_Put(F,Approx(XX(J)));
         Put(F,"  ");
         Txt_Put(F,Approx(UU(J)));
         New_Line(F);

         Txt_Put(FF,Approx(Lambda*(X(J)+Scal(Half))-Scal(Half)));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Mu*(U(J)+Scal(Half))-Scal(Half)));
         New_Line(FF);

         Txt_Put(FF,Approx(Lambda*(XX(J)+Scal(Half))-Scal(Half)));
         Put(FF,"  ");
         Txt_Put(FF,Approx(Mu*(UU(J)+Scal(Half))-Scal(Half)));
         New_Line(FF);

      end loop;
      Close(FFF);
      Close(FF);
      Close(F);
   end DrawInclusion6;



procedure FindPoint(J: in Integer; D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3: in Scalar;  Check: in out Boolean;JJ1: out Integer) is
      use Ada.Text_IO;
      use Txt_Int_IO;

      SU,IU,SX,IX: Rep;
begin

   if Contains(UD1,Zero) and Contains(XD1,Zero) then
      Message("Doubling.FindPoint",Giving_Up);
   elsif Contains(XD1,Zero) and (not Contains(UD1,Zero)) then
      if (UD3/UD1>Scal(Zero) and (Sup(XD3)>Zero and Inf(XD3)<Zero)) or
                         (UD2/UD1>Scal(Zero) and (Sup(XD2)>Zero and Inf(XD2)<Zero))  then
         JJ1:=J;
         Check:=False;
         SU:=Rep'Max(Sup(UD2),Sup(UD3));
         IU:=Rep'Min(Inf(UD2),Inf(UD3));

         SX:=Rep'Max(Sup(XD2),Sup(XD3));
         IX:=Rep'Min(Inf(XD2),Inf(XD3));

         if (SU>Sup(UD1) and IU<Inf(UD1)) and (SX>Sup(XD1) and IX<Inf(XD1)) then
            if (D1>D2 or D1>D3) then
               New_Line;
                      Put("J="); Put(Current_Output,J,4);Put("  Containment does not hold, 1...");
            end if;
         end if;
      end if;

   elsif (not Contains(XD1,Zero)) and Contains(UD1,Zero) then
      if (XD3/XD1>Scal(Zero) and (Sup(UD3)>Zero and Inf(UD3)<Zero)) or
        (XD2/XD1>Scal(Zero) and (Sup(UD2)>Zero and Inf(UD2)<Zero))  then
                          JJ1:=J;
                          Check:=False;
                          SU:=Rep'Max(Sup(UD2),Sup(UD3));
                          IU:=Rep'Min(Inf(UD2),Inf(UD3));

                          SX:=Rep'Max(Sup(XD2),Sup(XD3));
                          IX:=Rep'Min(Inf(XD2),Inf(XD3));

                          if (SU>Sup(UD1) and IU<Inf(UD1)) and (SX>Sup(XD1) and IX<Inf(XD1)) then
                             if (D1>D2 or D1>D3) then
                                New_Line;
                                Put("J="); Put(Current_Output,J,4);Put("  Containment does not hold, 2...");
                             end if;
                          end if;
      end if;
   else
      if (XD3/XD1>Scal(Zero) and UD3/UD1>Scal(Zero) ) or
        (XD2/XD1>Scal(Zero) and UD2/UD1>Scal(Zero) ) then
         JJ1:=J;
         Check:=False;
         SU:=Rep'Max(Sup(UD2),Sup(UD3));
         IU:=Rep'Min(Inf(UD2),Inf(UD3));

                          SX:=Rep'Max(Sup(XD2),Sup(XD3));
                          IX:=Rep'Min(Inf(XD2),Inf(XD3));

                          if (SU>Sup(UD1) and IU<Inf(UD1)) and (SX>Sup(XD1) and IX<Inf(XD1)) then
                             if (D1>D2 or D1>D3) then
                                Put("D1=");Show(D1);
                                Put("D2=");Show(D2);
                                Put("D3=");Show(D3);

                                New_Line;
                                Put("J="); Put(Current_Output,J,4);Put("  Containment does not hold, 3...");
                             end if;
                          end if;
                       end if;

                    end if;

end FindPoint;



   procedure  DerBounds(N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar;CenX1,CenY1,Sx1,Su1,Rad1,Ecc1: in Scalar; MaxD1,MaxD3,MaxT,MinD: out Scalar) is
      --Checking containment of Lambda(G(C_1)) in C_3
      use Reps, Reps.IO;
      -- use Reps_Ops;
      use Ada.Text_IO;
      K: constant Integer:=2000;
      KK: constant Integer:=2000;
      KKK: constant Integer:=2000;
      B1: Scalar;
      X3,U3,Y3,Y1,Y2,SS1,SS2,VS1,VS2,Y,X,U,Det,AD,Er1,Er2,Ei1,Ei2,Rad,EE,EEE: Scalar;
      DF,DG: Matrix(1..2,1..2);
      Ang,Delt,RDelt: Scalar;
      II: Integer;
   begin



      B1:=Sqrt(Scal(One)-Sqr(Ecc1));
      Delt:=Two*ScalPi/Scal(K);
      RDelt:=Rad1/Scal(KK);
      Ang:=Scal(Zero,Sup(Delt));
      MaxT:=Scal(Zero);
      MaxD1:=Scal(Zero);
      MaxD3:=Scal(Zero);
      MinD:=Scal(1000.0);

      II:=0;
      for J in 1..K loop
         Rad:=Scal(Zero,Sup(RDelt));
         for I in 1..KK loop
            if Scal(100*II)/Scal(K*KK)=Scal(100*II/(KK*K)) then
               JS(Current_Output,100*II/(K*KK));
            end if;
            II:=II+1;

            X:=(Sx1+(Cos(Ang)*Rad));
            U:=(Su1+(B1*Sin(Ang)*Rad));

            -----------------------------
            --COMPUTING a on E_1

            U3:=(Mu)*(U+Scal(Half))-Scal(Half);
            Y3:=(Lambda)*(X+Scal(Half))-Scal(Half);


            DG(1,1):=Scal(One);
            DG(2,2):=Scal(One);
            DG(2,1):=Scal(Zero);
            DG(1,2):=Scal(Zero);

            for JJ in 1..N loop
               X3:=Y3;
               Map(Phi,S,B,C,E,Lambda,Mu,Y3,U3);
               VS1:=DerVal1(B,C,E,Y3,X3,S);
               VS2:=DerVal2(B,C,E,Y3,X3,S);

               Y1:=-VS2/VS1;
               Y2:=-Scal(One)/VS1;

               SS1:=DerVal1(B,C,E,X3,Y3,S);
               SS2:=DerVal2(B,C,E,X3,Y3,S);

               DF(1,1):=Y1;
               DF(1,2):=Y2;
               DF(2,1):=SS1+SS2*Y1;
               DF(2,2):=SS2*Y2;

               Mult(DF,DG);
            end loop;


            AD:=Lambda*DG(1,1)+Mu*DG(2,2);
            Det:=Sqr( Lambda*DG(1,1)-Mu*DG(2,2)) +Scal(4.0)*Lambda*DG(1,2)*Mu*DG(2,1);


            if Inf(Det)>Zero then
               Ei2:=ScalZero;
               Ei1:=ScalZero;
               Er2:=Half*(AD+Sqrt(Det));
               Er1:=Half*(AD-Sqrt(Det));
            elsif Sup(Det)<Zero then
               Ei2:=Half*Sqrt(-Det);
               Ei1:=-Half*Sqrt(-Det);
               Er2:=Half*AD;
               Er1:=Half*AD;
            else
               Ei2:= Half*Sqrt(Scal(Zero,-Inf(Det)));
               Ei1:=-Half*Sqrt(Scal(Zero,-Inf(Det)));
               Er2:=Half*(AD+Sqrt(Scal(Zero,Sup(Det))));
               Er1:=Half*(AD-Sqrt(Scal(Zero,Sup(Det))));
            end if;


            if  Sup(abs(Ei1))>Zero or   Sup(abs(Ei2))>Zero then
               Put("WARNING: COMPLEX EIGENVALUES:");
               New_Line;
               Put("E1=");Show(Er1,False);Put("+I*");Show(Ei1);
               Put("E2=");Show(Er2,False);Put("+I*");Show(Ei2);
            end if;

            EE:=Sqrt(Sqr(Er2)+Sqr(Ei2));
            EEE:=Sqrt(Sqr(Er1)+Sqr(Ei1));
            if Sup(EE)>Sup(MaxT) then
               MaxT:=EE;
            end if;
            if Sup(EEE)>Sup(MaxT) then
               MaxT:=EEE;
            end if;

            -----------------------------------
            ---COMPUTING A_1 and b on E_1

            DG(1,1):=Scal(One);
            DG(2,2):=Scal(One);
            DG(2,1):=Scal(Zero);
            DG(1,2):=Scal(Zero);

            Y:=X;
            for JJ in 1..N loop
               X:=Y;
               Map(Phi,S,B,C,E,Lambda,Mu,Y,U);
               VS1:=DerVal1(B,C,E,Y,X,S);
               VS2:=DerVal2(B,C,E,Y,X,S);

               Y1:=-VS2/VS1;
               Y2:=-Scal(One)/VS1;

               SS1:=DerVal1(B,C,E,X,Y,S);
               SS2:=DerVal2(B,C,E,X,Y,S);

               DF(1,1):=Y1;
               DF(1,2):=Y2;
               DF(2,1):=SS1+SS2*Y1;
               DF(2,2):=SS2*Y2;

               Mult(DF,DG);
            end loop;

            AD:=Lambda*DG(1,1)+Mu*DG(2,2);
            Det:=Sqr( Lambda*DG(1,1)-Mu*DG(2,2)) +Scal(4.0)*Lambda*DG(1,2)*Mu*DG(2,1);


            if Inf(Det)>Zero then
               Ei2:=ScalZero;
               Ei1:=ScalZero;
               Er2:=Half*(AD+Sqrt(Det));
               Er1:=Half*(AD-Sqrt(Det));
            elsif Sup(Det)<Zero then
               Ei2:=Half*Sqrt(-Det);
               Ei1:=-Half*Sqrt(-Det);
               Er2:=Half*AD;
               Er1:=Half*AD;
            else
               Ei2:= Half*Sqrt(Scal(Zero,-Inf(Det)));
               Ei1:=-Half*Sqrt(Scal(Zero,-Inf(Det)));
               Er2:=Half*(AD+Sqrt(Scal(Zero,Sup(Det))));
               Er1:=Half*(AD-Sqrt(Scal(Zero,Sup(Det))));
            end if;


            if  Sup(abs(Ei1))>Zero or   Sup(abs(Ei2))>Zero then
               Put("WARNING: COMPLEX EIGENVALUES:");
               New_Line;
               Put("E1=");Show(Er1,False);Put("+I*");Show(Ei1);
               Put("E2=");Show(Er2,False);Put("+I*");Show(Ei2);
            end if;

            EE:=Sqrt(Sqr(Er2)+Sqr(Ei2));
            EEE:=Sqrt(Sqr(Er1)+Sqr(Ei1));
            if Inf(EE)<Inf(MinD) then
               MinD:=EE;
            end if;
            if Inf(EEE)<Inf(MinD) then
               MinD:=EEE;
            end if;
            if Sup(EE)>Sup(MaxD1) then
               MaxD1:=EE;
            end if;
            if Sup(EEE)>Sup(MaxD1) then
               MaxD1:=EEE;
            end if;


            -----------------------------------
            ---COMPUTING A_3 and b on Lambda(E_2)

            U3:=(Mu)*(U+Scal(Half))-Scal(Half);
            Y3:=(Lambda)*(Y+Scal(Half))-Scal(Half);

            DG(1,1):=Scal(One);
            DG(2,2):=Scal(One);
            DG(2,1):=Scal(Zero);
            DG(1,2):=Scal(Zero);

            for JJ in 1..N loop
               X3:=Y3;
               Map(Phi,S,B,C,E,Lambda,Mu,Y3,U3);
               VS1:=DerVal1(B,C,E,Y3,X3,S);
               VS2:=DerVal2(B,C,E,Y3,X3,S);

               Y1:=-VS2/VS1;
               Y2:=-Scal(One)/VS1;

               SS1:=DerVal1(B,C,E,X3,Y3,S);
               SS2:=DerVal2(B,C,E,X3,Y3,S);

               DF(1,1):=Y1;
               DF(1,2):=Y2;
               DF(2,1):=SS1+SS2*Y1;
               DF(2,2):=SS2*Y2;

               Mult(DF,DG);
            end loop;

            AD:=Lambda*DG(1,1)+Mu*DG(2,2);
            Det:=Sqr( Lambda*DG(1,1)-Mu*DG(2,2)) +Scal(4.0)*Lambda*DG(1,2)*Mu*DG(2,1);


            if Inf(Det)>Zero then
               Ei2:=ScalZero;
               Ei1:=ScalZero;
               Er2:=Half*(AD+Sqrt(Det));
               Er1:=Half*(AD-Sqrt(Det));
            elsif Sup(Det)<Zero then
               Ei2:=Half*Sqrt(-Det);
               Ei1:=-Half*Sqrt(-Det);
               Er2:=Half*AD;
               Er1:=Half*AD;
            else
               Ei2:= Half*Sqrt(Scal(Zero,-Inf(Det)));
               Ei1:=-Half*Sqrt(Scal(Zero,-Inf(Det)));
               Er2:=Half*(AD+Sqrt(Scal(Zero,Sup(Det))));
               Er1:=Half*(AD-Sqrt(Scal(Zero,Sup(Det))));
            end if;


            if  Sup(abs(Ei1))>Zero or   Sup(abs(Ei2))>Zero then
               Put("WARNING: COMPLEX EIGENVALUES:");
               New_Line;
               Put("E1=");Show(Er1,False);Put("+I*");Show(Ei1);
               Put("E2=");Show(Er2,False);Put("+I*");Show(Ei2);
            end if;

            EE:=Sqrt(Sqr(Er2)+Sqr(Ei2));
            EEE:=Sqrt(Sqr(Er1)+Sqr(Ei1));
            if Inf(EE)<Inf(MinD) then
               MinD:=EE;
            end if;
            if Inf(EEE)<Inf(MinD) then
               MinD:=EEE;
            end if;
            if Sup(EE)>Sup(MaxD3) then
               MaxD3:=EE;
            end if;
            if Sup(EEE)>Sup(MaxD3) then
               MaxD3:=EEE;
            end if;
            ----------------------------------
            ---COMPUTING A_3 and b on Lambda(E_4)

            U3:=-Mu*(U+Scal(Half))-Scal(Half);
            Y3:=(Lambda)*(Y+Scal(Half))-Scal(Half);


            DG(1,1):=Scal(One);
            DG(2,2):=Scal(One);
            DG(2,1):=Scal(Zero);
            DG(1,2):=Scal(Zero);

            for JJ in 1..N loop
               X3:=Y3;
               Map(Phi,S,B,C,E,Lambda,Mu,Y3,U3);
               VS1:=DerVal1(B,C,E,Y3,X3,S);
               VS2:=DerVal2(B,C,E,Y3,X3,S);

               Y1:=-VS2/VS1;
               Y2:=-Scal(One)/VS1;

               SS1:=DerVal1(B,C,E,X3,Y3,S);
               SS2:=DerVal2(B,C,E,X3,Y3,S);

               DF(1,1):=Y1;
               DF(1,2):=Y2;
               DF(2,1):=SS1+SS2*Y1;
               DF(2,2):=SS2*Y2;

               Mult(DF,DG);
            end loop;

            AD:=Lambda*DG(1,1)+Mu*DG(2,2);
            Det:=Sqr( Lambda*DG(1,1)-Mu*DG(2,2)) +Scal(4.0)*Lambda*DG(1,2)*Mu*DG(2,1);

            if Inf(Det)>Zero then
               Ei2:=ScalZero;
               Ei1:=ScalZero;
               Er2:=Half*(AD+Sqrt(Det));
               Er1:=Half*(AD-Sqrt(Det));
            elsif Sup(Det)<Zero then
               Ei2:=Half*Sqrt(-Det);
               Ei1:=-Half*Sqrt(-Det);
               Er2:=Half*AD;
               Er1:=Half*AD;
            else
               Ei2:= Half*Sqrt(Scal(Zero,-Inf(Det)));
               Ei1:=-Half*Sqrt(Scal(Zero,-Inf(Det)));
               Er2:=Half*(AD+Sqrt(Scal(Zero,Sup(Det))));
               Er1:=Half*(AD-Sqrt(Scal(Zero,Sup(Det))));
            end if;


            if  Sup(abs(Ei1))>Zero or   Sup(abs(Ei2))>Zero then
               Put("WARNING: COMPLEX EIGENVALUES:");
               New_Line;
               Put("E1=");Show(Er1,False);Put("+I*");Show(Ei1);
               Put("E2=");Show(Er2,False);Put("+I*");Show(Ei2);
            end if;

            EE:=Sqrt(Sqr(Er2)+Sqr(Ei2));
            EEE:=Sqrt(Sqr(Er1)+Sqr(Ei1));
            if Inf(EE)<Inf(MinD) then
               MinD:=EE;
            end if;
            if Inf(EEE)<Inf(MinD) then
               MinD:=EEE;
            end if;
            if Sup(EE)>Sup(MaxD3) then
               MaxD3:=EE;
            end if;
            if Sup(EEE)>Sup(MaxD3) then
               MaxD3:=EEE;
            end if;

            if Sup(MaxD3)>Sup(MaxT) then
               MaxT:=MaxD3;
            end if;


            -----------------------------------
            ---COMPUTING a on Lambda(E_2)

            U3:=Sqr(Mu)*(U+Scal(Half))-Scal(Half);
            Y3:=Sqr(Lambda)*(Y+Scal(Half))-Scal(Half);



            DG(1,1):=Scal(One);
            DG(2,2):=Scal(One);
            DG(2,1):=Scal(Zero);
            DG(1,2):=Scal(Zero);

            for JJ in 1..N loop
               X3:=Y3;
               Map(Phi,S,B,C,E,Lambda,Mu,Y3,U3);
               VS1:=DerVal1(B,C,E,Y3,X3,S);
               VS2:=DerVal2(B,C,E,Y3,X3,S);

               Y1:=-VS2/VS1;
               Y2:=-Scal(One)/VS1;

               SS1:=DerVal1(B,C,E,X3,Y3,S);
               SS2:=DerVal2(B,C,E,X3,Y3,S);

               DF(1,1):=Y1;
               DF(1,2):=Y2;
               DF(2,1):=SS1+SS2*Y1;
               DF(2,2):=SS2*Y2;

               Mult(DF,DG);
            end loop;


            ---COMPUTING A_3 and b on E_3

            AD:=Lambda*DG(1,1)+Mu*DG(2,2);
            Det:=Sqr( Lambda*DG(1,1)-Mu*DG(2,2)) +Scal(4.0)*Lambda*DG(1,2)*Mu*DG(2,1);


            if Inf(Det)>Zero then
               Ei2:=ScalZero;
               Ei1:=ScalZero;
               Er2:=Half*(AD+Sqrt(Det));
               Er1:=Half*(AD-Sqrt(Det));
            elsif Sup(Det)<Zero then
               Ei2:=Half*Sqrt(-Det);
               Ei1:=-Half*Sqrt(-Det);
               Er2:=Half*AD;
               Er1:=Half*AD;
            else
               Ei2:= Half*Sqrt(Scal(Zero,-Inf(Det)));
               Ei1:=-Half*Sqrt(Scal(Zero,-Inf(Det)));
               Er2:=Half*(AD+Sqrt(Scal(Zero,Sup(Det))));
               Er1:=Half*(AD-Sqrt(Scal(Zero,Sup(Det))));
            end if;


            if  Sup(abs(Ei1))>Zero or   Sup(abs(Ei2))>Zero then
               Put("WARNING: COMPLEX EIGENVALUES:");
               New_Line;
               Put("E1=");Show(Er1,False);Put("+I*");Show(Ei1);
               Put("E2=");Show(Er2,False);Put("+I*");Show(Ei2);
            end if;

            EE:=Sqrt(Sqr(Er2)+Sqr(Ei2));
            EEE:=Sqrt(Sqr(Er1)+Sqr(Ei1));
            if Sup(EE)>Sup(MaxT) then
               MaxT:=EE;
            end if;
            if Sup(EEE)>Sup(MaxT) then
               MaxT:=EEE;
            end if;
            ----------------------------------
            ---COMPUTING a on Lambda(E_4)

            U3:=-Sqr(Mu)*(U+Scal(Half))-Scal(Half);
            Y3:=Sqr(Lambda)*(Y+Scal(Half))-Scal(Half);


            DG(1,1):=Scal(One);
            DG(2,2):=Scal(One);
            DG(2,1):=Scal(Zero);
            DG(1,2):=Scal(Zero);

            for JJ in 1..N loop
               X3:=Y3;
               Map(Phi,S,B,C,E,Lambda,Mu,Y3,U3);
               VS1:=DerVal1(B,C,E,Y3,X3,S);
               VS2:=DerVal2(B,C,E,Y3,X3,S);

               Y1:=-VS2/VS1;
               Y2:=-Scal(One)/VS1;

               SS1:=DerVal1(B,C,E,X3,Y3,S);
               SS2:=DerVal2(B,C,E,X3,Y3,S);

               DF(1,1):=Y1;
               DF(1,2):=Y2;
               DF(2,1):=SS1+SS2*Y1;
               DF(2,2):=SS2*Y2;

               Mult(DF,DG);
            end loop;

            AD:=Lambda*DG(1,1)+Mu*DG(2,2);
            Det:=Sqr( Lambda*DG(1,1)-Mu*DG(2,2)) +Scal(4.0)*Lambda*DG(1,2)*Mu*DG(2,1);

            if Inf(Det)>Zero then
               Ei2:=ScalZero;
               Ei1:=ScalZero;
               Er2:=Half*(AD+Sqrt(Det));
               Er1:=Half*(AD-Sqrt(Det));
            elsif Sup(Det)<Zero then
               Ei2:=Half*Sqrt(-Det);
               Ei1:=-Half*Sqrt(-Det);
               Er2:=Half*AD;
               Er1:=Half*AD;
            else
               Ei2:= Half*Sqrt(Scal(Zero,-Inf(Det)));
               Ei1:=-Half*Sqrt(Scal(Zero,-Inf(Det)));
               Er2:=Half*(AD+Sqrt(Scal(Zero,Sup(Det))));
               Er1:=Half*(AD-Sqrt(Scal(Zero,Sup(Det))));
            end if;


            if  Sup(abs(Ei1))>Zero or   Sup(abs(Ei2))>Zero then
               Put("WARNING: COMPLEX EIGENVALUES:");
               New_Line;
               Put("E1=");Show(Er1,False);Put("+I*");Show(Ei1);
               Put("E2=");Show(Er2,False);Put("+I*");Show(Ei2);
            end if;

            EE:=Sqrt(Sqr(Er2)+Sqr(Ei2));
            EEE:=Sqrt(Sqr(Er1)+Sqr(Ei1));
            if Sup(EE)>Sup(MaxT) then
               MaxT:=EE;
            end if;
            if Sup(EEE)>Sup(MaxT) then
               MaxT:=EEE;
            end if;

            <<HERE>> null;

            Rad:=Scal(Sup(Rad),Sup( Scal(Sup(Rad))+Scal(Sup(RDelt))) );
         end loop;
         Ang:=Scal(Sup(Ang),Sup( Scal(Sup(Ang))+Scal(Sup(Delt))) );
      end loop;
   end DerBounds;



   procedure Containment0(N: in Integer; Phi,S: in Taylor;
B,C,E,Lambda,Mu,Sx0,Su0,Sx1,Su1,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1,LsN0,LsP0,LuN0,LuP0,LsN1,LsP1,LuN1,LuP1: in
Scalar;
CenX1,CenY1,S1,C1,Rad1,Ecc1: in Scalar) is
      use Reps, Reps.IO;
      -- use Reps_Ops;
      use Ada.Text_IO;
      K: constant Integer:=1000;
      KK: constant Integer:=500;
      KKK: constant Integer:=10;
      X1,U1:  Vector(0..K-1);
      XX0,UU0,XX1,UU1:  Vector(0..4*K-1);
      B1,Xc1,Uc1,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3: Scalar;
      Ang,Delt,EDel0,XDel0,EDel1,XDel1,Xi0,Eta0,Xi1,Eta1: Scalar;
      I1,K1,K2: Integer;
      JJ0: Integer:=0;
      JJ1: Integer:=0;
      Check: Boolean;
   begin
      B1:=Sqrt(Scal(One)-Sqr(Ecc1));
      Delt:=Two*ScalPi/Scal(X1'Length);
      Ang:=Scal(Zero,Sup(Delt));
      for J in X1'Range loop
         X1(J):=S1+Cos(Ang)*Rad1;
         U1(J):=C1+B1*Sin(Ang)*Rad1;
         Ang:=Scal(Sup(Ang),Sup( Scal(Sup(Ang))+Scal(Sup(Delt))) );
      end loop;


      XDel0:=(LsP0+LsN0)/Scal(K);
      EDel0:=(LuP0+LuN0)/Scal(K);
      XDel1:=(LsP1+LsN1)/Scal(K);
      EDel1:=(LuP1+LuN1)/Scal(K);

      Xi0:=-(LsN0);
      Eta0:=-(LuN0);
      Xi1:=-(LsN1);
      Eta1:=-(LuN1);
      for J in 0..K-1 loop
          XX0(J):=Eta0*Wx0+Xi0*Vx0+Sx0;
          UU0(J):=Eta0*Wu0+Xi0*Vu0+Su0;
          XX1(J):=Eta1*Wx1+Xi1*Vx1+Sx1;
          UU1(J):=Eta1*Wu1+Xi1*Vu1+Su1;

          XX0(J):=Lambda*(XX0(J)+Scal(Half))-Scal(Half);
          XX1(J):=Lambda*(XX1(J)+Scal(Half))-Scal(Half);
          UU0(J):=Mu*(UU0(J)+Scal(Half))-Scal(Half);
          UU1(J):=Mu*(UU1(J)+Scal(Half))-Scal(Half);

          Eta0:=Scal(Sup(Eta0),Sup(Scal(Sup(EDel0))+Scal(Sup(Eta0))));
          Eta1:=Scal(Sup(Eta1),Sup(Scal(Sup(EDel1))+Scal(Sup(Eta1))));
      end loop;


      Xi0:=-(LsN0);
      Eta0:=(LuP0);
      Xi1:=-(LsN1);
      Eta1:=(LuP1);
      for J in K-1..2*K-1 loop
          XX0(J):=Eta0*Wx0+Xi0*Vx0+Sx0;
          UU0(J):=Eta0*Wu0+Xi0*Vu0+Su0;
          XX1(J):=Eta1*Wx1+Xi1*Vx1+Sx1;
          UU1(J):=Eta1*Wu1+Xi1*Vu1+Su1;

          XX0(J):=Lambda*(XX0(J)+Scal(Half))-Scal(Half);
          XX1(J):=Lambda*(XX1(J)+Scal(Half))-Scal(Half);
          UU0(J):=Mu*(UU0(J)+Scal(Half))-Scal(Half);
          UU1(J):=Mu*(UU1(J)+Scal(Half))-Scal(Half);

          Xi0:=Scal(Sup(Xi0),Sup(Scal(Sup(XDel0))+Scal(Sup(Xi0))));
          Xi1:=Scal(Sup(Xi1),Sup(Scal(Sup(XDel1))+Scal(Sup(Xi1))));
      end loop;


      Xi0:=(LsP0);
      Eta0:=(LuP0);
      Xi1:=(LsP1);
      Eta1:=(LuP1);
      for J in 2*K-1..3*K-1 loop
          XX0(J):=Eta0*Wx0+Xi0*Vx0+Sx0;
          UU0(J):=Eta0*Wu0+Xi0*Vu0+Su0;
          XX1(J):=Eta1*Wx1+Xi1*Vx1+Sx1;
          UU1(J):=Eta1*Wu1+Xi1*Vu1+Su1;

          XX0(J):=Lambda*(XX0(J)+Scal(Half))-Scal(Half);
          XX1(J):=Lambda*(XX1(J)+Scal(Half))-Scal(Half);
          UU0(J):=Mu*(UU0(J)+Scal(Half))-Scal(Half);
          UU1(J):=Mu*(UU1(J)+Scal(Half))-Scal(Half);

          Eta0:=Scal(Inf(-Scal(Sup(EDel0))+Scal(Inf(Eta0))),Inf(Eta0));
          Eta1:=Scal(Inf(-Scal(Sup(EDel1))+Scal(Inf(Eta1))), Inf(Eta1));
      end loop;


      Xi0:=(LsP0);
      Eta0:=-(LuN0);
      Xi1:=(LsP1);
      Eta1:=-(LuN1);
      for J in 3*K-1..4*K-1 loop
          XX0(J):=Eta0*Wx0+Xi0*Vx0+Sx0;
          UU0(J):=Eta0*Wu0+Xi0*Vu0+Su0;
          XX1(J):=Eta1*Wx1+Xi1*Vx1+Sx1;
          UU1(J):=Eta1*Wu1+Xi1*Vu1+Su1;

          XX0(J):=Lambda*(XX0(J)+Scal(Half))-Scal(Half);
          XX1(J):=Lambda*(XX1(J)+Scal(Half))-Scal(Half);
          UU0(J):=Mu*(UU0(J)+Scal(Half))-Scal(Half);
          UU1(J):=Mu*(UU1(J)+Scal(Half))-Scal(Half);

          Xi0:=Scal(Inf(-Scal(Sup(XDel0))+Scal(Inf(Xi0))),Inf(Xi0));
          Xi1:=Scal(Inf(-Scal(Sup(XDel1))+Scal(Inf(Xi1))),Inf(Xi1));
      end loop;


      Xc1:=Scal(-0.55);
      Uc1:=Scal(-0.5);

      for J in XX1'Range loop
         UD1:=(UU0(J)-Uc1);
         XD1:=(XX0(J)-Xc1);
         D1:=Sqrt(Sqr(UD1)+Sqr(XD1));
         UD1:=UD1/D1;
         XD1:=XD1/D1;

         if J=XX1'First then K1:=X1'First; K2:=X1'Last;
         else K1:=JJ0-KK;K2:=JJ0+KK;
         end if;
         Check:=True;

         for I in K1..K2 loop
            I1:=(I mod X1'Length);
            UD3:=(U1(I1)-Uc1);
            XD3:=(X1(I1)-Xc1);
            D3:=Sqrt(Sqr(UD3)+Sqr(XD3));

            UD3:=UD3/D3;
            XD3:=XD3/D3;
            UD2:=(U1((I1+1) mod X1'Length)-Uc1);
            XD2:=(X1((I1+1) mod X1'Length)-Xc1);
            D2:=Sqrt(Sqr(UD2)+Sqr(XD2));

            UD2:=UD2/D2;
            XD2:=XD2/D2;

            FindPoint(J,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3,Check,JJ0);
         end loop;
         if Check then
            New_Line;
            Put("Point for component 0 in Containment0 not found...");
         end if;



         UD1:=(UU1(J)-Uc1);
         XD1:=(XX1(J)-Xc1);
         D1:=Sqrt(Sqr(UD1)+Sqr(XD1));
         UD1:=UD1/D1;
         XD1:=XD1/D1;

         if J=XX1'First then K1:=X1'First; K2:=X1'Last;
         else K1:=JJ1-KK;K2:=JJ1+KK;
         end if;
         Check:=True;

         for I in K1..K2 loop
            I1:=(I mod X1'Length);
            UD3:=(U1(I1)-Uc1);
            XD3:=(X1(I1)-Xc1);
            D3:=Sqrt(Sqr(UD3)+Sqr(XD3));
            UD3:=UD3/D3;
            XD3:=XD3/D3;
            UD2:=(U1((I1+1) mod X1'Length)-Uc1);
            XD2:=(X1((I1+1) mod X1'Length)-Xc1);
            D2:=Sqrt(Sqr(UD2)+Sqr(XD2));
            UD2:=UD2/D2;
            XD2:=XD2/D2;

            FindPoint(J,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3,Check,JJ1);
         end loop;
         if Check then
            New_Line;
            Put("Point for component 1 in Containment0 not found...");
         end if;
      end loop;

   end Containment0;



   procedure Containment1(N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar; CenX1,CenY1,S1,C1,Rad1,Ecc1:
in Scalar) is
      use Reps, Reps.IO;
      -- use Reps_Ops;
      use Ada.Text_IO;
      K: constant Integer:=2000;
      KK: constant Integer:=1000;
      KKK: constant Integer:=1;
      X1,U1:  Vector(0..K-1);
      XX1,UU1:  Vector(0..KKK*K-1);
      B1,Xc1,Uc1,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3: Scalar;
      Ang,Delt,Angg,Deltt: Scalar;
      II,I1,K1,K2: Integer;
      JJ1: Integer:=0;
      Check: Boolean;
   begin
      B1:=Sqrt(Scal(One)-Sqr(Ecc1));
      Delt:=Two*ScalPi/Scal(X1'Length);
      Deltt:=Two*ScalPi/Scal(XX1'Length);
      Ang:=Scal(Zero,Sup(Delt));
      Angg:=Scal(Zero,Sup(Deltt));
      II:=0;
      for J in X1'Range loop
         X1(J):=S1+Cos(Ang)*Rad1;
         U1(J):=C1+B1*Sin(Ang)*Rad1;
         for I in 0..KKK-1 loop
           XX1(II):=S1+Cos(Angg)*Rad1;
           UU1(II):=C1+B1*Sin(Angg)*Rad1;
           Angg:=Scal(Sup(Angg),Sup( Scal(Sup(Angg))+Scal(Sup(Deltt))) );
           II:=II+1;
         end loop;
         Ang:=Scal(Sup(Ang),Sup( Scal(Sup(Ang))+Scal(Sup(Delt))) );
      end loop;
      Xc1:=S1;
      Uc1:=C1;

      for J in XX1'Range loop
         for II in 1..N loop
            --            XX1(J):=Center(XX1(J));
            --            UU1(J):=Center(UU1(J));
            Map(Phi,S,B,C,E,Lambda,Mu,XX1(J),UU1(J));
         end loop;

         XX1(J):=Sqr(Lambda)*(XX1(J)+Scal(Half))-Scal(Half);
         UU1(J):=Sqr(Mu)*(UU1(J)+Scal(Half))-Scal(Half);
      end loop;

      for II in 1..N loop
         Map(Phi,S,B,C,E,Lambda,Mu,Xc1,Uc1);
      end loop;
      Xc1:=Center(Sqr(Lambda)*(Xc1+Scal(Half))-Scal(Half));
      Uc1:=Center(Sqr(Mu)*(Uc1+Scal(Half))-Scal(Half));
      Xc1:=Scal(-0.65);
      Uc1:=Scal(-0.5);

      for J in XX1'Range loop
         UD1:=(UU1(J)-Uc1);
         XD1:=(XX1(J)-Xc1);
         D1:=Sqrt(Sqr(UD1)+Sqr(XD1));
         UD1:=UD1/D1;
         XD1:=XD1/D1;

         if J=XX1'First then K1:=X1'First; K2:=X1'Last;
         else K1:=JJ1-KK;K2:=JJ1+KK;
         end if;
         Check:=True;

         for I in K1..K2 loop
            I1:=(I mod X1'Length);
            UD3:=(U1(I1)-Uc1);
            XD3:=(X1(I1)-Xc1);
            D3:=Sqrt(Sqr(UD3)+Sqr(XD3));
            UD3:=UD3/D3;
            XD3:=XD3/D3;
            UD2:=(U1((I1+1) mod X1'Length)-Uc1);
            XD2:=(X1((I1+1) mod X1'Length)-Xc1);
            D2:=Sqrt(Sqr(UD2)+Sqr(XD2));
            UD2:=UD2/D2;
            XD2:=XD2/D2;

            FindPoint(J,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3,Check,JJ1);
         end loop;
         if Check then
            Put("Point in Containment1 not found...");
         end if;
      end loop;
   end Containment1;


   procedure Containment2(N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar;CenX1,CenY1,S1,C1,Rad1,Ecc1: in Scalar) is
      use Reps, Reps.IO;
      -- use Reps_Ops;
      use Ada.Text_IO;
      K: constant Integer:=2000;
      KK: constant Integer:=2000;
      KKK: constant Integer:=1;
      X1,U1:  Vector(0..K-1);
      XX1,UU1,UM1:  Vector(0..KKK*K-1);
      B1,Xc1,Uc1,Umc1,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3: Scalar;
      Ang,Delt,Angg,Deltt: Scalar;
      II,I1,K1,K2: Integer;
      JJ1: Integer:=0;
      JJ2: Integer:=0;
      Check: Boolean;
   begin
      B1:=Sqrt(Scal(One)-Sqr(Ecc1));
      Delt:=Two*ScalPi/Scal(X1'Length);
      Deltt:=Two*ScalPi/Scal(XX1'Length);
      Ang:=Scal(Zero,Sup(Delt));
      Angg:=Scal(Zero,Sup(Deltt));
      II:=0;
      for J in X1'Range loop
         X1(J):=(S1+(Cos(Ang)*Rad1));
         U1(J):=(C1+(B1*Sin(Ang)*Rad1));
         Ang:=Scal(Sup(Ang),Sup( Scal(Sup(Ang))+Scal(Sup(Delt))) );
      end loop;

      for I in 1..2 loop
         for J in XX1'Range loop
            for II in 1..N loop
               Map(Phi,S,B,C,E,Lambda,Mu,X1(J),U1(J));
            end loop;


            if I=1 then


               XX1(J):=Lambda*(X1(J)+Scal(Half))-Scal(Half);
               UU1(J):=Mu*(U1(J)+Scal(Half))-Scal(Half);
               UM1(J):=-Mu*(U1(J)+Scal(Half))-Scal(Half);
            end if;
         end loop;
      end loop;

      Xc1:=Scal(-0.76);
      Uc1:=Scal(-0.48);
      Umc1:=Scal(-0.48);

      for J in XX1'Range loop
         UD1:=(UU1(J)-Uc1);
         XD1:=(XX1(J)-Xc1);
         D1:=Sqrt(Sqr(UD1)+Sqr(XD1));
         UD1:=UD1/D1;
         XD1:=XD1/D1;

         if J=XX1'First then K1:=X1'First; K2:=X1'Last;
         else K1:=JJ1-KK;K2:=JJ1+KK;
         end if;
         Check:=True;

         for I in K1..K2 loop
            I1:=(I mod X1'Length);
            UD3:=(U1(I1)-Uc1);
            XD3:=(X1(I1)-Xc1);
            D3:=Sqrt(Sqr(UD3)+Sqr(XD3));
            UD3:=UD3/D3;
            XD3:=XD3/D3;
            UD2:=(U1((I1+1) mod X1'Length)-Uc1);
            XD2:=(X1((I1+1) mod X1'Length)-Xc1);
            D2:=Sqrt(Sqr(UD2)+Sqr(XD2));

            UD2:=UD2/D2;
            XD2:=XD2/D2;

            FindPoint(J,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3,Check,JJ1);
         end loop;
            if Check then
               Put("Point in Containment2 not found...");
            end if;


            UD1:=(UM1(J)-Umc1);
            XD1:=(XX1(J)-Xc1);
            D1:=Sqrt(Sqr(UD1)+Sqr(XD1));
            UD1:=UD1/D1;
            XD1:=XD1/D1;

            if J=XX1'First then K1:=X1'First; K2:=X1'Last;
            else K1:=JJ2-KK;K2:=JJ2+KK;
            end if;
            Check:=True;

            for I in K1..K2 loop
               I1:=(I mod X1'Length);
               UD3:=(U1(I1)-Umc1);
               XD3:=(X1(I1)-Xc1);
               D3:=Sqrt(Sqr(UD3)+Sqr(XD3));
               UD3:=UD3/D3;
               XD3:=XD3/D3;
               UD2:=(U1((I1+1) mod X1'Length)-Uc1);
               XD2:=(X1((I1+1) mod X1'Length)-Xc1);
               D2:=Sqrt(Sqr(UD2)+Sqr(XD2));
               UD2:=UD2/D2;
               XD2:=XD2/D2;

               FindPoint(J,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3,Check,JJ2);
            end loop;
            if Check then
               Put("Point in Containment2 not found...");
            end if;

      end loop;
   end Containment2;




   procedure Containment3(N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar;
CenX1,CenY1,S1,C1,Rad1,Ecc1: in Scalar) is
      use Reps, Reps.IO;
      -- use Reps_Ops;
      use Ada.Text_IO;
      K: constant Integer:=1000;
      KK: constant Integer:=500;
      KKK: constant Integer:=1;
      X1,U1:  Vector(0..K-1);
      XX1,UU1:  Vector(0..KKK*K-1);
      B1,Xc1,Uc1,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3: Scalar;
      Ang,Delt,Angg,Deltt: Scalar;
      II,I1,K1,K2: Integer;
      JJ1: Integer:=0;
      Check: Boolean;
   begin
      B1:=Sqrt(Scal(One)-Sqr(Ecc1));
      Delt:=Two*ScalPi/Scal(X1'Length);
      Deltt:=Two*ScalPi/Scal(XX1'Length);
      Ang:=Scal(Zero,Sup(Delt));
      Angg:=Scal(Zero,Sup(Deltt));
      II:=0;
      for J in X1'Range loop
         X1(J):=S1+Cos(Ang)*Rad1;
         U1(J):=C1+B1*Sin(Ang)*Rad1;
         XX1(J):=Lambda*(X1(J)+Scal(Half))-Scal(Half);
         UU1(J):=Mu*(U1(J)+Scal(Half))-Scal(Half);
         Ang:=Scal(Sup(Ang),Sup( Scal(Sup(Ang))+Scal(Sup(Delt))) );
      end loop;

      Xc1:=Center(Lambda*(S1+Scal(Half))-Scal(Half));
      Uc1:=Center(Mu*(C1+Scal(Half))-Scal(Half));

      for J in XX1'Range loop
         UD1:=(UU1(J)-Uc1);
         XD1:=(XX1(J)-Xc1);
         D1:=Sqrt(Sqr(UD1)+Sqr(XD1));
         UD1:=UD1/D1;
         XD1:=XD1/D1;

         if J=XX1'First then K1:=X1'First; K2:=X1'Last;
         else K1:=JJ1-KK;K2:=JJ1+KK;
         end if;
         Check:=True;

         for I in K1..K2 loop
            I1:=(I mod X1'Length);
            UD3:=(U1(I1)-Uc1);
            XD3:=(X1(I1)-Xc1);
            D3:=Sqrt(Sqr(UD3)+Sqr(XD3));
            UD3:=UD3/D3;
            XD3:=XD3/D3;
            UD2:=(U1((I1+1) mod X1'Length)-Uc1);
            XD2:=(X1((I1+1) mod X1'Length)-Xc1);
            D2:=Sqrt(Sqr(UD2)+Sqr(XD2));
            UD2:=UD2/D2;
            XD2:=XD2/D2;

            FindPoint(J,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3,Check,JJ1);
         end loop;
         if Check then
            Put("Point in Containment3 not found...");
         end if;
      end loop;
   end Containment3;


   procedure Containment4(N: in Integer; Phi,S: in Taylor; B,C,E,Lambda,Mu: in Scalar;
                                                           CenX1,CenY1,S1,C1,Rad1,Ecc1: in Scalar) is
      use Reps, Reps.IO;
      -- use Reps_Ops;
      use Ada.Text_IO;
      K: constant Integer:= 20000;
      KK: constant Integer:=10000;
      KKK: constant Integer:=1;
      X1,U1:  Vector(0..K-1);
      XX1,UU1,UM1,XM1:  Vector(0..KKK*K-1);
      B1,Xc1,Uc1,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3: Scalar;
      Ang,Delt,Angg,Deltt: Scalar;
      II,I1,K1,K2: Integer;
      JJ1: Integer:=0;
      JJ2: Integer:=0;
      Check: Boolean;
   begin
      B1:=Sqrt(Scal(One)-Sqr(Ecc1));
      Delt:=Two*ScalPi/Scal(X1'Length);
      Deltt:=Two*ScalPi/Scal(XX1'Length);
      Ang:=Scal(Zero,Sup(Delt));
      Angg:=Scal(Zero,Sup(Deltt));
      II:=0;
      for J in X1'Range loop
         XX1(J):=S1+(Cos(Ang)*Rad1);
         UU1(J):=C1+(B1*Sin(Ang)*Rad1);
         --          for I in 0..KKK-1 loop
         --            XX1(II):=S1+Cos(Angg)*Rad1;
         --            UU1(II):=C1+B1*Sin(Angg)*Rad1;
         --            Angg:=Scal(Sup(Angg),Sup( Scal(Sup(Angg))+Scal(Sup(Deltt))) );
         --            II:=II+1;
         --          end loop;
         --XX1(J):=Center(XX1(J));
         --UU1(J):=Center(UU1(J));
         Ang:=Scal(Sup(Ang),Sup( Scal(Sup(Ang))+Scal(Sup(Delt))) );
      end loop;
      Xc1:=S1;
      Uc1:=C1;

      for J in XX1'Range loop
         for II in 1..N loop
            Map(Phi,S,B,C,E,Lambda,Mu,XX1(J),UU1(J));
         end loop;
         X1(J):=XX1(J);
         U1(J):=-(UU1(J)+Scal(Half))-Scal(Half);

         XX1(J):=Lambda*(X1(J)+Scal(Half))-Scal(Half);
         UU1(J):=-Mu*(U1(J)+Scal(Half))-Scal(Half);

         XM1(J):=Lambda*(X1(J)+Scal(Half))-Scal(Half);
         UM1(J):=Mu*(U1(J)+Scal(Half))-Scal(Half);
      end loop;

      for II in 1..N loop
         Map(Phi,S,B,C,E,Lambda,Mu,Xc1,Uc1);
      end loop;

      --Xc1:=(Lambda*(Xc1+Scal(Half))-Scal(Half));
      --Uc1:=(Mu*(Uc1+Scal(Half))-Scal(Half));

      --for II in 1..N loop
      --Map(Phi,S,B,C,E,Lambda,Mu,Xc1,Uc1);
      --end loop;

      --Xc1:=Center(Xc1);
      --Uc1:=Center(-(Uc1+Scal(Half))-Scal(Half));
      --Show(Xc1);
      --Show(Uc1);
      Xc1:=Scal(0.515);
      Uc1:=Scal(-0.55);
      --      Xc1:=Scal(0.4);
      --      Uc1:=Scal(-0.5);

      for J in XX1'Range loop
         for II in 1..N loop
            Map(Phi,S,B,C,E,Lambda,Mu,XX1(J),UU1(J));
            Map(Phi,S,B,C,E,Lambda,Mu,XM1(J),UM1(J));
         end loop;
         --                  XX1(J):=Center(XX1(J));
         --                  UU1(J):=Center(UU1(J));
         --                  XM1(J):=Center(XM1(J));
         --                  UM1(J):=Center(UM1(J));
         --                  X1(J):=Center(X1(J));
         --                  U1(J):=Center(U1(J));
      end loop;

      for J in XX1'Range loop
         UD1:=(UU1(J)-Uc1);
         XD1:=(XX1(J)-Xc1);
         D1:=Sqrt(Sqr(UD1)+Sqr(XD1));
         UD1:=UD1/D1;
         XD1:=XD1/D1;

         if J=XX1'First then K1:=X1'First; K2:=X1'Last;
         else K1:=JJ1-KK;K2:=JJ1+KK;
         end if;
         Check:=True;

         for I in K1..K2 loop
            I1:=(I mod X1'Length);
            UD3:=(U1(I1)-Uc1);
            XD3:=(X1(I1)-Xc1);
            D3:=Sqrt(Sqr(UD3)+Sqr(XD3));
            UD3:=UD3/D3;
            XD3:=XD3/D3;
            UD2:=(U1((I1+1) mod X1'Length)-Uc1);
            XD2:=(X1((I1+1) mod X1'Length)-Xc1);
            D2:=Sqrt(Sqr(UD2)+Sqr(XD2));
            UD2:=UD2/D2;
            XD2:=XD2/D2;

            FindPoint(J,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3,Check,JJ1);
         end loop;
         if Check then
            Put("Part 1, point in Containment4 not found...");
         end if;

         UD1:=(UM1(J)-Uc1);
         XD1:=(XM1(J)-Xc1);
         D1:=Sqrt(Sqr(UD1)+Sqr(XD1));
         UD1:=UD1/D1;
         XD1:=XD1/D1;

         if J=XX1'First then K1:=X1'First; K2:=X1'Last;
         else K1:=JJ2-KK;K2:=JJ2+KK;
         end if;
         Check:=True;

         for I in K1..K2 loop
            I1:=(I mod X1'Length);
            UD3:=(U1(I1)-Uc1);
            XD3:=(X1(I1)-Xc1);
            D3:=Sqrt(Sqr(UD3)+Sqr(XD3));
            UD3:=UD3/D3;
            XD3:=XD3/D3;
            UD2:=(U1((I1+1) mod X1'Length)-Uc1);
            XD2:=(X1((I1+1) mod X1'Length)-Xc1);
            D2:=Sqrt(Sqr(UD2)+Sqr(XD2));
            UD2:=UD2/D2;
            XD2:=XD2/D2;

            FindPoint(J,D1,D2,D3,UD1,XD1,UD2,XD2,UD3,XD3,Check,JJ2);
         end loop;
         if Check then
            Put("Part 2, point in Containment4 not found...");
         end if;

      end loop;
   end Containment4;



   procedure Containment5(Name: in String; Q,N: in Integer; LsP,LsN,LuP,LuN: in Rep; Lambda,Mu,P1,P2,CenX,CenY,Wx,Wu,Vx,Vu,S1,C1,S2,C2,Ecc1,Ecc2,Rad1,Rad2: in Scalar; Phi,S: in Taylor) is --FINE

      use Reps.IO;
      use Ada.Text_IO;

      X,Y,U,YY,UU,EDel,XDel,Xi,Eta,B1,B2: Scalar;
      F: File_Type;
      K: Integer:=10000;
      X1,U1: Vector(1..K);
      JJJ: Integer:=8;
   begin

      B1:=Sqrt(Scal(One)-Sqr(Ecc1));
      B2:=Sqrt(Scal(One)-Sqr(Ecc2));

      Create(F,Out_File,Name);
      for J in X1'Range loop
         X1(J):=S1+Cos(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad1;
         U1(J):=C1+B1*Sin(Two*Scal(J)*Scal(Rep(3.14159265))/Scal(K))*Rad1;

         Txt_Put(F,Approx(X1(J)));
         Put(F,"  ");
         Txt_Put(F,Approx(U1(J)));
         New_Line(F);
      end loop;

      XDel:=Rep(1.0)*(Scal(LsP+LsN))/Rep(Q);
      EDel:=Rep(1.0)*(Scal(LuP+LuN))/Rep(Q);

      --Upper side


      Eta:=Scal(LuP);
      Xi:=-Scal(LsN);
      for I in 1..Q loop
         X:=Eta*Wx+Xi*Vx;
         U:=Eta*Wu+Xi*Vu;
         Y:=X+P1;
         U:=U+P2;

         YY:=(Lambda**N)*(Y+Scal(Half))-Scal(Half);
         UU:=(Mu**N)*(U+Scal(Half))-Scal(Half);
         Txt_Put(F,Approx(YY));
         Put(F,"  ");
         Txt_Put(F,Approx(UU));
         New_Line(F);

         Xi:=Xi+XDel;

      end loop;

      Eta:=-Scal(LuN);
      Xi:=-Scal(LsN);
      for I in 1..Q loop
         X:=Eta*Wx+Xi*Vx;
         U:=Eta*Wu+Xi*Vu;
         Y:=X+P1;
         U:=U+P2;

         YY:=(Lambda**N)*(Y+Scal(Half))-Scal(Half);
         UU:=(Mu**N)*(U+Scal(Half))-Scal(Half);
         Txt_Put(F,Approx(YY));
         Put(F,"  ");
         Txt_Put(F,Approx(UU));
         New_Line(F);

         Xi:=Xi+XDel;
      end loop;

      Eta:=-Scal(LuN);
      Xi:= Scal(LsP);
      for I in 1..Q loop
         X:=Eta*Wx+Xi*Vx;
         U:=Eta*Wu+Xi*Vu;
         Y:=X+P1;
         U:=U+P2;


         YY:=(Lambda**N)*(Y+Scal(Half))-Scal(Half);
         UU:=(Mu**N)*(U+Scal(Half))-Scal(Half);
         Txt_Put(F,Approx(YY));
         Put(F,"  ");
         Txt_Put(F,Approx(UU));
         New_Line(F);

         Eta:=Eta+EDel;
      end loop;

      Eta:=-Scal(LuN);
      Xi:= -Scal(LsN);
      for I in 1..Q loop
         X:=Eta*Wx+Xi*Vx;
         U:=Eta*Wu+Xi*Vu;
         Y:=X+P1;
         U:=U+P2;

         YY:=(Lambda**N)*(Y+Scal(Half))-Scal(Half);
         UU:=(Mu**N)*(U+Scal(Half))-Scal(Half);
         Txt_Put(F,Approx(YY));
         Put(F,"  ");
         Txt_Put(F,Approx(UU));
         New_Line(F);

         Eta:=Eta+EDel;
      end loop;
      Close(F);


   end Containment5;


   function SetY(J,Q,N: in Integer; LsP0,LsN0,LuP0,LuN0,LsP1,LsN1,LuP1,LuN1: in Rep; Lambda,Mu,Px0,Pu0,Px1,Pu1,CenX,CenY,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1: in Scalar; Phi,S: in Taylor) return Scalar is

      use Reps.IO;
      use Ada.Text_IO;

      X,Y,U,YY,EDel,XDel,Xi,Eta: Scalar;
      F: File_Type;
      K: Integer:=10000;
   begin

      --Upper side
      if J<=Q then
         XDel:=Rep(1.0)*(Scal(LsP0+LsN0))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP0+LuN0))/Rep(Q);
         Eta:=Scal(LuP0);
         Xi:=-Scal(LsN0);
         for I in 1..J loop
            X:=Eta*Wx0+Xi*Vx0;
            U:=Eta*Wu0+Xi*Vu0;
            Y:=X+Px0;
            U:=U+Pu0;
            Xi:=Xi+XDel;
         end loop;
      elsif J<=2*Q then
         XDel:=Rep(1.0)*(Scal(LsP0+LsN0))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP0+LuN0))/Rep(Q);
         Eta:=-Scal(LuN0);
         Xi:=-Scal(LsN0);
         for I in 1..(J-Q) loop
            X:=Eta*Wx0+Xi*Vx0;
            U:=Eta*Wu0+Xi*Vu0;
            Y:=X+Px0;
            U:=U+Pu0;
            Xi:=Xi+XDel;
         end loop;
      elsif J<=3*Q then
         XDel:=Rep(1.0)*(Scal(LsP0+LsN0))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP0+LuN0))/Rep(Q);
         Eta:=-Scal(LuN0);
         Xi:= Scal(LsP0);
         for I in 1..(J-2*Q) loop
            X:=Eta*Wx0+Xi*Vx0;
            U:=Eta*Wu0+Xi*Vu0;
            Y:=X+Px0;
            U:=U+Pu0;
            Eta:=Eta+EDel;
         end loop;
      elsif J<=4*Q then
         XDel:=Rep(1.0)*(Scal(LsP0+LsN0))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP0+LuN0))/Rep(Q);
         Eta:=-Scal(LuN0);
         Xi:= -Scal(LsN0);
         for I in 1..(J-3*Q) loop
            X:=Eta*Wx0+Xi*Vx0;
            U:=Eta*Wu0+Xi*Vu0;
            Y:=X+Px0;
            U:=U+Pu0;
            Eta:=Eta+EDel;
         end loop;
      elsif J<=5*Q then
         XDel:=Rep(1.0)*(Scal(LsP1+LsN1))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP1+LuN1))/Rep(Q);
         Eta:=Scal(LuP1);
         Xi:=-Scal(LsN1);
         for I in 1..J-(4*Q) loop
            X:=Eta*Wx1+Xi*Vx1;
            U:=Eta*Wu1+Xi*Vu1;
            Y:=X+Px1;
            U:=U+Pu1;
            Xi:=Xi+XDel;
         end loop;
      elsif J<=6*Q then
         XDel:=Rep(1.0)*(Scal(LsP1+LsN1))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP1+LuN1))/Rep(Q);
         Eta:=-Scal(LuN1);
         Xi:=-Scal(LsN1);
         for I in 1..(J-5*Q) loop
            X:=Eta*Wx1+Xi*Vx1;
            U:=Eta*Wu1+Xi*Vu1;
            Y:=X+Px1;
            U:=U+Pu1;
            Xi:=Xi+XDel;
         end loop;
      elsif J<=7*Q then
         XDel:=Rep(1.0)*(Scal(LsP1+LsN1))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP1+LuN1))/Rep(Q);
         Eta:=-Scal(LuN1);
         Xi:= Scal(LsP1);
         for I in 1..(J-6*Q) loop
            X:=Eta*Wx1+Xi*Vx1;
            U:=Eta*Wu1+Xi*Vu1;
            Y:=X+Px1;
            U:=U+Pu1;
            Eta:=Eta+EDel;
         end loop;
      elsif J<=8*Q then
         XDel:=Rep(1.0)*(Scal(LsP1+LsN1))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP1+LuN1))/Rep(Q);
         Eta:=-Scal(LuN1);
         Xi:= -Scal(LsN1);
         for I in 1..(J-7*Q) loop
            X:=Eta*Wx1+Xi*Vx1;
            U:=Eta*Wu1+Xi*Vu1;
            Y:=X+Px1;
            U:=U+Pu1;
            Eta:=Eta+EDel;
         end loop;
         end if;
         YY:=(Lambda**N)*(Y+Scal(Half))-Scal(Half);
         return YY;
   end SetY;


   function SetU(J,Q,N: in Integer; LsP0,LsN0,LuP0,LuN0,LsP1,LsN1,LuP1,LuN1: in Rep; Lambda,Mu,Px0,Pu0,Px1,Pu1,CenX,CenY,Wx0,Wu0,Vx0,Vu0,Wx1,Wu1,Vx1,Vu1: in Scalar; Phi,S: in Taylor) return Scalar is

      use Reps.IO;
      use Ada.Text_IO;

      X,Y,U,UU,EDel,XDel,Xi,Eta: Scalar;
      F: File_Type;
      K: Integer:=10000;
   begin

      --Upper side
      if J<=Q then
         XDel:=Rep(1.0)*(Scal(LsP0+LsN0))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP0+LuN0))/Rep(Q);
         Eta:=Scal(LuP0);
         Xi:=-Scal(LsN0);
         for I in 1..J loop
            X:=Eta*Wx0+Xi*Vx0;
            U:=Eta*Wu0+Xi*Vu0;
            Y:=X+Px0;
            U:=U+Pu0;
            Xi:=Xi+XDel;
         end loop;
      elsif J<=2*Q then
         XDel:=Rep(1.0)*(Scal(LsP0+LsN0))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP0+LuN0))/Rep(Q);
         Eta:=-Scal(LuN0);
         Xi:=-Scal(LsN0);
         for I in 1..(J-Q) loop
            X:=Eta*Wx0+Xi*Vx0;
            U:=Eta*Wu0+Xi*Vu0;
            Y:=X+Px0;
            U:=U+Pu0;
            Xi:=Xi+XDel;
         end loop;
      elsif J<=3*Q then
         XDel:=Rep(1.0)*(Scal(LsP0+LsN0))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP0+LuN0))/Rep(Q);
         Eta:=-Scal(LuN0);
         Xi:= Scal(LsP0);
         for I in 1..(J-2*Q) loop
            X:=Eta*Wx0+Xi*Vx0;
            U:=Eta*Wu0+Xi*Vu0;
            Y:=X+Px0;
            U:=U+Pu0;
            Eta:=Eta+EDel;
         end loop;
      elsif J<=4*Q then
         XDel:=Rep(1.0)*(Scal(LsP0+LsN0))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP0+LuN0))/Rep(Q);
         Eta:=-Scal(LuN0);
         Xi:= -Scal(LsN0);
         for I in 1..(J-3*Q) loop
            X:=Eta*Wx0+Xi*Vx0;
            U:=Eta*Wu0+Xi*Vu0;
            Y:=X+Px0;
            U:=U+Pu0;
            Eta:=Eta+EDel;
         end loop;
      elsif J<=5*Q then
         XDel:=Rep(1.0)*(Scal(LsP1+LsN1))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP1+LuN1))/Rep(Q);
         Eta:=Scal(LuP1);
         Xi:=-Scal(LsN1);
         for I in 1..J-(4*Q) loop
            X:=Eta*Wx1+Xi*Vx1;
            U:=Eta*Wu1+Xi*Vu1;
            Y:=X+Px1;
            U:=U+Pu1;
            Xi:=Xi+XDel;
         end loop;
      elsif J<=6*Q then
         XDel:=Rep(1.0)*(Scal(LsP1+LsN1))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP1+LuN1))/Rep(Q);
         Eta:=-Scal(LuN1);
         Xi:=-Scal(LsN1);
         for I in 1..(J-5*Q) loop
            X:=Eta*Wx1+Xi*Vx1;
            U:=Eta*Wu1+Xi*Vu1;
            Y:=X+Px1;
            U:=U+Pu1;
            Xi:=Xi+XDel;
         end loop;
 elsif J<=7*Q then
         XDel:=Rep(1.0)*(Scal(LsP1+LsN1))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP1+LuN1))/Rep(Q);
         Eta:=-Scal(LuN1);
         Xi:= Scal(LsP1);
         for I in 1..(J-6*Q) loop
            X:=Eta*Wx1+Xi*Vx1;
            U:=Eta*Wu1+Xi*Vu1;
            Y:=X+Px1;
            U:=U+Pu1;
            Eta:=Eta+EDel;
         end loop;
    elsif J<=8*Q then
         XDel:=Rep(1.0)*(Scal(LsP1+LsN1))/Rep(Q);
         EDel:=Rep(1.0)*(Scal(LuP1+LuN1))/Rep(Q);
         Eta:=-Scal(LuN1);
         Xi:= -Scal(LsN1);
         for I in 1..(J-7*Q) loop
            X:=Eta*Wx1+Xi*Vx1;
            U:=Eta*Wu1+Xi*Vu1;
            Y:=X+Px1;
            U:=U+Pu1;
            Eta:=Eta+EDel;
         end loop;
         end if;
         UU:=(Mu**N)*(U+Scal(Half))-Scal(Half);
         return UU;
   end SetU;




   end Doubling;

